//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////
//
// FColorSelect.C - A simple button which allows to select the activated
//                  trace for a given color
//
// $Id: FColorSelect.C,v 1.3 2003/03/12 13:35:05 aarn Exp $
//
////////////////////////////////////////////////////////////////////////

#include "FColorSelect.moc"
#include "FColorSelect.h"
#include <minmax.h>

#include <qpainter.h>
#include <cstdio>
#include <algorithm>

FColorSelect::FColorSelect ( int num, QWidget * parent, const char * name )
  : QPushButton(parent, name), playerNo(-1), leftTeam(true)
  , autoExpLevel(-1), myNumber(num)
{
  buildMenus();
  setText("R Coach LDW"); // probabely biggest text to appear
  setFixedSize( sizeHint() );
  buttonPixm= new QPixmap( sizeHint() );
  
  col= FColors::agentColor( num );
  FColors::Flags f= FColors::agentFlags( num );

  setWMTrace(f.WM);
  setLogTrace(f.log);
  setSidePlayer( f.leftTeam, f.playerNo ); // update label done automatically

  setDrawTrace(f.drawTraces);
  setWMTrace(f.WM);
  setLogTrace(f.log);
  setAutoExpLevel(f.autoExpLevel);

  if (f.active)
    activate();
  else
    deactivate();

  connect( this, SIGNAL(clicked()), SLOT(popup()) );
}

FColorSelect::~FColorSelect()
{
  delete buttonPixm;
}

bool FColorSelect::isDrawTrace(int playerNo, bool leftTeam)
{
  return isActive() && ( playerNo == getPlayerNo() ) &&
    (leftTeam == isSideLeft()) && hasDrawTrace();
}

bool FColorSelect::isWMTrace(int playerNo, bool leftTeam)
{
  return isActive() && ( playerNo == getPlayerNo() ) &&
    (leftTeam == isSideLeft()) && hasWMTrace();
}

bool FColorSelect::isLogTrace(int playerNo, bool leftTeam)
{
  return isActive() && ( playerNo == getPlayerNo() ) &&
    (leftTeam == isSideLeft()) && hasLogTrace();
}

void FColorSelect::setSidePlayer( const bool left, const int no)
{
  bool stchng(false);
  // have to change team ?
  if (left != leftTeam) {
    stchng= true;
    leftTeam= left;
    if (leftTeam) 
      mainPopUp->changeItem("left team", 1);
    else
      mainPopUp->changeItem("right team", 1);
  }
  
  if (no != playerNo) {
    stchng= true;
    // delete old check
    playerNoMenu->setItemChecked( playerNo, false);
    playerNo=no;
    // set new check
    playerNoMenu->setItemChecked( playerNo, true);
    
    if (playerNo == 0)
      mainPopUp->changeItem("Coach", 2);
    else {
      char buf[20]= "";
      sprintf(buf, "Player %d", playerNo);
      mainPopUp->changeItem(buf, 2);
    }
  }
  if (stchng)
    activate();
  updateLabel();
}

void FColorSelect::toggleActive()
{
  mainPopUp->setItemChecked( 0, !isActive() );
  updateLabel();
}

void FColorSelect::toggleDrawTrace()
{
  mainPopUp->setItemChecked( 3, !hasDrawTrace() );
  activate();
  updateLabel();
}

void FColorSelect::toggleWM()
{
  mainPopUp->setItemChecked( 4, !hasWMTrace() );
  activate();
  updateLabel();
}

void FColorSelect::toggleLog()
{
  mainPopUp->setItemChecked( 5, !hasLogTrace() );
  activate();
  updateLabel();
}

void FColorSelect::setAutoExpLevel( const int level )
{
  int l = max(min(level,10),0);
  if (l != autoExpLevel) {
    autoExpMenu->setItemChecked( autoExpLevel, false);
    autoExpLevel=l;
    // set new check
    autoExpMenu->setItemChecked( l, true );
    emit sigAutoExpLevelChanged( myNumber, l );
  }
}

void FColorSelect::popup()
{
  // show the popup menu always at the same position
  mainPopUp->popup(mapToGlobal(rect().topRight()));
}

void FColorSelect::buildMenus()
{
  playerNoMenu= new QPopupMenu;
  playerNoMenu->insertItem("Coach", this, SLOT( setPlayer0() ),0 , 0 );
  playerNoMenu->insertItem("Player 1", this, SLOT( setPlayer1() ), 0, 1 );
  playerNoMenu->insertItem("Player 2", this, SLOT( setPlayer2() ), 0, 2 );
  playerNoMenu->insertItem("Player 3", this, SLOT( setPlayer3() ), 0, 3 );
  playerNoMenu->insertItem("Player 4", this, SLOT( setPlayer4() ), 0, 4 );
  playerNoMenu->insertItem("Player 5", this, SLOT( setPlayer5() ), 0, 5 );
  playerNoMenu->insertItem("Player 6", this, SLOT( setPlayer6() ), 0, 6 );
  playerNoMenu->insertItem("Player 7", this, SLOT( setPlayer7() ), 0, 7 );
  playerNoMenu->insertItem("Player 8", this, SLOT( setPlayer8() ), 0, 8 );
  playerNoMenu->insertItem("Player 9", this, SLOT( setPlayer9() ), 0, 9 );
  playerNoMenu->insertItem("Player 10", this, SLOT( setPlayer10() ), 0, 10 );
  playerNoMenu->insertItem("Player 11", this, SLOT( setPlayer11() ) , 0, 11 );
  playerNoMenu->setCheckable(true);

  autoExpMenu = new QPopupMenu;
  autoExpMenu->insertItem("don't expand", this, SLOT( setAutoExpLevel0() ),0 , 0 );
  autoExpMenu->insertItem("Level 1", this, SLOT( setAutoExpLevel1() ), 0, 1 );
  autoExpMenu->insertItem("Level 2", this, SLOT( setAutoExpLevel2() ), 0, 2 );
  autoExpMenu->insertItem("Level 3", this, SLOT( setAutoExpLevel3() ), 0, 3 );
  autoExpMenu->insertItem("Level 4", this, SLOT( setAutoExpLevel4() ), 0, 4 );
  autoExpMenu->insertItem("Level 5", this, SLOT( setAutoExpLevel5() ), 0, 5 );
  autoExpMenu->insertItem("Level 6", this, SLOT( setAutoExpLevel6() ), 0, 6 );
  autoExpMenu->insertItem("Level 7", this, SLOT( setAutoExpLevel7() ), 0, 7 );
  autoExpMenu->insertItem("Level 8", this, SLOT( setAutoExpLevel8() ), 0, 8 );
  autoExpMenu->insertItem("Level 9", this, SLOT( setAutoExpLevel9() ), 0, 9 );
  autoExpMenu->insertItem("expand all", this, SLOT( setAutoExpLevelInf() ), 0, 10 );
  autoExpMenu->setCheckable(true);

  mainPopUp= new QPopupMenu;
  mainPopUp->insertItem("active", this, SLOT( toggleActive() ), 0, 0 );
  mainPopUp->insertItem("left Team", this, SLOT( changeTeam() ), 0, 1 );
  mainPopUp->insertItem("Coach", playerNoMenu, 2);
  mainPopUp->insertItem("draw trace", this, SLOT( toggleDrawTrace() ), 0, 3 );
  mainPopUp->insertItem("Worldmodel", this, SLOT( toggleWM() ), 0, 4 );
  mainPopUp->insertItem("Logfile", this, SLOT( toggleLog() ), 0, 5 );
  mainPopUp->insertItem("Autoexpand Level", autoExpMenu, 6);
  mainPopUp->setCheckable(true);
}

void FColorSelect::updateLabel()
{
  char buf[20]= "";
  
  if(!isActive())
    strcpy(buf, "disabled");
  else {
    if (leftTeam)
      strcpy(buf, "L ");
    else 
      strcpy(buf, "R ");
    if (playerNo == 0)
      strcpy(buf+2, "Coach ");
    else 
      sprintf(buf+2, "Pl. %2d ", playerNo);
    if (hasDrawTrace())
      strcat(buf, "D");
    if (hasWMTrace())
      strcat(buf, "W");
    if (hasLogTrace())
      strcat(buf, "L");
  }

  QPainter p( buttonPixm, this );
  p.fillRect( buttonPixm->rect(), QBrush( backgroundColor() ) );
  p.setPen( QPen( col ) );
  p.drawText( rect(), AlignCenter, buf);
  p.end();

  setPixmap( *buttonPixm );
  
  emit stateChanged();
}
