//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// FLogWindow.h - Implements a simple logfile window using a QListView
//
// $Id: FLogWindow.C,v 1.1.1.1 2002/10/08 12:40:43 cm Exp $
//
////////////////////////////////////////////////////////////////////////

#include "FLogWindow.moc"
#include "FLogWindow.h"
#include "Trace.h"

#include <stdio.h>

void
FLogWindow::FLogToolTip::maybeTip( const QPoint & p ) {
  QListView * parent = dynamic_cast<QListView*>(parentWidget());
  QPoint pp = parent->viewport()->mapFromParent(p);
  QListViewItem * lvi = parent->itemAt(pp);

  if (lvi && lvi->depth() > 1) { // level 0, 1 contains player resp. module
    QRect r = parent->itemRect(lvi);
    if (r.isValid()) {
      QPoint tl = parent->viewport()->mapToParent(r.topLeft());
      QPoint br = parent->viewport()->mapToParent(r.bottomRight());
      FLogItem * fli = dynamic_cast<FLogItem*>(lvi);
      if (fli)
	tip(QRect(tl,br), fli->file+"("+fli->line+"): "+fli->function);
    }
  }
}

FLogWindow::FLogWindow(QWidget * parent, const char * name )
  : QListView( parent, name ), maxEntryIcons(FColors::numAgentColors())
    , lastLVI(0)
{
  entryIcons= new (QPixmap*)[FColors::numAgentColors()+1];
  autoExpLevel= new int[FColors::numAgentColors()+1];

  for (int i=-1; i < FColors::numAgentColors(); ++i) {
    entryIcons[i+1]= FColors::createColorIcon(i);
    autoExpLevel[i+1]=0;
  }

  addColumn("Time");
  addColumn("Logfile entry");

  contextMenu= new QPopupMenu;
  contextMenu->insertItem("Clear", this, SLOT(clearList()), 0, 0);
  contextMenu->insertItem("Only one step", this, SLOT(toggleClearMode()),
			  0, 1);
  contextMenu->setItemChecked(1, true);

  contextMenu->insertItem("autoexpand subtree", this, SLOT(toggleExpMode()),
			  0, 2);
  contextMenu->setItemChecked(2, true);
#ifdef QT2
  connect(this, SIGNAL(expanded(QListViewItem *)), SLOT(sigExpanded(QListViewItem *)));
#else
  connect(this, SIGNAL(doubleClicked(QListViewItem *)), SLOT(sigExpanded(QListViewItem *)));
#endif

  setCaption("FUNSSELMon LogWindow");
  setRootIsDecorated(true);
  setSorting(-1);

  resize(Config::getValueAsInt("FMonLogSizeX"),
	 Config::getValueAsInt("FMonLogSizeY"));
  
  for (int i = 0; i < 13; ++i) {
    players[i]=0;
    for (int j = 0; j < Trace::NUM_MODULES; ++j)
      modules[i][j]=0;
  }
  tt = new FLogToolTip(this);
}

FLogWindow::~FLogWindow()
{
  delete [] entryIcons;
  delete tt;
  delete [] autoExpLevel;
}

void FLogWindow::addLine( const int time, const int color, 
			  const int plNum, const char *msg)
{
  char buf[20] = "";
  int col( color>-1 ? color < maxEntryIcons ? color+1 : 0 : 0 );
  sprintf(buf,"%5.5d", time);
  QListViewItem *lvi= 
  new QListViewItem( this, lastLVI );
  lastLVI = lvi;
  lvi->setText(0,buf);
  lvi->setPixmap(1,*(entryIcons[col]));
  char *buf2= new char[strlen(msg)+10];  
  if (plNum > -1) {
    sprintf(buf2, "%d ", plNum);
  } else {
    strcpy(buf2, "??");
  }
  strncat(buf2, msg, strlen(msg)-1);
  lvi->setText(1, buf2);
}

void FLogWindow::addLine( const int time, const int color, 
			  const int plNum, const int module,
			  const int level, const char * file,
			  const char * line, const char * func,
			  const char * msg)
{
  char time_str[6] = "";
  char pl_str[3];
  int col( color>-1 ? color < maxEntryIcons ? color+1 : 0 : 0 );
  sprintf(time_str,"%5.5d", time);
  if (plNum > -1) {
    sprintf(pl_str, "%2d ", plNum);
  } else {
    strcpy(pl_str, "??");
  }
  
  int mod = min(max(module, 0), Trace::NUM_MODULES-1);
  int lev = max(level, 1);
  int pl  = min(max(plNum, 0), 12);

   // if we have a msg from a non existent player
  if (players[pl] == 0) {
    QListViewItem * plvi = new QListViewItem( this );
    players[pl] = plvi;
    plvi->setText(0, QString("Player ")+pl_str);
    plvi->setPixmap(0,*(entryIcons[col]));
    plvi->setText(1, "");
  }

  // if we have a msg from a non existent module
  if (modules[pl][mod] == 0) {
    QListViewItem * lvi = new QListViewItem( players[pl] );
    modules[pl][mod] = lvi;
    lvi->setPixmap(0,*(entryIcons[col]));
    lvi->setText(0, Trace::moduleStrings[mod].name);
    if (lvi->depth() <= autoExpLevel[max(col-1,0)]) {
      blockSignals(true); // to make sure subtree is not expanded
      players[pl]->setOpen(true);
      blockSignals(false);
    }

    FLogItem * lvi2 = new FLogItem(lev, lvi);
    lvi2->setPixmap(0,*(entryIcons[col]));
    lvi2->setText(0, time_str);
    lvi2->setText(1, msg);
    lvi2->file = file;
    lvi2->line = line;
    lvi2->function = func;
    if (lvi2->depth() <= autoExpLevel[max(col-1,0)]) {
      blockSignals(true);
      lvi->setOpen(true);
      blockSignals(false);
    }
  // we have a msg from an existing module
  } else {
    QListViewItem * fli = modules[pl][mod]->firstChild();
    while (true) {
      while (fli->nextSibling())
	fli = fli->nextSibling();
      if (((FLogItem*)fli)->level < lev) {
	if (fli->firstChild())
	  fli = fli->firstChild();
	else {
	  FLogItem * lvi = new FLogItem(lev, fli);
	  lvi->setText(0, time_str);
	  lvi->setPixmap(0, *(entryIcons[col]));
	  lvi->setText(1, msg);
	  lvi->file = file;
	  lvi->line = line;
	  lvi->function = func;
	  if (lvi->depth() <= autoExpLevel[max(col-1,0)]) {
	    blockSignals(true);
	    fli->setOpen(true);
	    blockSignals(false);
	  }
	  break;
	}
      } else {
	FLogItem * lvi = new FLogItem(lev, fli->parent(), fli);
	lvi->setPixmap(0, *(entryIcons[col]));
	lvi->setText(0, time_str);
	lvi->setText(1, msg);
	lvi->file = file;
	lvi->line = line;
	lvi->function = func;
	if (lvi->depth() <= autoExpLevel[max(col-1,0)]) {
	  blockSignals(true);
	  fli->parent()->setOpen(true);
	  blockSignals(false);
	}
	break;
      }
    }
  }
}

void FLogWindow::sigExpanded( QListViewItem * item ) 
{
  if (autoExpand() && item) {
    QListViewItem * lvi = item->firstChild();
    while (lvi) {
      if (lvi->isOpen()) 
#ifdef QT2
	emit expanded(lvi);
#else
        emit doubleClicked(lvi);
#endif
      else
	lvi->setOpen(true);
      lvi = lvi->nextSibling();
    }
  }
}

#ifdef QT2
void FLogWindow::contentsMouseReleaseEvent ( QMouseEvent * e )
#else
void FLogWindow::mouseReleaseEvent ( QMouseEvent * e )
#endif
{
  if (e->button() == RightButton)
    contextMenu->popup(e->globalPos());

  QListView::mouseReleaseEvent( e );
}
