#include <qapplication.h>
#include <qlcdnumber.h>
#include "FColors.h"
#include "Config.h"
#include "Exception.h"
#include "FieldMap.h"
#include "FieldObject.h"

#include <unistd.h>
#include "../libMRB/player.h"
#include "../libMRB/coordinates.h"
#include "../MasterControl/grade.h"
#include "../modules/passmodule/passeval.h"
#include "../modules/goalshotmodule/goalshoteval.h"
#include "../modules/positionmodule/PositionEval.h"
#include "AdvSkills.h"

#include <qwidget.h>

#include <pthread.h>

using namespace std;

// cwz-build-command: make RealPlayer

#define COUT(out)

//#define RPDEBUG(out) cout << out << endl

#ifndef RPDEBUG
#define RPDEBUG(out)
#endif

////////////////////////////////////////////////////////////////
//
//     thread part
//
/////////////////////////////////////////////////////////


enum PlCommand {
  PlNone,
  PlGo2,
  PlfastGo2,
  PlKick2,
  PlfastKick2,
  PlIntercept,
  PlDribble
};

enum LookCommand {
  LkNone,
  LkPos
};

typedef struct {
  pthread_mutex_t     mutex;  /* Protects access to value */
  bool                finished; 
  Worldmodel          *Snap;  /* Access protected by mutex */
  Vector2d            Pos;
  enum PlCommand      Command;
  Vector2d            LookPos;
  enum LookCommand     LkCommand;
} commun_struct_t;

commun_struct_t data = {
  PTHREAD_MUTEX_INITIALIZER,0,NULL,Vector2d(0,0),PlNone};

pthread_t Player;


///////////////////////////////////////////////////////////////////////
//
//      PlayGround Part
//
///////////////////////////////////////////////////////////////////////

class PlayGround : public FieldMap
{
public:
  PlayGround( QWidget *parent = 0, const char *name = 0):
    FieldMap(parent,name),wIsOpen(1) ,modifier(0)
  {    
    setFocusPolicy(StrongFocus);
    setEnabled(true);
    setFocus();
  };

  ~PlayGround(){  };
  
  void deleteList( list<FieldObject *> * &);
  
  bool isOpen()
  {return wIsOpen;};

  //  void pollSnapshots();

  int getCommand();
  int getLkCommand();

protected:
  void      keyPressEvent ( QKeyEvent * );  
  void      keyReleaseEvent ( QKeyEvent * );  
  void      mousePressEvent ( QMouseEvent * );  
  void      mouseDoubleClickEvent ( QMouseEvent * );  
  //  void      mouseMoveEvent ( QMouseEvent * );
  void      mouseReleaseEvent ( QMouseEvent *);
  void      closeEvent ( QCloseEvent *e);
  bool wIsOpen;
  char modifier;
};

void PlayGround::closeEvent ( QCloseEvent *e)
{
  wIsOpen= false;
};

void PlayGround::keyPressEvent ( QKeyEvent * qke )
{
  if (qke->isAutoRepeat())
    return;

  switch (qke->key()) {
  case 'L':
  case 'I':
  case 'D':
    modifier = qke->key();
    break;
  default: break;
  }
}

void PlayGround::keyReleaseEvent ( QKeyEvent * qke )
{
  if (qke->isAutoRepeat())
    return;
  if (qke->key() != modifier)
    return;
  modifier = 0;
}

void PlayGround::mousePressEvent ( QMouseEvent * qme )
{
  int Status;
  Status = pthread_mutex_lock (&data.mutex);
  if (Status != 0) {fprintf(stderr,"Lock error!\n"); exit(-1);}
      
  enum PlCommand Command = data.Command;
  enum LookCommand LkCommand = data.LkCommand;

  Status = pthread_mutex_unlock (&data.mutex);
  if (Status != 0) {fprintf(stderr,"Unlock error!\n"); exit(-1);}

  mouseDownPos= qme->pos();
  mouseDragPos= qme->pos();
  mouseDownTime= time;
  if (qme->button() == LeftButton && modifier == 0) 
    Command = PlGo2;
  else if (qme->button() == LeftButton && modifier == 'D')
    Command = PlDribble;
  else if (qme->button() == RightButton ||
	   (qme->button() == LeftButton && modifier == 'I')) 
    Command = PlIntercept;
  else if (qme->button() == MidButton ||
	   (qme->button() == LeftButton && modifier == 'L'))
    LkCommand = LkPos;

  Status = pthread_mutex_lock (&data.mutex);
  if (Status != 0) {fprintf(stderr,"Lock error!\n"); exit(-1);}

  data.Command = Command;
  data.LkCommand = LkCommand;
  data.Pos = getFC(qme->pos());

  Status = pthread_mutex_unlock (&data.mutex);
  if (Status != 0) {fprintf(stderr,"Unlock error!\n"); exit(-1);}
};

void PlayGround::mouseDoubleClickEvent ( QMouseEvent * qme )
{
  int Status;
  Status = pthread_mutex_lock (&data.mutex);
  if (Status != 0) {fprintf(stderr,"Lock error!\n"); exit(-1);}
      
  enum PlCommand Command = data.Command;
  enum LookCommand LkCommand = data.LkCommand;

  Status = pthread_mutex_unlock (&data.mutex);
  if (Status != 0) {fprintf(stderr,"Unlock error!\n"); exit(-1);}

  
  mouseDownPos= qme->pos();
  mouseDragPos= qme->pos();
  mouseDownTime= time;

  if (qme->button() == LeftButton && modifier == 0) 
      Command = PlfastGo2;
  else if (qme->button() == MidButton || (qme->button() == LeftButton && modifier == 'L')) 
    LkCommand = LkNone;

  Status = pthread_mutex_lock (&data.mutex);
  if (Status != 0) {fprintf(stderr,"Lock error!\n"); exit(-1);}

  data.Command = Command;
  data.LkCommand = LkCommand;
  data.Pos = getFC(qme->pos());

  Status = pthread_mutex_unlock (&data.mutex);
  if (Status != 0) {fprintf(stderr,"Unlock error!\n"); exit(-1);}
};

void PlayGround::mouseReleaseEvent ( QMouseEvent * qme )
{
  int Status;
  Status = pthread_mutex_lock (&data.mutex);
  if (Status != 0) {fprintf(stderr,"Lock error!\n"); exit(-1);}
      
  enum PlCommand Command = data.Command;

  Status = pthread_mutex_unlock (&data.mutex);
  if (Status != 0) {fprintf(stderr,"Unlock error!\n"); exit(-1);}

  if (qme->button() == RightButton) 
    Command=PlNone;

  Status = pthread_mutex_lock (&data.mutex);
  if (Status != 0) {fprintf(stderr,"Lock error!\n"); exit(-1);}

  data.Command = Command;

  Status = pthread_mutex_unlock (&data.mutex);
  if (Status != 0) {fprintf(stderr,"Unlock error!\n"); exit(-1);}
};


///////////////////////////////////////////////////////////////////////
//
//      PlayerMWidget
//
//////////////////////////////////////////////////////////////////////


class PlayerMWidget : public QWidget
{
    Q_OBJECT
public:
  PlayerMWidget( QWidget * parent=0, const char * name=0);
  ~PlayerMWidget();

  bool isOpen()
  {return wIsOpen;};

  void pollSnapshots();

protected:
  void      closeEvent ( QCloseEvent *e);

  virtual void resizeEvent ( QResizeEvent * );

private:
  void deleteList( list<FieldObject *>  * &l);
  char * Team;

  QLabel *leftTeam;
  QLabel *rightTeam;
  QLabel *timeDisp;
  QLabel *playMode;
  QLCDNumber *Stamina;
  QLabel *mode;

  PlayGround *Ground;
  bool wIsOpen;
  list<FieldObject *> * plobjs;
  FieldPlayer *Me;
};

PlayerMWidget::PlayerMWidget(QWidget * parent, const char * name) :
  QWidget(parent,name), wIsOpen(1)
{
  Ground=new PlayGround(this,"PlayGround");
  plobjs=0;

  leftTeam= new QLabel("0 left Team", this);
  leftTeam->setBackgroundColor( FColors::leftTeam.color() );
  leftTeam->setAlignment(AlignLeft | AlignVCenter);
  leftTeam->resize(leftTeam->sizeHint());

  rightTeam= new QLabel("0 other Team", this);
  rightTeam->setBackgroundColor( FColors::rightTeam.color() );
  rightTeam->setAlignment(AlignRight | AlignVCenter);
  rightTeam->resize(rightTeam->sizeHint());  

  timeDisp= new QLabel(" -2    ", this);
  timeDisp->setAlignment(AlignCenter);
  timeDisp->resize(timeDisp->sizeHint());

  playMode= new QLabel(" dies ist ein ganz langer Text   ", this);
  playMode->setAlignment(AlignCenter);
  playMode->resize(playMode->sizeHint());

  Stamina  = new QLCDNumber( 4, this, "stamina"  );

  mode= new QLabel("No Command set/No Look Set", this);
  mode->setAlignment(AlignCenter);
  mode->resize(mode->sizeHint());

  resize(Ground->width(),Ground->height()+30);
}

PlayerMWidget::~PlayerMWidget()
{
  delete Ground;
};

void PlayerMWidget::resizeEvent ( QResizeEvent * )
{
  int wid=(width()-2*5-timeDisp->width()-playMode->width())/2;

  leftTeam->setGeometry(1,1,wid,leftTeam->height());
  timeDisp->move( leftTeam->x()+leftTeam->width()+5,1);
  playMode->move( timeDisp->x()+timeDisp->width()+5,1);
  rightTeam->setGeometry(width()-wid,1,wid,rightTeam->height());
  Stamina->move(1,leftTeam->height()+5);
  mode->move(Stamina->x() + Stamina->width() + 5,leftTeam->height()+5);
  Ground->setGeometry( 1, 60, width(), height()-60 );
};
  

void PlayerMWidget::deleteList( list<FieldObject *>  * &l)
{
  if (!l)
    return;
  // check if there is a list
  for( list<FieldObject *>::iterator it= l->begin();
       it != l->end(); it++)
    delete (*it);
  // this cleans up pointers using a clear():
  delete l;
  l=0;
}

void PlayerMWidget::pollSnapshots()
{
  Worldmodel *PS= 0;
  int Status;
  Vector2d            Pos;
  enum PlCommand      Command;   
  enum LookCommand    LkCommand;
  Vector2d            LookPos;
  

  Status = pthread_mutex_lock (&data.mutex);
  if (Status != 0) {fprintf(stderr,"Lock error!\n"); exit(-1);}
  
  PS=data.Snap;
  data.Snap=0;
  Command=data.Command;

  Pos=data.Pos;
  LookPos=data.LookPos;
  LkCommand=data.LkCommand;
    
  Status = pthread_mutex_unlock (&data.mutex);
  if (Status != 0) {fprintf(stderr,"Unlock error!\n"); exit(-1);}

  if (PS!=0)
    {
      //      printf("Got Snapshot!!\n");
      
      char str[40];
      sprintf(str,"%i",(int)PS->time);
      timeDisp->setText(str);

      sprintf(str,"%i our team",PS->ourScore);
      leftTeam->setText(str);
      
      sprintf(str,"%i other team",PS->otherScore);
      rightTeam->setText(str);

      stringstream aux;
      aux << PS->playMode << char(0);
      const char *auxstr=aux.str().c_str();
      sscanf(auxstr,"PLAY_MODE_%s",str);
      playMode->setText(str);

      Stamina->display(PS->my.stamina);
      QString txt;
      switch (Command) {
      case PlNone: txt = "None"; break;
      case PlGo2: txt = "Go2"; break;
      case PlfastGo2: txt = "FastGo2"; break;
      case PlKick2: txt = "Kick2"; break;
      case PlfastKick2: txt = "FastKick2"; break;
      case PlIntercept: txt = "Intercept"; break;
      case PlDribble: txt = "Dribble"; break;
      }
      switch (LkCommand) {
      case LkNone: txt += " None"; break;
      case LkPos:  txt += " LookPos"; break;
      }

      mode->setText(txt);
      deleteList( plobjs );
      plobjs = new list<FieldObject *>;

      if ((Command==PlGo2) || (Command==PlfastGo2) || (Command=PlIntercept))
	plobjs->push_back( new FieldMarker( Pos,0,6666,""));
      if ((LkCommand==LkPos))
	plobjs->push_back( new FieldMarker( LookPos,0,6666,""));

      if (!Ground->showViewCone(true,PS->my.number)) 
	Ground->toggleViewConeLeft(PS->my.number);

      Me=new FieldPlayer(PS->my.pos,PS->my.bodyDir,PS->my.faceDir,
			 ( ( PS->viewWidth==VIEW_WIDTH_NORMAL ) ? M_PI/2 :
			   ( PS->viewWidth==VIEW_WIDTH_NARROW ) ? M_PI/4 :
			   ( PS->viewWidth==VIEW_WIDTH_WIDE ) ? M_PI: 0),
			   PS->my.number,true);
      plobjs->push_back( Me );

      QColor acol=FColors::agentColor(0);
      plobjs->push_back( new WFieldBall(PS->ball.pos,PS->ball.age,
					PS->ball.v,acol));
      
      int i;
      for (i=1; i<=11; i++)
	if ((i!=PS->my.number) && (PS->ourTeam[i].age<MAXAGE))
	  plobjs->push_back( new WFieldPlayer(PS->ourTeam[i].pos,
					      PS->ourTeam[i].bodyDir,
					      i,true,PS->ourTeam[i].age,
					      Vector2d(0,0),acol));
      acol=FColors::agentColor(1);
      for (i=1; i<=11; i++)
	if (PS->otherTeam[i].age<MAXAGE)
	  plobjs->push_back( new WFieldPlayer(PS->otherTeam[i].pos,
					      PS->otherTeam[i].bodyDir,
					      i,true,PS->otherTeam[i].age,
					      Vector2d(0,0),acol));
      
      Ground->redrawObjs(plobjs);

      delete PS;
    }
  // else RPDEBUG("no new snapshot");
  
};

void PlayerMWidget::closeEvent ( QCloseEvent *e)
{
  wIsOpen=false;
};
/*
void PlayerMWidget::pollSnapshots()
{
  Ground->pollSnapshots();
};
*/

///////////////////////////////////////////////////////////////
//
//   player part
//
/////////////////////////////////////////////////////////////

class marionette : public Player {
public:
  marionette(const char* team, bool goalie, const char *host, int port,
		int recn, const char * traceCfg, const char *configfile,
		bool FUNSSELmode, int clangMinVer, int clangMaxVer,
	     commun_struct_t * _data);
  ~marionette();

  virtual void act();
  
private:

  bool communicateTackle;   // for the use of emergency channel after tackling

  void sendAudioMsg();

  PassEval                *passeval;
  GoalShotEval            *goaleval;
  PositionEval            *poseval;
  TacticalModule          *tactical;
  commun_struct_t         *data;
};

marionette::marionette(const char* team, bool goalie, const char *host,
		       int port, int recn, const char *configfile,
		       const char * traceCfg, bool FUNSSELmode,
		       int clangMinVer, int clangMaxVer, 
		       commun_struct_t *_data) :
  Player(team, goalie, host, port, configfile, traceCfg, recn,
	 FUNSSELmode, clangMinVer, clangMaxVer)
{
  if (!skills)
    exit(-666);

  data=_data;
  
  tactical     = new TacticalModule(world, skills);
  /*  goaleval     = new GoalShotEval();
  passeval     = new PassEval(goaleval);
  
  poseval      = new PositionEval(goaleval, passeval);
  */

}

marionette::~marionette() {

  delete passeval;
  delete goaleval;
  delete poseval;
}

void PrintState( AdvSkills::KickState state)
{
  switch (state){
  case AdvSkills::KRANGE:
    printf("KRange\n");
    break;
  case AdvSkills::KFINISHING:
    printf("KFinishing\n");
    break;
  case AdvSkills::KTOOSHORT:
    printf("KTooShort\n");
    break;
  case AdvSkills::KPLACING:
    printf("KPlacing\n");
    break;
  case AdvSkills::KPLACEFORSKICK:
    printf("KPlaceForSKick\n");
    break;
  case AdvSkills::KSTOPPING:
    printf("KStopping\n");
    break;
  default:
    printf("unknown\n");
  };
};

void
marionette::act() {

  static AdvSkills::ExtKickCommand EK;

  SensorData audioData;   // audio message from teammate
  audioData = world->adjust();

  // if teammate has spoken, give message to tactical module
  if( audioData.type != SENSOR_TYPE_ERROR )
    tactical->decodeAudio( audioData );

  // handle rest of audio messages (e.g., from coach)
  AudioMsg audiomsg;
  while (getNextMsg(audiomsg)) {
    // give message to worldmodel
    if (world->feedbackAudio(audiomsg)) {
  }
  };

  int status;
  Vector2d            Pos;
  enum PlCommand      Command;   
  enum LookCommand    LkCommand;
  Vector2d            LookPos;

  status = pthread_mutex_lock (&data->mutex);
  if (status != 0) {fprintf(stderr,"Lock error!\n"); exit(-1);}
  
  if (data->Snap!=0)
    {
      delete data->Snap;
    };
  
  data->Snap=new Worldmodel(*world);
  Command=data->Command;
  if (Command==PlIntercept)
    data->LkCommand=LkNone;

  Pos=data->Pos;
  LookPos=data->LookPos;
  LkCommand=data->LkCommand;
    
  status = pthread_mutex_unlock (&data->mutex);
  if (status != 0) {fprintf(stderr,"Unlock error!\n"); exit(-1);}
  
  COUT("doing world feedback for tactical module...");
  tactical->feedbackWorld();

  AdvSkills::TurtleCommand tc; 
  
  AdvSkills::DribbleCommand dc;
  if (Command==PlDribble)
    {
      Angle a = (Pos - world->my.pos).getAngle();
      dc=skills->dribble(a, a + Angle(M_PI/4), 100);
      skills->exec(dc);
    }
  else
    {
      if (world->my.ballKickable(world->ball))
	{ 
	  if ((Command==PlIntercept) ||(Command==PlNone))
	    {
	      Angle a = (Pos - world->my.pos).getAngle();
	      dc=skills->dribble(a, a + Angle(M_PI/4), 100);
	      skills->exec(dc);
	    }
	  else 
	    {
	      RPDEBUG("kick\n");
	      EK=skills->kick(BALL_SPEED_MAX/4.0,Pos,2);
	      skills->exec(EK);
	      if ((EK.state==AdvSkills::KFINISHING)||
		  (EK.state==AdvSkills::KTOOSHORT))
	      	Command=PlNone;
	      //PrintState(EK.state);
	      //    Command=PlNone;
	    }
	}
      else //ball not kickable
	{
	  if ((world->playMode==PLAY_MODE_BEFORE_KICK_OFF) ||
	      (world->playMode==PLAY_MODE_GOAL_US) ||
	      (world->playMode==PLAY_MODE_GOAL_THEM))
	      
	    {
	      if ((Command==PlGo2) || (Command==PlfastGo2))
		{
		  skills->basicMove(Pos.getX(),Pos.getY());
		  Command=PlNone;
		}
	      else skills->noAction();
	    }
	  else //play mode "play on"
	    {
	      if (Command==PlGo2)
		{
		  tc= skills->goToPosition(Pos, 2, 50);
		}
	      else if (Command==PlfastGo2)
		{
		  tc= skills->goToPosition(Pos,2,100);
		}
	      else if (Command==PlIntercept)
		{
		  tc= skills->controlBall(world->my.stamina );
		}
	      
	      if (Command==PlNone)
		skills->noAction();
	      else
		{
		  if (tc.state != AdvSkills::TFINISHED)
		    skills->exec(tc);
		  else 
		    {
		      skills->noAction();
		      Command=PlNone;
		    };
		};
	    };
	};
    };

  // now turn head
  AdvSkills::TurnHeadCommand THC;

  if (LkCommand==LkNone)
    THC=skills->turnHead(world->ball);
  else if (LkCommand==LkPos)
    THC=skills->turnHead(LookPos);

  skills->exec(THC);


  status = pthread_mutex_lock (&data->mutex);
  if (status != 0) {fprintf(stderr,"Lock error!\n"); exit(-1);}

  data->Command=Command;
  
  status = pthread_mutex_unlock (&data->mutex);
  if (status != 0) {fprintf(stderr,"Unlock error!\n"); exit(-1);}

  // audio message generation
  sendAudioMsg();

  if (SYNCH_MODE)
    sendDone();

} //act


void marionette::sendAudioMsg()
{
  char msg[11];
  msg[10] = (char)0;

  // get player's action (to see if he tackled)
  CycleAction action;
  getLastAction( action );

  msg[9] = msgEncodeChars[char( 72 )];   // no passmodule
  MessageType msgType = tactical->determineMessageType( action, Grade::BAD );

  if( msgType != MSGTYPE_NONE )
    {
      world->encodeAudio( msgType, msg );   // communicate important worldmodel variables, msg[0]..msg[5]
      tactical->encodeAudio( msg );   // communicate important tactical module variables, msg[6]..msg[8]

      say_10( msg );   // soccerserver 8+ communication

      COUT( "Sending AudioMsg: " << msg[0] << msg[1] << msg[2] << msg[3]
	    << msg[4] << msg[5] << msg[6] << msg[7] << msg[8] << msg[9] );
    }
}

void * PlayerThread(void * Marion)
{
  marionette *a=(marionette*) Marion;
  a->play();

  int status;
  status = pthread_mutex_lock (&data.mutex);
  if (status != 0) {fprintf(stderr,"Lock error!\n"); exit(-1);}
  
  data.finished=true;

  status = pthread_mutex_unlock (&data.mutex);
  if (status != 0) {fprintf(stderr,"Unlock error!\n"); exit(-1);}

  return NULL;
};


void startPlayer(int argc, char **argv)
{
  char   *team_name= "MRB";
  char   *host= new char [strlen(DEFAULT_HOST)+1];
  strcpy (host, DEFAULT_HOST);
  int    port = PORT;
  int    reconnectNumber = 0;
  bool   isGoalie= false;
  char   *configfile= "agent.conf";
  char   *traceCfg= "trace.conf";
  bool   FUNSSELmode = false;
  int    clangMinVer = 7;
  int    clangMaxVer = 7;

  char   c;
  while((c = getopt( argc, argv, "Fgh:l:r:p:t:f:c:C:" ) ) != EOF ) {
    switch(c) {
    case 'g':
      isGoalie = true;
      break;
    case 'h':
      delete (host);
      host = optarg;
      break;
    case 'l':
      traceCfg = optarg;
      break;
    case 'r':
      reconnectNumber = atoi(optarg);
      if ((reconnectNumber < 1) || (reconnectNumber > 11))
	goto arg_err;
      break;
    case 'p':
      port = atoi(optarg);
      break;
    case 't':
      team_name = optarg;
      break;
    case 'f':
      configfile= optarg;
      break;
    case 'F':
      FUNSSELmode = true;
      break;
    case 'c':
       clangMinVer = atoi(optarg);
      break;
    case 'C':
       clangMaxVer = atoi(optarg);
      break;
    case '?':
    arg_err:		
    cerr << "Usage: " << argv[0] << "[-h HOSTNAME]"
      " [-l FILENAME] [-r num] [-p PORT] [-t TEAMNAME]" 
      "[-f FILENAME]" << endl
	 << "  -g           agent is goalie" << endl
	 << "  -h HOSTNAME  connects to HOSTNAME (default "<< DEFAULT_HOST << ")" << endl
	 << "  -l FILENAME  alternative configuration file for debug traces (default trace.conf)" << endl
	 << "  -r num       reconnect with that number" << endl
	 << "  -p PORT      port number (default 6000)" << endl
	 << "  -t TEAMNAME  sets the player's team name (default MRB)" << endl
	 << "  -f FILENAME  alternative configuration file for tactics (default agent.conf)" << endl
         << "  -F           enable FUNSSEL mode" << endl
         << "  -c num       minimal supported coachlanguage version (default 7)"<< endl
         << "  -C num       maximal supported coachlanguage version (default 7)"<< endl
         << "  -?           display this help screen"<< endl;
    exit(1);
    }
  }

  if (clangMinVer < 7 || clangMaxVer < 7 ||
      clangMinVer > 8 || clangMaxVer > 8 ) {
    cerr << "unsupported coachlanguage version\n supported versions: 7\n";
    exit(-1);
  }

  if (clangMinVer > clangMaxVer) {
    cerr << "argument to -c has to be less than argument to -C" << endl << endl;
    exit(-1);
  }


  cerr << "Mainz Rolling Brains RoboCup RealPlayer" << endl
       << "copyright 1998, 1999, 2000, 2001, 2002, 2003 by the Mainz Rolling Brains team" 
       << endl;


  
  marionette *a = new marionette(team_name, isGoalie, host, port, 
				 reconnectNumber, configfile, traceCfg,
				 FUNSSELmode, clangMinVer, clangMaxVer,
				 &data);
 
  int status = 
    pthread_create (&Player, NULL,  PlayerThread, 
		    (void *) a);     
  if (status != 0) {fprintf(stderr,"Create error!\n"); exit(-1);}


};



///////////////////////////////////////////////////////////////////////
//
//    main
//
//////////////////////////////////////////////////////
#include <unistd.h>
#include <sys/resource.h>
#include <errno.h>
#include "RealPlayer.moc"


int main (int argc, char **argv)
{
  struct rlimit rlim;
  getrlimit(RLIMIT_CORE, &rlim);
  rlim.rlim_cur = rlim.rlim_max;
  cerr << "core dump enabled." << endl;
  if (setrlimit(RLIMIT_CORE, &rlim) == -1)
    cerr << errno << endl;

  QApplication a(argc, argv);

  Config::readFile("funsselmonrc");
  try {
    FColors::init();
  } catch (Exception e) {
    cerr << e << endl;
    exit(666);
  }
  
  startPlayer(argc,argv);


  PlayerMWidget *MWidget=new PlayerMWidget();
  a.setMainWidget( MWidget );
  MWidget->show();
  
  while(MWidget->isOpen()) {
    // give time to qt
    a.processEvents(50);
    // give time to parser
    MWidget->pollSnapshots();
    // wait a moment
    usleep(20);
  }
  
  delete MWidget;
  
  exit(0);
};
