//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <unistd.h>
#include <iostream>
#include <strings.h>
#include "constants.h"
#include "mastercontrol.h"

using namespace std;


int main (int argc, char** argv) {
  char   *team_name= "MRB";
  char   *host= new char [strlen(DEFAULT_HOST)+1];
  strcpy (host, DEFAULT_HOST);
  int    port = PORT;
  int    reconnectNumber = 0;
  bool   isGoalie= false;
  char   *configfile= "agent.conf";
  char   *traceCfg= "trace.conf";
  bool   FUNSSELmode = false;
  int    clangMinVer = 7;
  int    clangMaxVer = 7;

  char   c;
  while((c = getopt( argc, argv, "Fgh:l:r:p:t:f:c:C:" ) ) != EOF ) {
    switch(c) {
    case 'g':
      isGoalie = true;
      break;
    case 'h':
      delete (host);
      host = optarg;
      break;
    case 'l':
      traceCfg = optarg;
      break;
    case 'r':
      reconnectNumber = atoi(optarg);
      if ((reconnectNumber < 1) || (reconnectNumber > 11))
	goto arg_err;
      break;
    case 'p':
      port = atoi(optarg);
      break;
    case 't':
      team_name = optarg;
      break;
    case 'f':
      configfile= optarg;
      break;
    case 'F':
      FUNSSELmode = true;
      break;
    case 'c':
       clangMinVer = atoi(optarg);
      break;
    case 'C':
       clangMaxVer = atoi(optarg);
      break;
    case '?':
    arg_err:		
    cerr << "Usage: " << argv[0] << "[-h HOSTNAME]"
      " [-l FILENAME] [-r num] [-p PORT] [-t TEAMNAME]" 
      "[-f FILENAME]" << endl
	 << "  -g           agent is goalie" << endl
	 << "  -h HOSTNAME  connects to HOSTNAME (default "<< DEFAULT_HOST << ")" << endl
	 << "  -l FILENAME  alternative configuration file for debug traces (default trace.conf)" << endl
	 << "  -r num       reconnect with that number" << endl
	 << "  -p PORT      port number (default 6000)" << endl
	 << "  -t TEAMNAME  sets the player's team name (default MRB)" << endl
	 << "  -f FILENAME  alternative configuration file for tactics (default agent.conf)" << endl
         << "  -F           enable FUNSSEL mode" << endl
         << "  -c num       minimal supported coachlanguage version (default 7)"<< endl
         << "  -C num       maximal supported coachlanguage version (default 7)"<< endl
         << "  -?           display this help screen"<< endl;
    exit(1);
    }
  }

  if (clangMinVer < 7 || clangMaxVer < 7 ||
      clangMinVer > 8 || clangMaxVer > 8 ) {
    cerr << "unsupported coachlanguage version\n supported versions: 7\n";
    exit(-1);
  }

  if (clangMinVer > clangMaxVer) {
    cerr << "argument to -c has to be less than argument to -C" << endl << endl;
    exit(-1);
  }

  cerr << "Mainz Rolling Brains RoboCup player" << endl
       << "copyright 1998, 1999, 2000, 2001, 2002, 2003 by the Mainz Rolling Brains team" 
       << endl;

  MasterControl *a = new MasterControl(team_name, isGoalie, host, port, 
				       reconnectNumber, configfile, traceCfg,
				       FUNSSELmode, clangMinVer, clangMaxVer);
  a->play();

  delete (a);
}
