//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#define MOD_MC

#include "Trace.h"

#include "mastercontrol.h"

#include "drawMacros.h"
#include "Config.h"
#include "mcommand.h"
#include "AdvSkills.h"

#include <sstream>
#include <string>

using namespace std;


#undef USETIMETRACE

// trace macro: uncomment the line you want.

//#define USETIMETRACE
//#define SHOWOFFSIDELINE

#ifdef SHOWOFFSIDELINE
#warning: MC debug trace SHOWOFFSIDELINE is activated!
#endif

#ifndef USETIMETRACE
#define TIMETRACE(marker)
#define INITTTRACE
#define TOTALTTRACE
#else
#warning: MC debug trace TIMETRACE is activated!
#include <sys/resource.h>
#include <time.h>

static rusage start, total_start, end;

#define INITTTRACE \
getrusage(RUSAGE_SELF, &start); \
getrusage(RUSAGE_SELF, &total_start); \
cout << getTime() << " TT: starting" << endl; 

#define TIMETRACE(marker) \
getrusage(RUSAGE_SELF, &end); \
cout << getTime() << " TT: " marker " " << \
((1.0e-3*end.ru_utime.tv_usec   + 1.0e3*end.ru_utime.tv_sec) - \
 (1.0e-3*start.ru_utime.tv_usec + 1.0e3*start.ru_utime.tv_sec)) << endl; \
start = end;

#define TOTALTTRACE \
getrusage(RUSAGE_SELF, &end); \
cout << getTime() << " TT: total time______________ " << \
((1.0e-3*end.ru_utime.tv_usec         + 1.0e3*end.ru_utime.tv_sec) - \
 (1.0e-3*total_start.ru_utime.tv_usec + 1.0e3*total_start.ru_utime.tv_sec)) << endl;
#endif


MasterControl::MasterControl(const char* team, bool goalie, const char *host,
			     int port, int recn, const char *configfile,
			     const char * traceCfg, bool FUNSSELmode,
			     int clangMinVer, int clangMaxVer) :
  Player(team, goalie, host, port, configfile, traceCfg, recn,
	 FUNSSELmode, clangMinVer, clangMaxVer)
{
  if (!skills)
    exit(-666);

  tactical     = new TacticalModule(world, skills);

  goaleval     = new GoalShotEval();
  passeval     = new PassEval(goaleval);
  poseval      = new PositionEval(goaleval, passeval);

  standard     = new StandardSituationModule(world, tactical, skills, goalie, passeval, goaleval, poseval);
  pass         = new PassModule(world, tactical, skills, passeval, goaleval);
  ballhandling = new BallHandlingModule(world, tactical, skills);
  goalshot     = new GoalShotModule(world, tactical, skills, goaleval);
  position     = new PositionModule(goalie, world, tactical, skills, passeval, goaleval, poseval);
}


MasterControl::~MasterControl()
{
  delete tactical;
  delete standard;
  delete pass;
  delete passeval;
  delete ballhandling;
  delete goalshot;
  delete goaleval;
  delete position;
  delete poseval;
}

void
MasterControl::act()
{
  INITTTRACE;

  //[!] HACK: sollte nur aufgerufen werden, wenn der Player Type gendert wurde
  skills->readjustToPlayer(world->my);

  skills->blockTurnHead();

  SensorData audioData;          // audio message from teammate
  audioData = world->adjust();   // adjust worldmodel with new sensor information

  // if teammate has spoken, give message to tactical module
  if( audioData.type != SENSOR_TYPE_ERROR )
    tactical->feedbackAudio( audioData );

  // handle rest of audio messages (e.g., from coach)
  AudioMsg audiomsg;
  while( getNextMsg( audiomsg ) )
    {
      // give message to worldmodel
      if( world->feedbackAudio( audiomsg ) )
	{
	  VALUE( "Got coachmsg: " << audiomsg );

	  if( !standard->feedbackAudio( audiomsg ) )
	    cerr << "STD: COULD NOT READ COACHMSG" << endl;
	  else
	  if( !position->feedbackAudio( audiomsg ) )
	    cerr << "position: COULD NOT READ COACHMSG" << endl;
	}
    }

  // traces to check offside line
#ifdef SHOWOFFSIDELINE
  {
    stringstream ost;
    ost << "MC offside-line: ourOffsideLine: " 
	<< world->ourOffsideLine(MAXAGE) 
	<< " otherOffsideLine: "
      	<< world->otherOffsideLine(MAXAGE) 
	<< (char)0;
    string str = ost.str(); 
    Comm::sendMCmd(MLog(str));
    // leak delete str;
    Comm::sendMCmd(MDVLine(world->ourOffsideLine(MAXAGE)));
  } 
#endif

  tactical->feedbackWorld();
  standard->feedbackWorld();
  pass->feedbackWorld();
  ballhandling->feedbackWorld();
  goalshot->feedbackWorld();
  position->feedbackWorld();

  TIMETRACE("after modules feed back ");

  // get all action evaluations from modules
  Grade stndG = standard->evaluate();
  TIMETRACE("after standard eval     ");

  Grade passG = pass->evaluate();
  TIMETRACE("after pass     eval     ");

  Grade handG = ballhandling->evaluate();
  TIMETRACE("after ballh    eval     ");

  Grade goalG = goalshot->evaluate();
  TIMETRACE("after goalshot eval     ");

  Grade posnG = position->evaluate();
  TIMETRACE("after position eval     ");

  // HACK
  if( world->playMode == PLAY_MODE_PENALTY_READY_US ||
      world->playMode == PLAY_MODE_PENALTY_TAKEN_US )
    {
      passG.set_IMPOSSIBLE();
      posnG.set_IMPOSSIBLE();
    }

  enum { DEC_STAND, DEC_PASS, DEC_HAND, DEC_GOAL, DEC_POSN } decision = DEC_POSN;

  // simple bugfix: avoid undefined grades
  
  if (!stndG.is_defined()) {
    stndG.set_IMPOSSIBLE();
    cerr << "MC: " << getTime() << ", Player No: " 
	 << world->my.number << " stndG undefined! Changed to impossible"
	 << endl;
  }

  if (!goalG.is_defined()) {
    goalG.set_IMPOSSIBLE();
    cerr << "MC: " << getTime() << ", Player No: " 
	 << world->my.number << " goalG undefined! Changed to impossible"
	 << endl;
  }

  if (!handG.is_defined()) {
    handG.set_IMPOSSIBLE();
    cerr << "MC: " << getTime() << ", Player No: " 
	 << world->my.number << " handG undefined! Changed to impossible"
	 << endl;
  }

  if (!passG.is_defined()) {
    passG.set_IMPOSSIBLE();
    cerr << "MC: " << getTime() << ", Player No: " 
	 << world->my.number << " passG undefined! Changed to impossible"
	 << endl;
  }

  if (!posnG.is_defined()) {
    posnG.set_IMPOSSIBLE();
    cerr << "MC: " << getTime() << ", Player No: " 
	 << world->my.number << " posnG undefined! Changed to impossible"
	 << endl;
  }

  // handle absolute values
  
  if (stndG.is_ABSOLUTE()) 
    decision= DEC_STAND;
  else if (goalG.is_ABSOLUTE()) 
    decision= DEC_GOAL;
  else if (handG.is_ABSOLUTE())
    decision= DEC_HAND;
  else if (passG.is_ABSOLUTE())
    decision= DEC_PASS;
  else if (posnG.is_ABSOLUTE())
    decision= DEC_POSN;

  // Magic decision tree

  else if (!goalG.is_worse_than( handG ) &&
	   !goalG.is_worse_than( passG ) &&
	   !goalG.is_worse_than( posnG ) ) 
    decision = DEC_GOAL;
  else if (!passG.is_worse_than( handG ) && 
	   !passG.is_worse_than( posnG ) ) 
    decision = DEC_PASS;
  else if (!handG.is_worse_than( posnG ) &&
	   !handG.is_worse_than( passG ) ) 
    decision = DEC_HAND;
  else if (!passG.is_worse_than( posnG )) 
    decision = DEC_PASS;
  else 
    decision = DEC_POSN;
  
  // do actions

  VALUE( "Grades: std="<< stndG << ", goal=" << goalG << ", pass="
	 << passG << ", bhand=" << handG << ", pos=" << posnG );

  switch (decision) {
  case DEC_STAND:
    DECISION( "Standardsitmodule acts" );
    VALUE( "with grade " << stndG );
    standard->act();
    break;
  case DEC_PASS:
    DECISION( "Passmodule acts" );
    VALUE( "with grade " << passG );
    pass->act();
    break;
  case DEC_HAND:
    DECISION( "Ballhandlingmodule acts" );
    VALUE( "with grade " << handG );
    ballhandling->act();
    break;
  case DEC_GOAL:
    DECISION( "Goalshotmodule acts" );
    VALUE( "with grade " << goalG );
    goalshot->act();
    break;
    // default action: positioning ...
  case DEC_POSN:
  default:
    DECISION( "Positioningmodule acts" );
    VALUE( "with grade " << posnG );
    position->act();
  } // switch

  TIMETRACE("after module act        ");

  // now collect all TurnHeadWishes...
  TurnHeadWish stdTHW, passTHW, ballhTHW, goalTHW, posTHW;

  stdTHW = standard->turnHeadEvaluate();
  TIMETRACE("after standard THeval   ");

  passTHW = pass->turnHeadEvaluate();
  TIMETRACE("after pass     THeval   ");

  ballhTHW = ballhandling->turnHeadEvaluate();
  TIMETRACE("after ballh    THeval   ");

  goalTHW = goalshot->turnHeadEvaluate();
  TIMETRACE("after goalshot THeval   ");
  posTHW = position->turnHeadEvaluate();
  TIMETRACE("after position THeval   ");

  // execute turn head command
  skills->exec( tactical->calcTurnHead( goalTHW, passTHW, ballhTHW, posTHW, stdTHW ) );
  TIMETRACE("after calcTurnHead      ");

  // audio message generation
  sendAudioMsg();

  TOTALTTRACE;
} //act


void MasterControl::sendAudioMsg()
{
  char msg[11];
  msg[10] = (char)0;

  // get player's action (to see if he tackled)
  CycleAction action;
  getLastAction( action );

  Grade passGrade = pass->encodeAudio( msg );   // communicate player's self-evaluation as passpartner, msg[9]
  
  WBall simBall;
  MessageType msgType = tactical->determineMessageType( action, passGrade, simBall );

  if( msgType != MSGTYPE_NONE )
    {
      // communicate important worldmodel variables, msg[0]..msg[5]
      world->encodeAudio( msgType, msg, simBall );   
      // communicate important tactical module variables, msg[6]..msg[8]
      tactical->encodeAudio( msg );

      say_10( msg );   // soccerserver 8+ communication

      VALUE( "Sending AudioMsg: " << msg[0] << msg[1] << msg[2] << msg[3]
	     << msg[4] << msg[5] << msg[6] << msg[7] << msg[8] << msg[9] );
    }
}
