#include <iostream>
#include <iomanip>
#include <fstream>
#include <strstream>
#include <netinet/in.h>
#include <string>
#include <SServer.h>
#include <sys/stat.h>
#include <errno.h>

#include <MsgStore.h>

#include "goalKickStats.h"

// cwz-build-command: gmake TARGET=goalKickStats

const int grid = 20;
const int grid1d = 200;
const double grid1dstretch = 0.25;

const double kick_dist = PLAYER_SIZE + BALL_SIZE + KICKABLE_MARGIN;

// team identification

#if 0
// teams from aarn-soccerlog
const int n_teams = 10;
const char *teams[n_teams] = {
  "11MonkeysZ00",
  "ATH2000neu",
  "BS2001",
  "Cyberoos01",
  "FCPortugal00",
  "MRB01",
  "Robolog01",
  "TutGroove01",
  "VirtualWerder01",
  "YowAI00" 
};
#else
// teams from GO2
const int n_teams = 15;
const char *teams[n_teams] = {
  "MRB",
  "DrWeb",
  "Matrix",
  "Arvand",
  "UTUtd2002",
  "BS02",
  "BalticLuebeck",
  "UvA_Trilearn",
  "ATH_2001_PB_Edt",
  "BUGS",
  "RoboLog",
  "ERA-Polytech2002",
  "Persepolis",
  "AllemaniACs",
  "Brainstormers02"
};
#endif


bool attached;
string team[2];
int    team_id[2] = { 0, 0 };

// goal kick statistics
int kicked_goal[n_teams][grid][grid] = { { { 0 } } };
int self_goal[n_teams][grid][grid] = { { { 0 } } };
int got_goal[n_teams][grid][grid] = { { { 0 } } };
int passed[n_teams][grid][grid] = { { { 0 } } };
int good_pass[n_teams][grid1d] = { { 0 } };
int bad_pass[n_teams][grid1d] = { { 0 } };
int self_pass[n_teams][grid1d] = { { 0 } };
int good_pass_t[n_teams][grid1d] = { { 0 } };
int bad_pass_t[n_teams][grid1d] = { { 0 } };
int self_pass_t[n_teams][grid1d] = { { 0 } };
int intercepted[n_teams][grid][grid] = { { { 0 } } };
int dribbled[n_teams][grid][grid] = { { { 0 } } };

int score[n_teams][n_teams]  = { { 0 } };
int events2[n_teams][n_teams] = { { 0 } };
int events[n_teams] = { { 0 } };

// who kicked last
int cur_player;
int cur_team;
int cur_time;
// and where
Vector2d cur_pos;

void tryAttachTeams(const SoccerMessage &msg)
{
  SServer::showinfo_t2 *show2 = getShowInfo(msg);
  if (!show2)
    return;

  team_id[0] = team_id[1] = -1;

  if (show2->team[0].name[0]) {
    team[0] = show2->team[0].name;
    for (int t = 0; t < n_teams; t++) {
      if (team[0] == teams[t] ||
	  team[0] == string(teams[t]) + "-2")
	team_id[0] = t;
    }
    if (team_id[0] == -1)
	cerr << "could not attach team " << team[0] << endl;
  }
  if (show2->team[1].name[0]) {
    team[1] = show2->team[1].name;
    for (int t = 0; t < n_teams; t++) {
      if (team[1] == teams[t] ||
	  team[1] == string(teams[t]) + "-2")
	team_id[1] = t;
    }
    if (team_id[1] == -1)
	cerr << "could not attach team " << team[1] << endl;
  }
}

int shotGoal(SServer::showinfo_t2 *show2)
{
  if (show2->pmode == SServer::PM_AfterGoal_Left)
    return 0;
  else if (show2->pmode == SServer::PM_AfterGoal_Right)
    return 1;
  return -1;
}

void print_gridded_data(const char *name, int data[grid][grid], int norm)
{
  ofstream out(name);
  
  for (int gx = 0; gx < grid; gx++) {
    double x = PITCH_LENGTH*(double(gx + 0.5)/grid - 0.5);
    for (int gy = 0; gy < grid; gy++) {
      double y = PITCH_WIDTH*(double(gy + 0.5)/grid - 0.5);
      if (norm > 0)
	out << x << " " << y << " " << double(data[gx][gy])/norm
	    << endl;
      else
	out << x << " " << y << " *" << endl;
    }
    out << endl;
  }
}

void print_gridded_1ddata(const char *name, int data[grid], int data2[grid], int norm)
{
  ofstream out(name);
  
  for (int gy = 0; gy < grid1d; gy++) {
    if (norm > 0)
      out << gy << " " << gy*grid1dstretch << " " << double(data[gy])/norm << " " << double(data2[gy])/norm
	  << endl;
    else
      out << gy << " *" << endl;
  }
}

int main(int argc, const char **argv)
{
  ofstream log("goalKickStats.log");

  try {
    for (int arg = 1; arg < argc; arg++) {
      log << "analyze " << argv[arg] << endl;

      cerr << 100*arg/argc << "% done" << endl;
      MsgStore store(argv[arg]);
      attached = false;

      cur_player = -1;
      for (int time = 0; time <= 6000; time++) {
	if (!store.fetchTime(time))
	  break;

	if ((time % 100) == 0)
	  cerr << "time " << time << "\r";

	SoccerMessage msg;
	while ((msg = store.getNextMessage()).isValid()) {
	  if (!attached) {
	    tryAttachTeams(msg);
	    if (team_id[0] != -1 && team_id[1] != -1)
	      attached = 1;
	  }
	  
	  if (!attached)
	    continue;
	  
	  SServer::showinfo_t2 *show2 = getShowInfo(msg);
	  if (!show2)
	    continue;
 
	  // get score after game (no extension)
	  if (time == 6000) {
	    events[team_id[0]]++;
	    events[team_id[1]]++;
	    events2[team_id[0]][team_id[1]]++;
	    events2[team_id[1]][team_id[0]]++;
	  
	    score[team_id[0]][team_id[1]] += ntohs(show2->team[0].score);
	    score[team_id[1]][team_id[0]] += ntohs(show2->team[1].score);
	    // finished
	    break;
	  }
	  
	  Vector2d bpos = getBallPos(show2);

	  // using Noda's strange constants for field size
	  // since we cannot include constants.h
	  int cx = int(floor((0.5 + cur_pos.getX()/PITCH_LENGTH)*grid));
	  int cy = int(floor((0.5 + cur_pos.getY()/PITCH_WIDTH)*grid));
	  if (cx < 0)
	    cx = 0;
	  if (cx >= grid)
	    cx = grid - 1;
	  // turn around field for second team
	  if (cur_team == 1)
	    cx = grid - 1 - cx;

	  if (cy < 0)
	    cy = 0;
	  if (cy >= grid)
	    cy = grid - 1;

	  // check for goal
	  bool exit = false;
	  switch (shotGoal(show2)) {
	  case 0:
	    if (cur_player == -1)
	      continue;
	    if (cur_team == 0) {
	      kicked_goal[team_id[cur_team]][cx][cy]++;
	      got_goal[team_id[cur_team ^ 1]][grid - 1 - cx][cy]++;
	      log << time << ": goal left "
		  << team_id[cur_team] << " " << team_id[cur_team ^ 1] << endl;
	      exit = true; 
	    }
	    else {
	      self_goal[team_id[cur_team]][cx][cy]++;
	      log << time << ": goal self left"
		  << team_id[cur_team] << endl;
	      exit = true; 
	    }
	    break;
	  case 1:
	    if (cur_player == -1)
	      continue;
	    if (cur_team == 1) {
	      kicked_goal[team_id[cur_team]][cx][cy]++;
	      got_goal[team_id[cur_team ^ 1]][grid - 1 - cx][cy]++;
	      log << time << ": goal right "
		  << team_id[cur_team ^ 1] << " " << team_id[cur_team] << endl;
	      exit = true; 
	    }
	    else {
	      log << time << ": goal self right "
		  << team_id[cur_team ^ 1] << endl;
	      self_goal[team_id[cur_team]][cx][cy]++;
	      exit = true; 
	    }
	    break;
	  default:
	    break;
	  }

	  if (exit)
	    break;

	  // search for kicker
	  int new_player = -1;
	  Vector2d new_pos;
	  int new_team = -1;
	  for (int i = 0; i < 22; i++) {
	    Vector2d ppos = getPlayerPos(show2, i);
	    if (playerKicked(show2, i) &&
		sqrnorm(ppos - bpos) < sqr(kick_dist)) {
	      new_player = i;
	      new_pos = ppos;
	      new_team = (i >= 11) ? 1 : 0;
	    }
	  }

	  // no one kicked
	  if (new_player == -1)
	    break;

	  int pass_t = time - cur_time;
	  int pass_d = int(norm(new_pos - cur_pos)/grid1dstretch);
	  if (pass_t >= grid1d)
	    pass_t = grid1d - 1;

	  if (pass_d >= grid1d)
	    pass_d = grid1d - 1;
	  
	  if (cur_player != -1) {
	    if (cur_player == new_player) {
	      self_pass_t[team_id[cur_team]][pass_t]++;
	      self_pass[team_id[cur_team]][pass_d]++;

	      if (time > cur_time + 4) {
		dribbled[team_id[cur_team]][cx][cy]++;
		log << time << ": dribbled "
		    << team_id[cur_team] << " " << cur_player << endl;
	      }
	    }
	    else if (new_team == cur_team) {
	      good_pass_t[team_id[cur_team]][pass_t]++;
	      good_pass[team_id[cur_team]][pass_d]++;
	      passed[team_id[cur_team]][cx][cy]++;
	      log << time << ": passed "
		  << team_id[cur_team] << " " << cur_player << " " << new_player << endl;
	    }
	    else {
	      bad_pass_t[team_id[cur_team]][pass_t]++;
	      bad_pass[team_id[cur_team]][pass_d]++;

	      int nx = int(floor((0.5 + new_pos.getX()/PITCH_LENGTH)*grid));
	      int ny = int(floor((0.5 + new_pos.getY()/PITCH_WIDTH)*grid));   
	      intercepted[team_id[new_team]][nx][ny]++;
	      log << time << ": icpted " << team_id[new_team] << " "
		  << cur_player << " "<< team_id[new_team ^ 1] << " "
		  << new_player << endl;
	    }
	  }
	  cur_pos = new_pos;
	  cur_time = time;
	  cur_team = new_team;
	  cur_player = new_player;
	  break;
	}
      }
      cerr << "done" << endl;
    }
    
    for (int team = 0; team < n_teams; team++) {
      strstream name;
      name << "kicked_goals_" << teams[team] << (char) 0;
      print_gridded_data(name.str(), kicked_goal[team],
			 events[team]);
      name.seekp(0,ios::beg);
      name << "self_goals_" << teams[team] << (char) 0;
      print_gridded_data(name.str(), self_goal[team],
			 events[team]);
      name.seekp(0,ios::beg);
      name << "got_goals_" << teams[team] << (char) 0;
      print_gridded_data(name.str(), got_goal[team],
			 events[team]);
      name.seekp(0,ios::beg);
      name << "passed_" << teams[team] << (char) 0;
      print_gridded_data(name.str(), passed[team],
			 events[team]);
      name.seekp(0,ios::beg);
      name << "intercepted_" << teams[team] << (char) 0;
      print_gridded_data(name.str(), intercepted[team],
			 events[team]);
      name.seekp(0,ios::beg);
      name << "dribbled_" << teams[team] << (char) 0;
      print_gridded_data(name.str(), dribbled[team],
			 events[team]);
      name.seekp(0,ios::beg);
      name << "good_pass_" << teams[team] << (char) 0;
      print_gridded_1ddata(name.str(), good_pass[team], good_pass_t[team],
			   events[team]);
      name.seekp(0,ios::beg);
      name << "bad_pass_" << teams[team] << (char) 0;
      print_gridded_1ddata(name.str(), bad_pass[team], bad_pass_t[team],
			   events[team]);
      name.seekp(0,ios::beg);
      name << "self_pass_" << teams[team] << (char) 0;
      print_gridded_1ddata(name.str(), self_pass[team], self_pass_t[team],
			   events[team]);
    }
    // print scores
    ofstream out("scores");
    out.precision(2);
    
    out << setw(16) << "team";
    for (int a = 0; a < n_teams; a++)
      out << " " << teams[a];
    out << endl;
    
    for (int a = 0; a < n_teams; a++) {
      out << setw(16) << teams[a];
      for (int b = 0; b < n_teams; b++) {
	if (events2[a][b])
	  out << setw(strlen(teams[b]) + 1) << double(score[a][b])/events2[a][b];
	else
	  out << setw(strlen(teams[b]) + 1) << "*";
      }
      out << endl;
    }

    out << endl << "events measured" << endl << setw(16) << "team";
    for (int a = 0; a < n_teams; a++)
      out << " " << teams[a];
    out << endl;
    
    for (int a = 0; a < n_teams; a++) {
      out << setw(16) << teams[a];
      for (int b = 0; b < n_teams; b++)
	out << setw(strlen(teams[b]) + 1) << events2[a][b];
      out << endl;
    }
  }
  catch (Exception e) {
    cerr << "caught exception "<< e << endl;
    return -1;
  }
  return 0;
}
