//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// 	$Id: AdvSkills.C,v 1.12 2003/06/26 12:03:33 aarn Exp $
//      $Manager: aarn (Axel Arnold)$
//
//////////////////////////////////////////////////////////////////////

// for axels wonderful emacs-extension
// cwz-build-command: gmake libMRB.a

#include "constants.h"
#include "worldmodel.h"
#include "AdvSkills.h"
#include "mytrace.h"
#include "minmax.h"
#include <sstream>

using namespace std;

#define MOD_SKILLS
#include "Trace.h"
#define SKILLS_LEVEL 250
#define NAMEDLINE(name, start, dir) { \
    Vector2d end(start + dir); \
    LEVELDRW(SKILLS_LEVEL, MDLine(start, end)); \
    LEVELDRW(SKILLS_LEVEL, MDLabel(end, name)); }

/////////////////////////////////////////////////// ENTIRE SKILLS STUFF

AdvSkills *skills = 0;

/*********************************** The constructor ************************/

AdvSkills::AdvSkills(Comm *c, Worldmodel *w):
  // pseudo constants
  precision(1e-5),
  minSpeed(0.1),
  minInterceptTime(30),
  goToThreshStretch(0.008),
  turnMinAngle(DEG_TO_RAD(15)),
  directThreshold(0.1),
  laybackErr(0.25),
  data_error(0.1),
  safeCircle(PLAYER_SIZE+BALL_SIZE+0.5*KICKABLE_MARGIN),
  stopableV(0.5*KICK_POWER_RATE*MAXPOWER),
  dribbleBodyErr(DEG_TO_RAD(10)),
  dribbleHoldErr(DEG_TO_RAD(30)),
  dribbleMinDash(PLAYER_RAND),
  dribbleMinBallV(BALL_RAND),
  controlMaxSteps(20),
  // real vars
  comm(c),
  world(w),
  blockTH(true),
  expmode(false)
{
  if (skills) {
    cerr << "second AdvSkills object created, exiting" << endl;
    exit(-1);
  }
  skills = this;
}


void AdvSkills::readjustToPlayer(const WPlayer& my)
{
  safeCircle = my.size + BALL_SIZE + 0.5 * my.kickableMargin;
}


/****************************** the advanced skills ************************/
#include "AdvSkillsKick.C"
#include "AdvSkillsIntercept.C"
#include "AdvSkillsDribble.C"

/********************************* the basic skills ************************/

void AdvSkills::basicDash (double d) {
  turnParam = 0;
  turnHeadParam = 0;
  blockTH = false;
  comm->sendDash(d);
}

void AdvSkills::basicKick (double a,double b) {
  turnParam = 0;
  turnHeadParam = 0;
  blockTH = false;
  comm->sendKick(a,b);
}

void AdvSkills::basicTurn (double d) {
  turnParam = d;
  turnHeadParam = 0;
  blockTH = false;
  comm->sendTurn(d);
}

void AdvSkills::basicCatch (double d) {
  turnParam = 0;
  turnHeadParam = 0;
  blockTH = false;
  comm->sendCatch(d);
}

void AdvSkills::noAction(void) {
  turnParam = 0;
  turnHeadParam = 0;
  blockTH = false;
  comm->noAction();
}

void AdvSkills::attentionTo(int pNum, bool ourTeam)
{
  comm->sendAttentionTo(pNum, ourTeam);
}

void AdvSkills::attentionToOff(void)
{
  comm->sendAttentionToOff();
}

void AdvSkills::basicPointTo(double dist, double dir) {
  turnParam = 0;
  turnHeadParam = 0;
  blockTH = false;
  comm->sendPointTo(dist, dir);
}

void AdvSkills::basicPointToOff(void) {
  turnParam = 0;
  turnHeadParam = 0;
  blockTH = false;
  comm->sendPointToOff();
}

void AdvSkills::basicTackle(double power) {
  turnParam = 0;
  turnHeadParam = 0;
  blockTH = false;
  comm->sendTackle(power);
}

void AdvSkills::basicMove(double x, double y) {
  turnParam = 0;
  turnHeadParam = 0;
  blockTH = false;
  comm->sendMove(x,y);
}

void AdvSkills::sendTurnNeck(double dir)
{
  turnHeadParam = dir;
  comm->sendTurnNeck(dir);
}

void AdvSkills::sendChangeView(view_width width, view_quality quality)
{
  comm->sendChangeView(width, quality);
}

void AdvSkills::sendSenseBody()
{
  comm->sendSenseBody();
}

/*********************************** catch *******************************/

AdvSkills::CatchCommand AdvSkills::goalieCatch( Angle dir )
{
  CatchCommand com;
  com.catchDir = dir;

  // check if player is goalie, if catching is not banned and if ball is in catchable area
  if(!comm->amIGoalie()) {
    com.state = CNOTGOALIE;
    return com;
  }
  if (world->cyclesSinceCatch <= CATCH_BAN_CYCLES) {
    com.state = CBAN;   
    return com;
  }
  if (!world->my.ballCatchable( world->ball, com.catchDir - world->my.bodyDir )) {
    com.state = CRANGE;
    return com;
  }
  com.state = COK;
  return com;
}

void AdvSkills::exec(const AdvSkills::CatchCommand &com) {
  basicCatch(com.catchDir - world->my.bodyDir);
}

/*********************************** turn *******************************/
AdvSkills::TurnCommand AdvSkills::turn(Angle a)
{
  TurnCommand com;

  // NOT an angle since we need to know if we turn to hard (inertiaMoment)
  double delta = double(a - world->my.bodyDir) * world->my.inertiaMoment();
  com.state = (fabs(delta) < turnMinAngle) ? RFINISHED : ROK;

  if (delta > M_PI * MAXMOMENT / 180.0) {
    delta = M_PI * MAXMOMENT / 180.0;
    com.state = RANGLE;
  }

  if (delta < M_PI * MINMOMENT / 180.0) {
    delta = M_PI * MINMOMENT / 180.0;
    com.state = RANGLE;
  }

  com.turnMoment = delta;
  com.bodyDir = world->my.bodyDir +
    Angle(com.turnMoment / world->my.inertiaMoment());
  return com;
}

AdvSkills::TurnCommand AdvSkills::turnTo(const WObject &obj)
{
  return turn((obj.pos - world->my.pos).getAngle());
}

void AdvSkills::exec(const AdvSkills::TurnCommand &com) { 
  basicTurn(com.turnMoment);
}

/***************************** turnHead *******************************/

AdvSkills::TurnHeadCommand AdvSkills::turnHead(Angle a)
{
  TurnHeadCommand com;
  com.state = THOK;
  com.vwidth = VIEW_WIDTH_NORMAL;

  if (blockTH) {
    cerr << "call of Skills::turnHead before command set" << endl;
    com.state = THBAN;
    return com;
  }

  Angle myNextBodyDir = world->my.bodyDir + Angle(turnParam/world->my.inertiaMoment());

  a -= myNextBodyDir;

  if (double(a) > M_PI * MAXNECKANG / 180.0) {
    a = M_PI * MAXNECKANG / 180.0;
    com.state = THANGLE;
  }
  else if (double(a) < M_PI * MINNECKANG / 180.0) {
    a = M_PI * MINNECKANG / 180.0;
    com.state = THANGLE;
  }

  // NOT an angle, since we need to know if we turn too hard
  double delta = double(a) - double(world->my.headAngle);

  if ((com.state == THOK) && (fabs(delta) < turnMinAngle)) {
    com.state = THFINISHED;
  }

  if (delta > M_PI * MAXNECKMOMENT / 180.0) {
    delta = M_PI * MAXNECKMOMENT / 180.0;
    com.state = THANGLE;
  }
  else if (delta < M_PI * MINNECKMOMENT / 180.0) {
    delta = M_PI * MINNECKMOMENT / 180.0;
    com.state = THANGLE;
  }

  com.faceDir = world->my.faceDir + Angle(delta);
  com.turnHeadMoment = delta;
  return com;
}

AdvSkills::TurnHeadCommand
AdvSkills::turnHead(const WObject &obj)
{
  // [!] remember dashes and take them into account here 
  return turnHead((obj.pos - world->my.pos - world->my.v).getAngle());
}

void AdvSkills::enterExpTurnHeadTurn(double _th)
{
  if (!expmode) {
    btParam  = turnParam;
    bblockTH = blockTH;
    expmode  = true;
    blockTH  = false;
  }

  turnParam = _th;
}

void AdvSkills::leaveExpTurnHeadTurn()
{
  if (!expmode)
    return;

  turnParam = btParam;
  expmode  = false;
  blockTH  = bblockTH;
}

void AdvSkills::exec(const AdvSkills::TurnHeadCommand &com) {
  if (com.state == THBAN)
    return;
  
  // change of view width
  if (com.vwidth != world->viewWidth)
    comm->sendChangeView(com.vwidth, VIEW_QUALITY_HIGH);
  
  comm->sendTurnNeck(com.turnHeadMoment);
}

/*********************************** pointto ******************************/

AdvSkills::PointToCommand
AdvSkills::pointTo(const Vector2d &pos)
{
  PointToCommand com;
  WPlayer my(world->my);
  com.off = false;
  com.dist = norm(pos - my.pos);
  com.dir  = (pos - my.pos).getAngle() - my.faceDir - 
    Angle(turnParam) -  Angle(turnHeadParam);

  if (world->cyclesSincePointTo <= POINTTO_BAN_CYCLES) {
    com.state = PTBAN;   
    return com;
  }
  com.state = PTOK;
  return com;
}

AdvSkills::PointToCommand
AdvSkills::pointToOff()
{
  PointToCommand com;
  com.off = true;
  if (world->cyclesSincePointTo <= POINTTO_BAN_CYCLES) {
    com.state = PTBAN;   
    return com;
  }
  com.state = PTOK;
  return com;
}

void AdvSkills::exec(const AdvSkills::PointToCommand &com) {
  if (com.state == PTBAN)
    return;
  
  if (com.off)
    basicPointToOff();
  else
    basicPointTo(com.dist, com.dir);
}

/*********************************** tackle ******************************/
double
AdvSkills::tackleProb() {
  Vector2d p2b(world->ball.pos - world->my.pos);
  p2b.turn(-world->my.bodyDir);

  const double d(p2b.getX() > 0 ? TACKLE_DIST : TACKLE_BACK_DIST);
  const double w(TACKLE_WIDTH);
  const double dwp(power(d*w,TACKLE_EXP));

  const double P(1 - (power(p2b.getX()*w, TACKLE_EXP) +
		      power(fabs(p2b.getY())*d, TACKLE_EXP)) /
	  	     dwp);
    
  return max(0., min(1., P));
}

AdvSkills::TackleCommand
AdvSkills::tackle()
{
  TackleCommand com;

  com.state = TOK;

  const WBall &ball(world->ball);
  const WPlayer &my(world->my);

  if (my.ballKickable(ball)) {
    com.state = TKICKABLE;
    return com;
  }

  com.failProb = 1-tackleProb();

  if (com.failProb > .5) {
    com.state = TRANGE;
    return com;
  }

  com.power = MAXPOWER/TACKLE_POWER_RATE;

  Vector2d dir;
  dir.setRadAng(MAXPOWER, my.bodyDir);

  com.dir = ball.v+dir;

  double nrm(norm(com.dir));
  if (nrm) 
    com.dir.stretch(1./nrm);

  return com;
}

AdvSkills::TackleCommand
AdvSkills::tackle(const double &dist, const double &tgt_v)
{
  TackleCommand com;

  com.state = TOK;

  const WBall &ball(world->ball);
  const WPlayer &my(world->my);

  if (my.ballKickable(ball)) {
    com.state = TKICKABLE;
    return com;
  }

  com.failProb = 1-tackleProb();

  if (com.failProb > .5) {
    com.state = TRANGE;
    return com;
  }

  double tgtv = max(tgt_v, 0.);

  Vector2d dir, v(ball.v), tgt(ball.v);
  dir.setRadAng(1., my.bodyDir);

  double p, r;
  p = v*dir;
  r = sqr(p) - (sqrnorm(v) - sqr(dist));

  double x1(-p);
  if (r >= 0)
    x1 += sqrt(r);
  else {
    com.state = TTOOFAST;
    x1 += fabs(p); // assume dist == |v|
  }
  tgt += x1*dir;

  double nrm(norm(tgt));
  if (nrm) {
    com.dir = (1./nrm)*tgt;

    tgtv = tgtv + (1-BALL_DECAY)*nrm;
    if (tgtv > BALL_SPEED_MAX) {
      tgtv = BALL_SPEED_MAX;
      com.state = TTOOSHORT;
    }
    tgt.stretch(tgtv/nrm);
  } else {
    double nrmv(norm(v));
    com.dir = nrmv ? (-1./nrmv)*v : v;
  }

  com.power = norm(tgt-v);
  if (x1 >= 0) {
    if (com.power > MAXPOWER)
      com.power = MAXPOWER;
  }
  else {
    com.power *= -1;
    if (com.power < MINPOWER)
      com.power = MINPOWER;
  }

  com.power /= TACKLE_POWER_RATE;

  return com;
}

void AdvSkills::exec(const AdvSkills::TackleCommand &com) {
  if (com.state == TRANGE) {
    cerr << world->time << "(" << comm->getMyTeam() 
	 << " No. " << world->my.number << ")" 
	 << " AdvSkills::exec(TackleCommand): ball too far away to tackle! (failprob=" << com.failProb << ")" << endl;
    noAction();
    return;
  }
  if (com.state == TKICKABLE) {
    cerr << world->time << "(MRB No. " << world->my.number << ")" 
	 << " AdvSkills::exec(TackleCommand): ball in kickrange" << endl;
    noAction();
    return;
  }

  basicTackle(com.power);
}
