//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <iostream>
#include <stdlib.h>
#include <assert.h>

#include "Config.h"

using namespace std;

struct Entry {
	Entry *next;
	char *key;
	char *value;
};

Entry *first=0;


static void setEntry (const char *key, const char *value)
{
	Entry *e;
	for (e = first; e; e=e->next) {
		if (strcmp (key, e->key) == 0) {
			free(e->value);
			e->value = strdup(value);
			return;
		}
	}
	
	e = new Entry;
	e->key = strdup(key);
	e->value = strdup(value);
	e->next = first;
	first = e;
}

static void removeComment (char *str)
{
	char *comment = strchr (str, '#');
	if (comment) {
		*comment = 0;
	}
}

static void parseLine (char *str, int count)
{
	removeComment(str);
	char *equalSign = strchr (str, '=');
	if (!equalSign) {
		char dummy[256];
		if (sscanf (str, "%s", dummy) == 1) {
			cerr << "error parsing config file. line " << count << ": can't find '=' sign." << endl;
		}
		return;
	}

	*equalSign = 0;
	
	char key [256];
	char value [256];
	if (sscanf (str, "%s", key) != 1) {
		cerr << "error parsing config file. line " << count << ": can't parse key" << endl;
		return;
	}
	equalSign++; 
    
	while ( *equalSign == ' ')
	  equalSign++;

	strcpy(value,equalSign);
	//cerr << "key: " << key << " value: ." << value << endl;
	/*
	double x,y;
	if (sscanf (equalSign+1, "%lf , %lf", &x, &y) == 2) {
		sprintf (value, "%f,%f", x, y);
	} else if (sscanf (equalSign+1, "%s", value) != 1) {
		cerr << "error parsing config file. line " << count << ": can't parse value" << endl;
		return;
	}
	*/
	setEntry (key, value);
	// cerr << key << ":" << value << endl;
}


const char * Config::getValue (const char *key)
{
	assert(key);
	Entry *e;
	for (e = first; e; e=e->next) {
		if (strcmp (key, e->key) == 0) {
			return e->value;
		}
	}
	cerr << "WARNING: configuration entry for '" << key << "' not found." << endl;
	return 0;
}

double Config::getValueAsDouble(const char *key)
{
	const char *val = getValue(key);
	if (val == 0) {
		cerr << "FATAL: no value for key '" << key << "' found." << endl;
		exit(10);
	}
	double result;
	if (sscanf(val, "%lf", &result) != 1) {
		cerr << "FATAL: value '" << val << "' for key '" << key << "' not a double." << endl;
		exit(10);
	}
	return result;
}

double Config::getValueAsDoubleFromList(const char *key, int No)
{
	const char *val = getValue(key);
	const char *nextval=val;
	if (val == 0) {
		cerr << "FATAL: no value for key '" << key << "' found." << endl;
		exit(10);
	}
	double result;

	int i=0;
	while ( (No!=i) && ((int) nextval!=1) )
	  {
	    nextval = strchr (nextval, ',')+1;
	    i++;
	  }
	
	if ((int) nextval==1) {
	  cerr << "FATAL: value '" << val << "' for key '" << key << "' has not enough entries for No: "<< No << "." << endl;
	  exit(10);
	}
	    
	if (sscanf(nextval, "%lf", &result) != 1) {
		cerr << "FATAL: value '" << nextval << "' for key '" << key << "' not a double." << endl;
		exit(10);
	}
	return result;
}

int Config::getValueAsInt (const char *key)
{
	const char *val = getValue(key);
	if (val == 0) {
		cerr << "FATAL: no value for key '" << key << "' found." << endl;
		exit(10);
	}
	int result;
	if (sscanf(val, "%d", &result) != 1) {
		cerr << "FATAL: value '" << val << "' for key '" << key << "' not an integer." << endl;
		exit(10);
	}
	return result;
}

int Config::getValueAsIntFromList (const char *key, int No)
{
	const char *val = getValue(key);
	const char *nextval=val;
	if (val == 0) {
		cerr << "FATAL: no value for key '" << key << "' found." << endl;
		exit(10);
	}
	int result;

	int i=0;
	while ( (No!=i) && ((int) nextval!=1) )
	  {
	    nextval = strchr (nextval, ',')+1;
	    i++;
	  }
	
	if ((int) nextval==1) {
	  cerr << "FATAL: value '" << nextval << "' for key '" << key << "' has not enough entries for No: "<< No << "." << endl;
	  exit(10);
	}
	    
	if (sscanf(nextval, "%d", &result) != 1) {
		cerr << "FATAL: value '" << nextval << "' for key '" << key << "' not an integer." << endl;
		exit(10);
	}
	return result;
}

Vector2d Config::getValueAsVector (const char *key)
{
	const char *val = getValue(key);
	if (val == 0) {
		cerr << "FATAL: no value for key '" << key << "' found." << endl;
		exit(10);
	}
	double x,y;
	if (sscanf(val, "%lf , %lf", &x,&y) != 2) {
		cerr << "FATAL: value '" << val << "' for key '" << key << "' not an vector." << endl;
		exit(10);
	}
	return Vector2d(x,y);
	
}


bool Config::readFile (const char *name) 
{
	FILE *f = fopen (name, "r");
	if (!f) {
		return false;
	}

	int line=1;
	char str [1024];
	while (fgets (str, sizeof(str), f) == str) {
		int len;
		if (str[len=strlen(str)-1] == '\n') {
			str[len]=0;
		}
		parseLine(str, line);
		line++;
	}
	fclose(f);
	return true;
}

#if 0
int main()
{
	Config::readFile("config.txt");

	cerr << "node0:" << Config::getValue("node0") << endl;
	Config::getValue("bla");
	return 0;
}
#endif
