//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "onlinecoach.h"
#include "Config.h"
#include "Trace.h"

using namespace std;

#undef TRACE
#define TRACE(cmd)

Onlinecoach::Onlinecoach (const char*  team_name,             
			  const char*  host,
			  int          port,
			  const char*  configfile,
			  const char*  traceCfg,
                          bool         FUNSSELmode) :

  world(NULL),
  comm(NULL)
{
  TRACE (cout << "Onlinecoach: initializing config reader" << endl);

  // try to initialize config file
  if (!Config::readFile(configfile)) {
    cerr << "PANIC: Could not read config file: " << configfile << endl;
    exit(-1);
  }

  TRACE (cout << "Onlinecoach: generating Comm" << endl);
    
  comm = new Comm(team_name,
		  CLIENT_ONLINECOACH,
		  host, port,
		  0,
		  FUNSSELmode);

  comm_error ce;

  if (comm->isServerAlive()) {
    TRACE (cout << "Onlinecoach: connecting" << endl);
    ce = comm->connect();
  }
  else
    ce = COMM_ERROR;

#ifdef MRBDEBUG
  theTrace_init(FUNSSELmode, traceCfg);
#endif

  switch (ce) {
  case (COMM_OK) : 
      
    TRACE (cout << "Onlinecoach: generating Worldmodel" << endl);
    
#ifdef MRBDEBUG
    theTrace->setPlayer(comm->getMyNumber(), comm->getMySide());
#endif

    world = new Worldmodel (comm);

    break;
  case (COMM_ERROR): 
    cerr << "Onlinecoach: unspecified error when initializing !" 
	 << endl; break;
  case (COMM_ERROR_SERVER_NOT_FOUND):
    cerr << "Onlinecoach: could not find server's host !"<< endl; break;
  case (COMM_ERROR_SOCKET_INIT):
    cerr << "Onlinecoach: could not create and bind socket !" << endl; break;
  case (COMM_ERROR_SERVER_INIT):
    cerr << "Onlinecoach: could not deliver checkin message to server !" 
	 << endl; break;
  case (COMM_ERROR_PARSE):
    cerr << "Onlinecoach: could not parse init response !" << endl; break;
  case (COMM_ERROR_NO_DATA):
    cerr << "Onlinecoach: did not receive response from server !" 
	 << endl; break;
  }
       
}


Onlinecoach::~Onlinecoach () {

  TRACE (cout << "Onlinecoach: entering destructor, closing down" << endl);

  delete (world);
  delete (comm);

  TRACE (cout << "Onlinecoach: leaving destructor" << endl);

#ifdef MRBDEBUG
  theTrace_stop();
#endif
}


void 
Onlinecoach::coach() {

  if (comm->isServerAlive()) {
    TRACE (cout << "Onlinecoach: started" << endl);

    while (comm->isServerAlive() && !comm->isGameOver()) {
      
      TRACE (cout << "Onlinecoach: waiting for signals and handling them" << endl);
      comm->drainSocket();

      if (comm->isActionNeeded()) {
	
	TRACE (cout << "Onlinecoach: adjusting worldmodel" << endl);
	world->adjust();
	
	TRACE (cout << "Onlinecoach: current timestep is " << getTime() << endl);
	
#ifdef MRBDEBUG
	theTrace->write(*world);
#endif

	TRACE (cout << "Onlinecoach: analyzing situation" << endl);
	analyze();
	  
	if (getPlayMode() != PLAY_MODE_PLAY_ON) {
	  TRACE (cout << "Onlinecoach: play_mode != play_on, acting" << endl);
	  act();
	}
	
	if( SYNCH_MODE ) {
	    sendDone();
	    TRACE(cout << world->time << " Onlinecoach: done" << endl);
	}
      }
      else {
	TRACE(cout << "Onlinecoach: no new timestep yet" << endl);
      }

    } // while server alive

    TRACE (cout << "Onlinecoach: server is gone, leaving" << endl);
    
  } else { 
    // server never was alive
    cerr << "Onlinecoach: could not connect to sserver, closing down" << endl;
  }
     
} //play



