//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "player.h"
#include "Config.h"
#include "Trace.h"

using namespace std;

#undef TRACE
#define TRACE(cmd)

Player::Player (const char* team_name,             
		bool        isGoalie, 
		const char* host,
		int         port,
		const char* configfile,
		const char* traceCfg,
		int         reconnectNumber,
                bool        FUNSSELmode,
		int         clangMinVer,
		int         clangMaxVer) :

  skills(NULL), 
  world(NULL),
  comm(NULL)
{
  TRACE (cout << "Player: initializing config reader" << endl);

  // try to initialize config file
  if (!Config::readFile(configfile)) {
    cerr << "PANIC: Could not read config file: " << configfile << endl;
    exit(-1);
  }

  TRACE (cout << "Player: generating Comm" << endl);

  TRACE (cout << "Player: initializing Comm" << endl);
  
  comm = new Comm(team_name, 
		  isGoalie ? CLIENT_GOALIE : CLIENT_FIELDPLAYER,
		  host, port,
		  reconnectNumber,
		  FUNSSELmode,
		  clangMinVer,
		  clangMaxVer);
  
  comm_error ce;

  if (comm->isServerAlive())
    ce = comm->connect();
  else
    ce = COMM_ERROR;


#ifdef MRBDEBUG
  theTrace_init(FUNSSELmode, traceCfg);
#endif
  switch (ce) {
  case (COMM_OK): 
    
    TRACE (cout << "Player: generating Worldmodel and AdvSkills" << endl);
    
#ifdef MRBDEBUG
    theTrace->setPlayer(comm->getMyNumber(), comm->getMySide());
#endif

    world  = new Worldmodel (comm);
    skills = new AdvSkills (comm, world);

    break;
  case (COMM_ERROR): 
    cerr << "Player: unspecified error when initializing !" << endl; break;
  case (COMM_ERROR_SERVER_NOT_FOUND):
    cerr << "Player: could not find server's host !"<< endl; break;
  case (COMM_ERROR_SOCKET_INIT):
    cerr << "Player: could not create and bind socket !" << endl; break;
  case (COMM_ERROR_SERVER_INIT):
    cerr << "Player: could not deliver checkin message to server !" 
	 << endl; break;
  case (COMM_ERROR_PARSE):
    cerr << "Player: could not parse init response !" << endl; break;
  case (COMM_ERROR_NO_DATA):
    cerr << "Player: did not receive response from server !" << endl; break;
  }
} 


Player::~Player () {

  cerr << "player ends" << endl;
  TRACE (cout << "Player: entering destructor, closing down" << endl);
  delete (skills);
  delete (world);
  delete (comm);

#ifdef MRBDEBUG
  theTrace_stop();
#endif
  TRACE (cout << "Player: leaving destructor" << endl);

}


void 
Player::play() {

  if (comm->isServerAlive()) {
    TRACE (cout << "Player: started" << endl);
    
    while ( comm->isServerAlive() && !comm->isGameOver() ) {

      TRACE (cout << "Player: waiting msgs" << endl);
      comm->drainSocket();

      if (comm->isActionNeeded()) {

	TRACE (cout << "Player: need action for time " << getTime() << endl); 

	TRACE (cout << "Player: acting" << endl);

	act();

	if( SYNCH_MODE )
	    sendDone();

#ifdef MRBDEBUG
	theTrace->write(*world);
#endif

      } else {
	TRACE(cout << "Player: do not need new action yet" << endl);
      }

    } // while server alive

    TRACE (cout << "Player: server is gone, leaving" << endl);

  } else {
    // server never was alive
    cerr << "Player: could not connect to sserver, closing down" << endl;
  }

} //play
