//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef __PLAYER_H__
#define __PLAYER_H__

#include <unistd.h>
#include <iostream>
#include <fstream>
#include <sstream>

#include "constants.h"
#include "Comm.h"
#include "worldmodel.h"
#include "AdvSkills.h"
#include "audiomgr.h"


class Player {

public:


  // Constructor / Destructor
  ///////////////////////////////////////////////////////////////////////////

  // Generates and logs in a player client at the given soccerserver.
  // After the constructor has finished all properties of the agent are set.
  // 
  // Call this constructor first in your derived classes constructors before
  // you implement any further initialization of the agent.

  Player (const char* team_name,             // name of own team
	  bool        isGoalie,              // init as goalie ?
	  const char* host         = DEFAULT_HOST,
	  int         port         = PORT,   
	  const char* configfile   = "agent.conf",
	  const char* traceCfg     = "trace.conf",
	  int         reconnectNumber = 0,   // reconnecting (only if number>0)
	  bool FUNSSELmode = false,          // tells if we can send draw commands etc.
	  int         clangMinVer = 7,       // which coachlanguage versions should
	  int         clangMaxVer = 7);      // be supported

  // Starts execution of the Player client. 
  // After calling this method, the agent will repeatedly call act
  // every time it needs to send an action while the server is running.
  //
  // You cannot override this method.
  void play();

  // Called once every time player needs to send an action to the server.
  //
  // Override this method virtual in your derived classes to implement 
  // player behaviour.
  virtual void act() = 0;

  
  // Cleans up and logs out
  //
  // You can put own code in your derived destructors. As the final action call
  // this destructor so the agent will leave the simulation properly 
  // and release its resources (files, sockets, windows, etc).
  virtual ~Player();  


  
  // general information about the client himself and the server state
  /////////////////////////////////////////////////////////////////////

  // obvious
  int          getMyNumber()  {return comm->getMyNumber();}
  bool         amIGoalie()    {return comm->amIGoalie();}
  const char*  getMyTeam()    {return comm->getMyTeam();}
  playing_side getMySide()    {return comm->getMySide();}
  play_mode    getPlayMode()  {return comm->getPlayMode();}

  void         getLastAction(CycleAction& action) {comm->getLastAction(action);}

  // returns the timestep the server is in, our current time.
  // (There is only ONE current time: This is it.)
  long         getTime() {return comm->getServerTime();} 

  // sends an audiomessage to other clients immediately
  void         say(AudioMsg &msg) {comm->say(msg);}

  // sends a short audiomessage to other clients (server 8.*)
  void         say_10( char * msg ) {comm->say_10(msg);}

  // sends "(done)" to the server in synch_mode
  void         sendDone( void ) {comm->sendDone();};

  // returns true, if a new audiomessage is available in msg, false otherwise
  bool         getNextMsg(AudioMsg &msg) {return comm->getNextMsg(msg);}

  // the clients modules
  //////////////////////////////////////////////////////////////////////////

  AdvSkills *skills;
  Worldmodel *world;

private:

  Comm *comm;

  std::ofstream out;

};


#endif
