//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "ballhandlingmodule.h"
#define FUNSSELMACROS
#include "drawMacros.h"
#include <sstream>

#define MOD_HAND
#include "Trace.h"

using namespace std;

BallHandlingModule::BallHandlingModule( Worldmodel * _world, TacticalModule * _tact, AdvSkills * _skills )
{
  world  = _world;
  tact   = _tact;
  skills = _skills;

  opponentControlsBall = false;
  lookingForBall = false;
  lastGoalieInterceptTime = -2;

  placeCycles = 0;
  haveSeenBallSincePass = false;
  action = WAIT;
  lastAction = WAIT;
  lastActCycles = ETERNAL;
}


/////////////////////////////////////////////////////////////////////////
//
// FEEDBACK
//
/////////////////////////////////////////////////////////////////////////

void 
BallHandlingModule::feedbackWorld()
{
  // store variables that are needed very often
  my   = world->my;
  ball = world->ball;

  if( lastActCycles < ETERNAL )
    lastActCycles++;

  if( lastAction == PLAYER_PLACE && lastActCycles == 1 )
    placeCycles++;
  else
    placeCycles = 0;

  // Soccerserver checks for <= stamina_limits, so rather stay slightly above:
  availableStamina = max( 0.0, my.stamina - EFFORT_DEC_THR * STAMINA_MAX - 1 );

  if( ball.age <= 1 )
    lookingForBall = false;

  opponentControlsBall = false;
  for( int i = 1; i <= 11; i++ )
    if( world->otherTeam[i].distTo( ball ) < world->otherTeam[i].kickRange() )
      opponentControlsBall = true;

  // opponent might have placed ball
  if( ball.age == 1 && !opponentControlsBall )
    {
      WBall simBall = ball;
      simBall.pos -= ( 1.0 / BALL_DECAY ) * simBall.v;
      for( int i = 1; i <= 11; i++ )
	if( world->otherTeam[i].distTo( simBall ) < world->otherTeam[i].kickRange() )
	  opponentControlsBall = true;
    }
}


/////////////////////////////////////////////////////////////////////////
//
// EVALUATE
//
/////////////////////////////////////////////////////////////////////////
#include "ballhandlingEvaluate.C"


/////////////////////////////////////////////////////////////////////////
//
// ACT
//
/////////////////////////////////////////////////////////////////////////
#include "ballhandlingAct.C"
