//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef __GOALSHOTMODULE_H__
#define __GOALSHOTMODULE_H__

#include "grade.h"
#include "goalshoteval.h"

#include "worldmodel.h"
#include "tacticalmodule.h"
#include "audiomgr.h"
#include "AdvSkills.h"
#include "mcommand.h"
#include <sstream>


//////////////////////////////////////////////////////////////////////
//
// class GoalShotModule
//
//////////////////////////////////////////////////////////////////////

class GoalShotModule
{
public:
  // constructor of module GoalShot 
  GoalShotModule( Worldmodel * _world, TacticalModule * _tact, AdvSkills * _skills, GoalShotEval * _goalshoteval );

  // destructor of module GoalShot 
  ~GoalShotModule() {}
  
  // feed information about new world and tactics into module
  bool feedbackWorld();

  // Do the evaluation for the next action
  Grade evaluate();

  // Calculate a turn head wish
  TurnHeadWish turnHeadEvaluate();

  // Perform the action for this module  
  bool act();

private:
  
  // player's worldmodel
  Worldmodel *world;

  // player's tactical module
  TacticalModule *tact;

  // player's advanced skills
  AdvSkills *skills;

  // evaluation class
  GoalShotEval *goalshoteval;

  // did the player perform a goalshot that has been finished?
  bool finishedKick;

  // did the player start a goalshot that takes more than one cycle?
  bool startedKick;

  // should we tackle?
  bool tackleToGoal;

  // a calculated tackle command
  AdvSkills::TackleCommand tackleCommand;
  
  // a calculated kick command
  AdvSkills::ExtKickCommand kickCommand;

  // the player himself
  WPlayer my;

  // the time the player has to prepare his kick-command
  long availableKickTime;
  
  // the point on the goal line that player should aim at
  Vector2d goalTarget;

}; // class GoalShotModule

#endif











