//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////
#define FUNSSELMACROS

#include "passmodule.h"
#include "passeval.h"
#include "goalshoteval.h"
#include "drawMacros.h"
#include "mcommand.h"

using namespace std;

// the TRACEs
/////////////////////

#undef USEPASSTRACE
#undef USEDRAWTRACE

#ifdef MRBDEBUG

// comment out the following line to disable trace
//#define USEPASSTRACE
//#define USEDRAWTRACE

#endif

using namespace std;

#ifdef USEPASSTRACE
#define PASSTRACE(cmd) {ostringstream ost;ost << cmd; Comm::sendMCmd(MLog(ost.str()));}
#warning PassModule debug trace PASSTRACE is activated!
#else
#define PASSTRACE(cmd)
#endif


#ifdef USEDRAWTRACE
#define DRAWTRACE(cmd) cmd
#warning PassModule debug trace DRAWTRACE is activated
#else
#define DRAWTRACE(cmd)
#endif


//////////////////////////////////////////////////////////////////////
//
// class PassModule
//
//////////////////////////////////////////////////////////////////////

PassModule::PassModule(Worldmodel * _world, TacticalModule * _tact, AdvSkills * _skills,
		       PassEval * _passeval, const GoalShotEval * _goalshoteval):
  world(_world),
  tact(_tact),
  skills(_skills),
  callMeAgain(false),
  passeval(_passeval),
  goalshoteval(_goalshoteval),
  timeForKick(0),
  lookAfterPass(false),
  lookAfterPassTime(-3)   // hope that's enough
{
  turnHeadGrade.set_IMPOSSIBLE();
}


//////////////////////////////////////////////////////////////////////
//
// bool PassModule::feedbackWorld
//
//////////////////////////////////////////////////////////////////////

bool PassModule::feedbackWorld()
{
  // if pass is over check if it failed
  if( lookAfterPass )
    {
      // if ball velocity does not point to target something must be wrong
      if ( fabs( angleBetween( world->ball.v, lookAfterPassPos - world->my.pos ) ) > PI_4 ||
	   world->ball.v.getLength() < 0.1 )   // collision? [C] 0.1?
	{
	  // tactical action can not stay TA_PASS
	  if( tact->ourAction.state == TA_PASS &&
	      tact->ourAction.performer == world->my.number )
	    {
	      PASSTRACE( " PASS: Pass seems to have gone bananas -> change tacticalAction" );

	      tact->ourAction.state = TA_UNKNOWN;
	      tact->ourAction.quality = TQ_GUESSED;
	      tact->ourAction.age = 0;
	      tact->ourAction.performer = UNKNOWN;
	      tact->ourAction.receiver = UNKNOWN;
	    }
	}
    }

  return true;
}


//////////////////////////////////////////////////////////////////////
//
// bool PassModule::encodeAudio
//
//////////////////////////////////////////////////////////////////////

Grade PassModule::encodeAudio( char* msg )
{
  long c = 0;
  PropOf prop;
  prop.grade.set_POOR(); 
  // [MW] in case we haven't seen the firstResponsiblePlayer we don't want to report that
  // we are not able to intercept a pass. If we report BAD that could be missunderstood with
  // "just tackled".
  prop.xIndex = 0;
  prop.yIndex = 0;

  // if I am the goalie I never want a pass!
  if( world->my.isGoalie == GOALIE_TRUE )
    c = 72;
  else
    {
      if( tact->firstResponsiblePlayer != UNKNOWN )
	if( world->ourTeam[tact->firstResponsiblePlayer].age <= 2 && 
	    tact->attentionTo == tact->firstResponsiblePlayer )
	  if( world->ourTeam[tact->firstResponsiblePlayer].distTo( world->ball ) < VISIBLE_DISTANCE )
	    prop = passeval->ourEvalProp( world, tact, tact->firstResponsiblePlayer, world->my.number );
      
      PASSTRACE( "PASS: eval -> grading pass: " << prop
		 << ", pl.age: " << world->ourTeam[tact->firstResponsiblePlayer].age );
      
      c = ( prop.grade.is_PERFECT() ? 0 :
	    prop.grade.is_GOOD() ? 1 :
	    prop.grade.is_AVERAGE() ? 2 :
	    prop.grade.is_POOR() ? 3 : 72 );

      if( c != 72 )
	{
	  if( prop.xIndex >= 3 )
	    c += 4 * 15;
	  else
	    c += 4 * ( prop.yIndex + 2 + 5 * prop.xIndex );
	}
    }

  msg[9] = msgEncodeChars[(int)char( c )];
  return prop.grade;
}


//////////////////////////////////////////////////////////////////////
//
// Grade PassModule::evaluate
//
//////////////////////////////////////////////////////////////////////

Grade PassModule::evaluate()
{
  // deactivate turnHeadGrade
  turnHeadGrade.set_IMPOSSIBLE();

  // a priori no emergency kick situation
  performEmergencyPass = false;

  // maybe we started a pass one cycle ago
  // if ball is already too fast, keep this decision and perform a rekick
  // otherwise calculate pass target again (if the data has not changed
  // the result should be the same!)
  if( callMeAgain && world->my.ballKickable( world->ball ) &&
      world->ball.v.getLength() > 0.8 * skills->stopableV )
    {
      PASSTRACE( "PASS: eval -> rekick needed and ball is fast => ABSOLUTE" );
      Grade g( Grade::ABSOLUTE );
      returnGRADE( g );
    }

  callMeAgain = false;

  // time for looking after ball is over
  if( lookAfterPass && lookAfterPassTime + 2 < world->time )
    {
      lookAfterPass = false;
      PASSTRACE( "PASS: eval -> don't want to look after the ball anymore" );
    }

  // prop for best pass
  PropOf prop;
  prop.score = passeval->SCORE_POOR + 1;
  prop.grade.set_BAD();
  prop.number = UNKNOWN;
  prop.offside = false;
  prop.pos = world->otherGoal.pos;
  prop.dist = world->my.distTo( prop.pos );
  prop.vel = BALL_SPEED_MAX;
  prop.xIndex = 0;
  prop.yIndex = 0;

  // prop for best pass with worst agescore
  PropOf lookProp;
  lookProp.score = passeval->SCORE_POOR + 1;
  lookProp.agescore = 0;
  lookProp.grade.set_BAD();
  lookProp.number = UNKNOWN;
  lookProp.offside = false;
  lookProp.pos = world->otherGoal.pos;
  lookProp.dist = world->my.distTo( prop.pos );
  lookProp.vel = BALL_SPEED_MAX;
  lookProp.xIndex = 0;
  lookProp.yIndex = 0;

  // if ball is near evaluate passes and maybe turn head to pass partner
  if( ( world->my.distTo( world->ball.pos ) < VISIBLE_DISTANCE && world->ball.age <= 3 ) ||
      world->my.ballKickable( world->ball ) )
    {
      PASSTRACE( "PASS: eval -> feeling ball & ball is safe; try to look into the right direction" );

      DRAWTRACE(Comm::sendMCmd(CIRCLE(world->my.pos,passeval->MAXPASS_LENGTH)));
      DRAWTRACE(Comm::sendMCmd(CIRCLE(world->my.pos,passeval->SHORTPASS_LENGTH)));

      // time available for kick
      timeForKick = passeval->timeForPass( world, world->my.number );
      double maxBallSpeed = BALL_SPEED_MAX;

      /*
      if( timeForKick == 1 && world->my.ballKickable( world->ball ) )
	maxBallSpeed = MAXPOWER * world->my.kickPowerRate( world->ball );
      */

      double offsideLine = world->ourOffsideLine( MAXAGE );

      PropOf helpProp;

      for( int panumber = 1; panumber <= 11; panumber++ )
	{
	  if( panumber == world->my.number )
	    continue;
	  if( world->ourTeam[panumber].age > MAXAGE )
	    continue;
	  if( panumber == world->ourGoalieNo )
	    continue;
	  if( world->ourGoalieNo == UNKNOWN && world->ourTeam[panumber].pos.getX() < -FIELD_LENGTH / 3.0 )
	    continue;
	  if( world->my.distTo( world->ourTeam[panumber] ) > passeval->MAXPASS_LENGTH )
	    continue;

	  // check if we have got audio information from that player
	  bool useAudio = false;

	  if( tact->playerPassInfo[panumber].age < 4 )   // !!! [C]
	    {
	      PASSTRACE( "PASS: eval -> audio info from player " << panumber << ": grade: "
			 << tact->playerPassInfo[panumber].grade
			 << " age: " << tact->playerPassInfo[panumber].age
			 << " Index: " << tact->playerPassInfo[panumber].xIndex
			 << "," << tact->playerPassInfo[panumber].yIndex );

	      // either our passpartner doesn't want a pass or he has just tackled
	      if( tact->playerPassInfo[panumber].grade.is_BAD() )
		{
		  PASSTRACE( "PASS: eval -> will NOT pass to this player" );
		  continue;
		}
	      if( world->ourTeam[panumber].age >= tact->playerPassInfo[panumber].age )
		{
		  useAudio = true;
		  PASSTRACE( "PASS: eval -> will use this information" );
		}
	      else
		{
		  PASSTRACE( "PASS: eval -> will NOT use this information" );
		}
	    }

	  bool defensive = world->ourTeam[panumber].pos.getX() < 0.0 && !tact->ourTeamOffender( panumber );

	  // now check all grid points (indices [i][j])
	  for( int i = 0; i <= ( defensive ? 1 : 10 ); i++ )
	    for( int j = ( defensive ? 1 : 0 ); j <= ( defensive ? 3 : 4 ); j++ )
	      {
		// check all points but the ones marked with "+"
		if( i < 3 || j == 2 )
		  {
		    // calculate the coordinates of this point
		    helpProp.pos = world->ourTeam[panumber].pos + passeval->RPASS_LATTICELENGTH * Vector2d( i, j - 2 );
		    passeval->evalPass( true, world, tact, world->my.number, panumber, offsideLine,
					timeForKick, maxBallSpeed, helpProp );
		    helpProp.xIndex = i;
		    helpProp.yIndex = j - 2;

		    // maybe use audio information
		    if( useAudio && !helpProp.grade.is_BAD() &&
			i == tact->playerPassInfo[panumber].xIndex &&
			j == tact->playerPassInfo[panumber].yIndex )
		      {
			if( tact->playerPassInfo[panumber].grade.is_better_than( helpProp.grade ) )
			  {
			    PASSTRACE( "upgrade pass because of audio information" );
			    helpProp.score -= 50;
			    helpProp.grade = passeval->scoreToGrade( helpProp.score );
			  }
			else if( tact->playerPassInfo[panumber].grade.is_worse_than( helpProp.grade ) )
			  {
			    PASSTRACE( "downgrade pass because of audio information" );
			    helpProp.score += 50;
			    helpProp.grade = passeval->scoreToGrade( helpProp.score );
			  }
		      }

		    if( helpProp.score < prop.score )
		      prop = helpProp;

		    if( world->ourTeam[panumber].coneTime < 2 && 
			world->ourTeam[panumber].age > 0 &&
			helpProp.agescore > passeval->SCORE_PERFECT &&
		        helpProp.score - helpProp.agescore < passeval->SCORE_GOOD &&
		        lookProp.score - lookProp.agescore > helpProp.score - helpProp.agescore &&
			lookProp.agescore <= helpProp.agescore &&
			fabs( world->my.bodyAngleTo( helpProp.pos ) ) < fabs( DEG_TO_RAD( MAXNECKANG + 15 ) ) )
		      { 
			// if this pass is better than lookProp pass and the ageScore is
			// equal or worse, look for this pass
			lookProp = helpProp;
		      }
		  }
	      }
	} // player loop

      if( lookProp.number != UNKNOWN )
	{
	  lookPos = lookProp.pos;

	  if( lookProp.score - lookProp.agescore <= passeval->SCORE_GOOD )
	    turnHeadGrade.set_PERFECT();
	  else
	    turnHeadGrade.set_GOOD();

	  if( lookProp.agescore < passeval->SCORE_GOOD / 2 )
	    turnHeadGrade.downgrade();

	  PASSTRACE( "PASS: eval -> looking for this pass: " << lookProp
		     << ", pl.age: " << world->ourTeam[lookProp.number].age
		     << ", turnHeadGrade: " << turnHeadGrade
		     << ", bodyAng: " << fabs( world->my.bodyAngleTo( lookProp.pos ) ) );
	}
      else
	{
	  PASSTRACE( "PASS: no lookPos found!" );
	  turnHeadGrade.set_IMPOSSIBLE();
	}

      // this was the old way of taking age of pass direction into account
      // maybe we should not delete it...
      /*
      if( prop.grade.is_better_than( Grade::POOR ) )
	{
	  DRAWTRACE( Comm::sendMCmd(MDLine(world->my.pos,prop.pos)) );

	  long passDirAge = world->dirAge( ( prop.pos - world->my.pos ).getAngle() );
	  turnHeadGrade.set_PERFECT();
	  if( passDirAge < 4 )
	    {
	      turnHeadGrade.downgrade();
	      if( passDirAge < 2 )
		turnHeadGrade.downgrade();
	    }
	}
      */
    }

  // kicking (and passing) impossible
  if( !world->my.ballKickable( world->ball ) )
    { 
      PASSTRACE( "PASS: eval -> can't kick the ball => IMPOSSIBLE" );
      Grade g( Grade::IMPOSSIBLE );
      returnGRADE( g );
    }

  // if no player pass was found try an emergency kick
  if( prop.grade.is_BAD() )
    {
      performEmergencyPass = true;
      targetOfEmergencyPass = getEmergencyPass();
      Grade g( Grade::BAD );
      // this is just a hack to keep other modules from making mistakes
      if( world->playMode != PLAY_MODE_PLAY_ON )
	g.upgrade();
      returnGRADE( g );
    }

  // downgrade pass if player is the goalie and should catch instead
  if( prop.grade.is_PERFECT() &&
      world->my.number == world->ourGoalieNo &&
      world->my.inOurPenaltyArea() &&
      world->my.cyclesSinceTackle > TACKLE_CYCLES &&
      world->playMode == PLAY_MODE_PLAY_ON )
    {
      prop.grade.downgrade();
    }

  // if there might have been a ball collision let ballhandling module do the job
  if( world->collState != NONE && prop.grade.is_PERFECT() )
    prop.grade.downgrade();

  // remember pass target
  chosenPass = prop;

  PASSTRACE( " PASS: eval -> best pass: " << prop << ", time for kick : " << timeForKick );
  DRAWTRACE( Comm::sendMCmd(MDPosition(prop.pos)) );
  returnGRADE( prop.grade );
}


//////////////////////////////////////////////////////////////////////
//
// PassModule::turnHeadEvaluate
//
//////////////////////////////////////////////////////////////////////

TurnHeadWish
PassModule::turnHeadEvaluate()
{
  TurnHeadWish thw;

  if( lookAfterPass )
    {
      thw.lookAtPos( Grade::ABSOLUTE, lookAfterPassPos, VIEW_WIDTH_NARROW );
      PASSTRACE( "PASS: turnHeadEvaluate -> want to see if pass is successful, grade = ABSOLUTE" );
    }
  else if( turnHeadGrade.is_better_than( Grade::IMPOSSIBLE ) )
    {
      //[MW] let player attack to its job
      if( tact->ourBallControl != BC_SAFE && world->ball.pos.getX() < 0.0 )
	turnHeadGrade.downgrade();

      PASSTRACE( "PASS: turnHeadEvaluate -> want to look at passpartner "
		 << chosenPass.number << "at " << lookPos << ", grade = " << turnHeadGrade );
      thw.lookAtPos( turnHeadGrade, lookPos, VIEW_WIDTH_NARROW );
    }
  else
    {
      thw.lookNowhere( Grade::IMPOSSIBLE, VIEW_WIDTH_NARROW );
      PASSTRACE( "PASS: turnHeadEvaluate -> don't want to turn head, grade = IMPOSSIBLE" );
    }

  return thw;
}


//////////////////////////////////////////////////////////////////////
//
// PassModule::act
//
//////////////////////////////////////////////////////////////////////

bool PassModule::act()
{
  // if the ball is not kickable, do nothing
  if( !world->my.ballKickable( world->ball ) )
    {
      callMeAgain = false;
      skills->noAction();
      PASSTRACE( "PASS: act -> did no action because ball not kickable" );
      return false;
    }

  // we have to perform an emergency kick (because we did not find a passpartner)
  if( performEmergencyPass )
    {
      callMeAgain = false;
      PASSTRACE( "PASS: act -> performed emergency kick; dir = "
		 << RAD_TO_DEG( targetOfEmergencyPass.getAngle() ) );

      kickparams = skills->kick( targetOfEmergencyPass, AdvSkills::KICKAWAY );
      skills->exec( kickparams );
      return true;
    }

  // if we are called again because a previous kick is not finished, perform a rekick
  if( callMeAgain )
    {
      kickparams = skills->rekick( kickparams );
      PASSTRACE( "PASS: act -> did a rekick" );
    }
  // else perform the calculated kick
  else
    {
      // ball should be stoppable when reaching passpartner
      chosenPass.vel = min( chosenPass.vel, skills->stopableV );
      kickparams = skills->kick( chosenPass.vel, chosenPass.pos, timeForKick );
    }

  skills->exec( kickparams );

  // simulate ball and player and check if ball will leave kickrange
  WBall simBall = world->ball;
  simBall.v = kickparams.v;
  simBall.simulate();

  if( dist( world->my.pos + world->my.v, simBall.pos ) > world->my.kickRange() )
    {
      callMeAgain = false;
      // report the kick
      tact->reportOurTacticalAction( TA_PASS, simBall, world->my.number, chosenPass.number );

      // kick is finished, prepare for looking after ball
      lookAfterPass     = true;
      lookAfterPassTime = world->time;
      lookAfterPassPos  = chosenPass.pos;
      PASSTRACE( "PASS: act -> pass is finished" );
    }
  else
    {
      // pass is not finished, prepare for rekick
      callMeAgain = true;
      PASSTRACE( "PASS: act -> pass is not finished, module might be called again next cycle" );
    }

  DRAWTRACE(Comm::sendMCmd(MDLine(world->my.pos,chosenPass.pos)));
  DRAWTRACE(Comm::sendMCmd(MDVLine(world->ourOffsideLine(MAXAGE))));

  PASSTRACE( " PASS: act -> trying to pass to " << chosenPass.number
	     << " at pos" << chosenPass.pos 
	     << " with " << timeForKick << " cycles time to kick" );
  return true;
}


//////////////////////////////////////////////////////////////////////
//
// PassModule::getEmergencyPass
//
//////////////////////////////////////////////////////////////////////

Vector2d PassModule::getEmergencyPass()
{
  // 9 possible directions for emergency kick
  Angle kickangle[9] = { -PI_2, -PI_4 - PI_8, -PI_4, -PI_8, 0.0, PI_8, 
			 PI_4, PI_4 + PI_8, PI_2 };

  if( world->playMode == PLAY_MODE_CORNER_KICK_US )
    {
      kickangle[0] = -PI;
      kickangle[1] = -PI + PI_8;
      kickangle[2] = -PI + PI_4;
      kickangle[3] = -PI_2 - PI_8;
      kickangle[4] = -PI_2; 
      kickangle[5] =  PI_2;
      kickangle[6] =  PI_2 + PI_8;
      kickangle[7] =  PI_2 + PI_4; 
      kickangle[8] =  PI - PI_8;
    }
  

  // evaluations of the 9 possible directions
  Grade dir[9];
  for( int i = 0; i < 9; i++ )
    dir[i].set_PERFECT();

  Vector2d my_pos = world->my.pos;

  const int AREA_LEFT   = 1;
  const int AREA_MIDDLE = 2;
  const int AREA_RIGHT  = 3;

  const int SI_DEFENSIVE = 1;
  const int SI_MIDFIELD  = 2;
  const int SI_OFFENSIVE = 3;

  int my_area = ( my_pos.getY() < -FIELD_WIDTH / 6.0 ? AREA_LEFT :
		  my_pos.getY() > FIELD_WIDTH / 6.0 ? AREA_RIGHT : AREA_MIDDLE );
  
  int my_situation = ( my_pos.getX() < -FIELD_LENGTH / 6.0 ? SI_DEFENSIVE :
		       my_pos.getX() > FIELD_LENGTH / 6.0 ? SI_OFFENSIVE : SI_MIDFIELD );

  // suppress certain directions due to play mode, situation and my position
  if( world->playMode == PLAY_MODE_FREE_KICK_US || world->playMode == PLAY_MODE_KICK_IN_US )
    {
      switch( my_situation )
	{
	case SI_DEFENSIVE :
	  switch( my_area )
	    {
	    case AREA_LEFT:
	      dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD(); dir[3].set_BAD();
	      dir[4].set_BAD(); dir[7].set_BAD(); dir[8].set_BAD();
	      break;
	    case AREA_MIDDLE:
	      // does not appear during kick ins
	      dir[0].set_BAD(); dir[1].set_BAD(); dir[7].set_BAD(); dir[8].set_BAD();
	      break;
	    case AREA_RIGHT:
	      dir[0].set_BAD(); dir[1].set_BAD(); dir[4].set_BAD(); dir[5].set_BAD();
	      dir[6].set_BAD(); dir[7].set_BAD(); dir[8].set_BAD();
	      break;
	    }
	  break;
	case SI_MIDFIELD:
	  switch( my_area )
	    {
	    case AREA_LEFT:
	      dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD();
	      dir[3].set_BAD(); dir[4].set_BAD(); dir[8].set_BAD();
	      break;
	    case AREA_MIDDLE:
	      // does not appear during kick ins
	      dir[0].set_BAD(); dir[1].set_BAD(); dir[7].set_BAD(); dir[8].set_BAD();
	      break;
	    case AREA_RIGHT:
	      dir[0].set_BAD(); dir[4].set_BAD(); dir[5].set_BAD(); 
	      dir[6].set_BAD(); dir[7].set_BAD(); dir[8].set_BAD();
	      break;
	    }
	  break;
	case SI_OFFENSIVE:
	  switch( my_area )
	    {
	    case AREA_LEFT:
	      dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD();
	      dir[3].set_BAD(); dir[4].set_BAD();
	      break;
	    case AREA_MIDDLE:
	      // does not appear during kick ins
	      break;
	    case AREA_RIGHT:
	      dir[4].set_BAD(); dir[5].set_BAD(); 
	      dir[6].set_BAD(); dir[7].set_BAD(); dir[8].set_BAD();
	      break;
	    }
	  break;
	}
    }
  else
    if( world->playMode == PLAY_MODE_CORNER_KICK_US )
      {
	if( my_area == AREA_LEFT )
	  {
	    dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD();
	    dir[3].set_BAD(); dir[4].set_BAD(); dir[5].set_BAD();
	  }
	else
	  {
	    dir[4].set_BAD(); dir[5].set_BAD(); dir[6].set_BAD();
	    dir[7].set_BAD(); dir[8].set_BAD(); dir[0].set_BAD();
	  }
      }
    else
      {
	switch( my_situation )
	  {
	  case SI_DEFENSIVE :
	    switch( my_area )
	      {
	      case AREA_LEFT   :
		dir[5].set_BAD(); dir[6].set_BAD(); dir[7].set_BAD(); dir[8].set_BAD();
		break;
	      case AREA_MIDDLE :
		break;
	      case AREA_RIGHT  :
		dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD(); dir[3].set_BAD();
		break;
	      }
	    break;
	    
	  case SI_MIDFIELD :
	    switch( my_area )
	      {
	      case AREA_LEFT   :
		dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD();
		dir[6].set_BAD(); dir[7].set_BAD(); dir[8].set_BAD();
		break;
	      case AREA_MIDDLE :
		dir[0].set_BAD(); dir[1].set_BAD();
		dir[7].set_BAD(); dir[8].set_BAD();
		break;
	      case AREA_RIGHT  :
		dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD();
		dir[6].set_BAD(); dir[7].set_BAD(); dir[8].set_BAD();
		break;
	      }
	    break;
	    
	  case SI_OFFENSIVE :
	    switch( my_area )
	      {
	      case AREA_LEFT   :
		dir[0].set_BAD(); dir[1].set_BAD(); dir[2].set_BAD(); dir[3].set_BAD();
		break;
	      case AREA_MIDDLE :
		dir[0].set_BAD(); dir[1].set_BAD();
		dir[7].set_BAD(); dir[8].set_BAD();
		break;
	      case AREA_RIGHT  :
		dir[5].set_BAD(); dir[6].set_BAD(); dir[7].set_BAD(); dir[8].set_BAD();
		break;
	      }
	    break;
	    
	  default :   // should not happen
	    break;
	  }
      }

  for( int i = 0; i < 9; i++ )
    {
      // first check if a short and narrow corridor is free
      if( world->otherTeamInCorridor( my_pos,
				      my_pos + radAng2Vector2d( 15.0, kickangle[i] ),
				      5.0, PI_2 ) > 0 )
	dir[i].downgrade();
      // ...then try it with a longer and broader one
      if( world->otherTeamInCorridor( my_pos,
				      my_pos + radAng2Vector2d( 17.0, kickangle[i] ), 
				      8.0, PI_2 ) > 0 )
	dir[i].downgrade();
      // ...finally test the longest and broadest corridor
      if( world->otherTeamInCorridor( my_pos,
				      my_pos + radAng2Vector2d( 20.0, kickangle[i] ), 
				      10.0, PI_2 ) > 0 )
	dir[i].downgrade();
      // check if target is in field
      if( !inField( my_pos + radAng2Vector2d( 20.0, kickangle[i] ) ) )
	dir[i].set_BAD();
    }

  // search for best grade
  int bestDir = ( my_area == AREA_LEFT ? 6 : my_area == AREA_RIGHT ? 2 : 4 );
  for( int i = 0; i < 9; i++ )
    if( dir[i].is_better_than( dir[bestDir] ) )
      bestDir = i;

  DRAWTRACE(Comm::sendMCmd(MDLine(my_pos,my_pos + radAng2Vector2d(15.0,kickangle[bestDir]))));

  PASSTRACE( "PASS: getEmergencyPass -> emergency kick into direction " << kickangle[bestDir] );

  // determine kick vector
  return radAng2Vector2d( BALL_SPEED_MAX, kickangle[bestDir] );
}
