//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifdef MRBDEBUG
//#define USEPOSFTRACE
//#define USEAREATRACE
#endif


#ifndef USEAREATRACE
#define AREATRC(output)
#else
#warning PositionEval AREATRACE aktiviert!!
#define FUNSSELMACROS
#include "drawMacros.h"
#define AREATRC(output) Comm::sendMCmd(output)
#endif

#ifndef USEPOSFTRACE
#define POSFTRACE(output)
#else
#warning POSFTRACE aktiviert!!
#define POSFTRACE(output) Comm::sendMCmd(output)
#define FUNSSELMACROS
#include "drawMacros.h"
#define USEPOSLOG
#endif

#ifndef USEPOSLOG
#define POSLOG(output)
#else
#warning POSLOG aktiviert !!
#define POSLOG(output) { ostringstream ost;\
	ost << "POSLOG: " << output << (char)0;\
 Comm::sendMCmd(MLog(ost.str()));}
#endif


#include "PositionEval.h"
#include "PositionFunctions.h"
#include "Config.h"


using namespace std;

#define sign(x) ((x)>0?1:-1)

const double MAX_MINX_STEP =.25;//.3//.17;
//max speed of our defense when going to other goal
const double START_SLOW    =-FIELD_LENGTH/2+1.2*PENALTY_AREA_LENGTH;
//after this point the defenders are slowed down

const double BALL_MIN_POSITION= 30;//35;
//how far our defenders try to stand behind the ball (relative)

const double MAX_AREA=FIELD_LENGTH/3.;
const double MIN_AREA=FIELD_LENGTH/4.;
const double DIST_TO_OFFSIDE=2;

const double MAX_OFFENDER_X_POS=FIELD_LENGTH/2-PENALTY_AREA_LENGTH/2;
const double MIN_DEFENDER_X_POS=-FIELD_LENGTH/2+0.5*PENALTY_AREA_LENGTH;

const double DIST_TO_FIELD_BORDER = 5.0;

const double BALL_FACTOR_OFF_INC = 1 / 20.0;
const double BALL_FACTOR_OFF_DEC = 1 / 30.0;
const double BALL_FACTOR_DEF_INC = 1 / 30.0;
const double BALL_FACTOR_DEF_DEC = 1 / 10.0;

PositionEval::PositionEval(const GoalShotEval* _goalshoteval, 
			   const PassEval* _passeval)
  : goalshoteval(_goalshoteval), passeval(_passeval)
{
  int i;
  
  readAgent_conf();

  for (i=1;i<10;i++)
    RealPositions[i].Pos=Vector2d(0,0);

  minX=0;
  maxX=0;
  ballFactorDef=0;
  ballFactorOff=0;

  defNo=Config::getValueAsInt("StartFormationDef");
  offNo=Config::getValueAsInt("StartFormationOff");
}

PositionEval::~PositionEval()
{
}

Grade PositionEval::ourEval(const Worldmodel* , int ) const
{
  returnGRADE(Grade());
}

	
Vector2d PositionEval::GetPlayerPosition(const Worldmodel* world, int PlayerNo) const 
{
  if ((PlayerNo>1) && (PlayerNo<12) && (world->ourTeam[PlayerNo].tacticalNo!=UNKNOWN) ) 
    return(RealPositions[world->ourTeam[PlayerNo].tacticalNo-2].Pos);
  else return(Vector2d(0,0));
};

double PositionEval::GetMinX() const
{
  return(minX);
};
  
double PositionEval::GetMaxX() const
{
  return(maxX);
};

void PositionEval::CalMinMaxX(const Worldmodel *world) 
{
  if (time<world->time)
    oldminX=minX;
  //save minX

  double ballminPos=BALL_MIN_POSITION*(1-Step(MIN_DEFENDER_X_POS,-FIELD_LENGTH/2
					      +2 *PENALTY_AREA_LENGTH,
					      world->ball.pos.getX()));
  //Calculate the max. and min. positions of our players
  maxX=world->ourOffsideLine(MAXAGE);
  maxX-=DIST_TO_OFFSIDE;
  POSFTRACE(VLINE(maxX));

  minX=min(world->ball.pos.getX(),maxX * ballminPos / 100.0 );

  //We don't want to occupie a too small room 
  if (maxX-minX<MIN_AREA * (double) (100-ballminPos) / 100.0 )
    minX=maxX-MIN_AREA * (double) (100-ballminPos) / 100.0;
  
  //and not a too big room
  if (maxX-minX>MAX_AREA * (double) (100-ballminPos) / 100.0 )
     {
       if (world->ball.pos.getX() < (minX+maxX)/2)
	 maxX=minX+MAX_AREA * (double) (100-ballminPos) / 100.0;
    };

  minX=min(maxX-100/(100-ballminPos)*(maxX-minX),0.);
  minX=max(minX,MIN_DEFENDER_X_POS);
  
  //defenders shouldn't be in the others halfside
  if (minX>0)
    minX=0;
  //its not very useful if the offenders are too far in the others penalty
  //area
  if (maxX>MAX_OFFENDER_X_POS)
    maxX=MAX_OFFENDER_X_POS;

  if ((minX>START_SLOW) && (oldminX+MAX_MINX_STEP<minX))
    {
      minX=oldminX+MAX_MINX_STEP;
      POSLOG("oldminX : "<< oldminX<<" minX : " <<minX<<" MAX_MINX_STEP: "<<
	     MAX_MINX_STEP);
    };

  //take care of the standartsituations
  switch (world->playMode){
    /*
  case PLAY_MODE_CORNER_KICK_THEM:
    minX=FIELD_LENGTH/2;
    break;*/
  case PLAY_MODE_GOAL_THEM:
  case PLAY_MODE_GOAL_US:
    minX=-MIN_AREA;
    oldminX=-MIN_AREA;
    break;
  case PLAY_MODE_CORNER_KICK_US:
    maxX=FIELD_LENGTH/2+4-PENALTY_AREA_LENGTH;
    break;
  case PLAY_MODE_FREE_KICK_THEM:
    if (world->ball.inOtherPenaltyArea())
      {
	maxX=FIELD_LENGTH/2-PENALTY_AREA_LENGTH-6;
	if (maxX-minX<MIN_AREA)
	  {
	    oldminX=minX=maxX-MIN_AREA;
	  };
      };
    break;
  case PLAY_MODE_FREE_KICK_US:
    if (world->ball.inOurPenaltyArea())  
      {
	oldminX=minX=-(FIELD_LENGTH/2-PENALTY_AREA_LENGTH)+4;
	if (maxX-minX<MIN_AREA)
	  maxX=minX+MIN_AREA;
      };
    break;
  case PLAY_MODE_GOAL_KICK_US:
    {
      oldminX=minX=PENALTY_AREA_LENGTH-FIELD_LENGTH/2+4;
      if (maxX-minX<MIN_AREA)
	maxX=minX+MIN_AREA;
      break;
    };
  case PLAY_MODE_GOAL_KICK_THEM:
    {
      maxX=FIELD_LENGTH/2-PENALTY_AREA_LENGTH-6;
      if (maxX-minX<FIELD_LENGTH/4.)
	  oldminX=minX=maxX-FIELD_LENGTH/4.; 
    };
  default:
    break;
  };

  defensive=Step(minX+ballminPos*(maxX-minX)/100.0,maxX,world->ball.pos.getX());
  offensive=1-defensive;
  //offensive=Triangle(1.,minX-maxX,world->ball.pos.getX()-maxX);
  //defensive=Triangle(1.,maxX-minX,world->ball.pos.getX()-minX);

  //if (maxX<world->ball.pos.getX()) 
  //  offensive=1;
  //if (minX+ballminPos*(maxX-minX)/100.>world->ball.pos.getX())
  //  defensive=1;//ball behind defenders
  
  POSFTRACE(VLINE(minX));
  POSFTRACE(VLINE(maxX)); 
  
  POSFTRACE(VLINE(minX+ballminPos*(maxX-minX)/100.0));
  POSLOG("MaxX: " << maxX << " MinX: " << minX << " oMinX: " << oldminX  
	 << " off.: " << offensive << 
	 " dev.: " << defensive << " ballminPos: " << ballminPos);
  POSLOG("Ballpos " << world->ball.pos); 

  time=world->time;
};

void PositionEval::CalYPos(Vector2d BallPos)
{
  //4 positions	which describe the middle of our formation

  Vector2d Def( - FIELD_LENGTH / 2.0, 0.0);
  Vector2d DefMid( -FIELD_LENGTH / 6.0, 0.0);
  Vector2d OffMid( FIELD_LENGTH / 4.0, 0);
  Vector2d Off( + FIELD_LENGTH / 2.0, 0);


  //They are shifted towards the ball y pos according to their x distance
  const double MaxShiftDef=PENALTY_AREA_WIDTH / 4.0;
  const double MaxShiftDefMid=FIELD_WIDTH / 6.0;
  const double MaxShiftOffMid=FIELD_WIDTH / 8.0;
  const double MaxShiftOff=PENALTY_AREA_WIDTH / 6.0;

  //define the width of our formations at these points (actually it's the half of the width)
  const double WidthDef    = PENALTY_AREA_WIDTH / 4.0;
  const double WidthDefMidForDef = FIELD_WIDTH * 1.3 / 6.0;
  const double WidthOffMidForDef = FIELD_WIDTH * 1.5 / 6.0;
  const double WidthDefMidForOff = FIELD_WIDTH * 1.0 / 3.0;
  const double WidthOffMidForOff = FIELD_WIDTH * 1.0 / 3.0;
  const double WidthOff    = PENALTY_AREA_WIDTH / 3.2;

  Def.setY(Triangle(MaxShiftDef,FIELD_LENGTH / 1.2, 
		    Def.getX()-BallPos.getX())
	   *(1-Triangle(1,FIELD_WIDTH / 2.0,BallPos.getY()))
	   *sign(BallPos.getY()));
  DefMid.setY(Triangle(MaxShiftDefMid,FIELD_LENGTH / 1.2, 
		       DefMid.getX()-BallPos.getX())
	      *(1-Triangle(1,FIELD_WIDTH / 2.0,BallPos.getY()))
	      *sign(BallPos.getY()));
  OffMid.setY(Triangle(MaxShiftOffMid,FIELD_LENGTH / 1.2, 
		       OffMid.getX()-BallPos.getX())
	      *(1-Triangle(1,FIELD_WIDTH / 2.0,BallPos.getY()))
	      *sign(BallPos.getY()));
  Off.setY(Triangle(MaxShiftOff,FIELD_LENGTH / 1.2, 
		    Off.getX()-BallPos.getX())
	   *(1-Triangle(1,FIELD_WIDTH / 2.0,BallPos.getY()))
	   *sign(BallPos.getY()));

  POSFTRACE(LINE(Def,DefMid));
  POSFTRACE(LINE(DefMid,OffMid));
  POSFTRACE(LINE(OffMid,Off));
    
  // now we have the center in y direction. We only need to calculate the
  // width of our formation


#ifdef USEPOSFTRACE
  Vector2d DefW(0,WidthDef);
  Vector2d DefMidWD(0,WidthDefMidForDef);
  Vector2d OffMidWD(0,WidthOffMidForDef);
  Vector2d DefMidWO(0,WidthDefMidForOff);
  Vector2d OffMidWO(0,WidthOffMidForOff);
  Vector2d OffW(0,WidthOff);
  POSFTRACE(LINE(Def+DefW,DefMid+DefMidWD));
  POSFTRACE(LINE(Def-DefW,DefMid-DefMidWD));
  POSFTRACE(LINE(DefMid+DefMidWD,OffMid+OffMidWD));
  POSFTRACE(LINE(DefMid-DefMidWD,OffMid-OffMidWD));
  POSFTRACE(LINE(DefMid+DefMidWO,OffMid+OffMidWO));
  POSFTRACE(LINE(DefMid-DefMidWO,OffMid-OffMidWO));
  POSFTRACE(LINE(OffMid+OffMidWO,Off+OffW));
  POSFTRACE(LINE(OffMid-OffMidWO,Off-OffW));
#endif  


  double WidthDefMid;
  double WidthOffMid;

  for (int i=0;i<10;i++)
    {
      PlayerPos &Player=RealPositions[i];

      if (Player.tact==PlayerPos::DEFENSE || Player.tact==PlayerPos::DEFMIDFIELD )
	{
	  WidthDefMid=WidthDefMidForDef;
	  WidthOffMid=WidthOffMidForDef;
	}
      else
	{
	  WidthDefMid=WidthDefMidForOff;
	  WidthOffMid=WidthOffMidForOff;
	}	   

      if (Player.Pos.getX()<DefMid.getX())
	Player.Pos.setY(Player.Pos.getY() //the relative pos 
			*(WidthDefMid+(WidthDef-WidthDefMid)*
			  Step(Def.getX(),DefMid.getX(),Player.Pos.getX()))
			// times the width
			+DefMid.getY()+(Def.getY()-DefMid.getY())*
			Step(Def.getX(),DefMid.getX(),Player.Pos.getX()));
  // plus the center of our formation at certain x pos
      else if (Player.Pos.getX()<OffMid.getX())    
	Player.Pos.setY(Player.Pos.getY() //the relative pos 
			*(WidthOffMid+(WidthDefMid-WidthOffMid)*
			  Step(DefMid.getX(),OffMid.getX(),Player.Pos.getX()))
			// times the width
			+OffMid.getY()+(DefMid.getY()-OffMid.getY())*
			Step(DefMid.getX(),OffMid.getX(),Player.Pos.getX()));
      else if (Player.Pos.getX()<Off.getX())    
	Player.Pos.setY(Player.Pos.getY() //the relative pos 
			*(WidthOff+(WidthOffMid-WidthOff)*
			  Step(OffMid.getX(),Off.getX(),Player.Pos.getX()))
			// times the width
			+Off.getY()+(OffMid.getY()-Off.getY())*
			Step(OffMid.getX(),Off.getX(),Player.Pos.getX()));
      
      // check if the position is inside of the field area
      if ( fabs(Player.Pos.getY()) > FIELD_WIDTH / 2.0 - DIST_TO_FIELD_BORDER )
	Player.Pos.setY(sign( Player.Pos.getY() ) * ( FIELD_WIDTH / 2.0 - DIST_TO_FIELD_BORDER ) );
    }

};

void PositionEval::CalXPos(Vector2d BallPos,double OffsideLine) 
{
  for (int i=0;i<10;i++)
    {
      PlayerPos &Player=RealPositions[i];

      //auf minX und maxX renormieren
      Player.Pos.setX(Player.Pos.getX()*(maxX-minX)/(100)+minX);
      
      //die Area anpassen
      Player.Area.setX(Player.Area.getX()*(maxX-minX)/100);
      
      /*
      //offensive players may go to the offside line
      if ( Player.tact==PlayerPos::OFFENSE )
	Player.Max.setX(OffsideLine);
      else
	Player.Max.setX(Player.Max.getX()*(maxX-minX)/100+minX);
      */
    };

};


void PositionEval::CalPlayerPositions(const Worldmodel *world) 

{
  int j;

  // update ball factors
  if ( offensive < 0.5 )
    {
      ballFactorOff-=BALL_FACTOR_OFF_DEC;
      ballFactorDef-=BALL_FACTOR_DEF_DEC;
    }
  else
    {
      ballFactorOff+=BALL_FACTOR_OFF_INC;
      ballFactorDef+=BALL_FACTOR_DEF_INC;
    }
  // check boundaries
  if (ballFactorOff > 1.0)
    ballFactorOff = 1.0;
  if (ballFactorDef > 1.0)
    ballFactorDef = 1.0;
  if (ballFactorOff < 0.0)
    ballFactorOff = 0.0;
  if (ballFactorDef < 0.0)
    ballFactorDef = 0.0;

  // Set position array with relative positions
  for (j=0;j<5;j++)
    {
      //Defense
      RealPositions[j]=DefFormations[defNo][j];
      RealPositions[j].Pos=DefFormations[defNo][j].Pos+ballFactorDef*DefFormations[defNo][j].BMod;
      RealPositions[j].Pos=RealPositions[j].Pos+
	Step(MIN_DEFENDER_X_POS,-FIELD_LENGTH/2+2 *PENALTY_AREA_LENGTH, world->ball.pos.getX())
	* DefFormations[defNo][j].FMod;
      //Offense
      RealPositions[j+5]=OffFormations[offNo][j];
      RealPositions[j+5].Pos=OffFormations[offNo][j].Pos+ballFactorOff*OffFormations[offNo][j].BMod;
      RealPositions[j+5].Pos=RealPositions[j+5].Pos+
	Step(MIN_DEFENDER_X_POS,-FIELD_LENGTH/2+2 *PENALTY_AREA_LENGTH, world->ball.pos.getX())
	* OffFormations[offNo][j].FMod;
	
    }

  // Calculate the positions on the field
  CalXPos(world->ball.pos,world->ourOffsideLine(MAXAGE)-DIST_TO_OFFSIDE);
  CalYPos(world->ball.pos);

  for (j=0;j<10;j++)
    {
      PlayerPos &Player=RealPositions[j];
      //Abwehr fr Ecken optimieren
      double ballx;
      double bally;
      ballx=Player.ballforceX;
      bally=Player.ballforceY;  
      double BallDistX=world->ball.pos.getX()-Player.Pos.getX();
      if (BallDistX<0)
	Player.Pos.setX(Player.Pos.getX()+5*sign(BallDistX)
			*Triangle(ballx,30,BallDistX)
			*Triangle(bally,FIELD_WIDTH / 2.0,world->ball.pos.getY()-Player.Pos.getY()));
    };

#ifdef POSFTRACE
  for (j=0;j<10;j++)
    POSFTRACE(CIRCLE(RealPositions[j].Pos,3));
  POSFTRACE(LINE(GetPlayerPosition(world,world->my.number),world->my.pos));
#endif


#ifdef USEAREATRACE
  PlayerPos RealPos=
    RealPositions[world->ourTeam[world->my.number].tacticalNo-2];
  Vector2d Min;
  Vector2d Max;
  GetPlayerArea(world,world->my.number,Min,Max);
  Vector2d tmp1=Min;
  Vector2d tmp2=Vector2d(Max.getX(),Min.getY());
  AREATRC(LINE(tmp1,tmp2));
  tmp1=Max;
  AREATRC(LINE(tmp1,tmp2));
  tmp2=Vector2d(Min.getX(),Max.getY());
  AREATRC(LINE(tmp1,tmp2));
  tmp1=Min;
  AREATRC(LINE(tmp1,tmp2));
#endif
};

void PositionEval::GetPlayerArea(const Worldmodel* world,int PlayerNo, 
				 Vector2d &Min, Vector2d &Max) const
{
  if ( (PlayerNo>1) && (PlayerNo<12) && 
       (world->ourTeam[PlayerNo].tacticalNo!=UNKNOWN) )
    {
      Min=RealPositions[world->ourTeam[PlayerNo].tacticalNo-2].Pos
	- RealPositions[world->ourTeam[PlayerNo].tacticalNo-2].Area;
      Max=RealPositions[world->ourTeam[PlayerNo].tacticalNo-2].Pos
	+ RealPositions[world->ourTeam[PlayerNo].tacticalNo-2].Area;

      // the offense may go to the offsideline
      if ( RealPositions[world->ourTeam[PlayerNo].tacticalNo-2].tact
	   ==PlayerPos::OFFENSE )
	Max.setX( world->ourOffsideLine(MAXAGE) );

      // check if the area is inside of the field
      if ( fabs( Max.getY() ) > FIELD_WIDTH / 2.0 - DIST_TO_FIELD_BORDER )
	Max.setY(sign( Max.getY() ) * ( FIELD_WIDTH / 2.0 - DIST_TO_FIELD_BORDER ) );
      if ( fabs( Min.getY() ) > FIELD_WIDTH / 2.0 - DIST_TO_FIELD_BORDER )
	Min.setY(sign( Min.getY() ) * ( FIELD_WIDTH / 2.0 - DIST_TO_FIELD_BORDER ) );
    }
  else 
    {
      Min=Vector2d(0,0);
      Max=Vector2d(0,0);
    }
};

bool PositionEval::IsInPlayerArea(const Worldmodel* world,int PlayerNo, Vector2d Pos) const
{
  Vector2d pMin;
  Vector2d pMax;
  GetPlayerArea(world,PlayerNo,pMin,pMax);
  return ((Pos.getX()<pMax.getX()) && (Pos.getX()>pMin.getX()) &&
	  (Pos.getY()<pMax.getY()) && (Pos.getY()>pMin.getY()));
};

int PositionEval::GetPlayerWatchNo(const Worldmodel* world,int PlayerNo, int WatchNo) const
{
  if ((PlayerNo>2) && (PlayerNo<11) && (WatchNo>=0) && (WatchNo<=2) && 
      (world->ourTeam[PlayerNo].tacticalNo!=UNKNOWN) )
    {
      int tactNo=world->ourTeam[PlayerNo].tacticalNo-2;
      tactNo= ( tactNo>4 ? OffFormations[offNo][tactNo-5].Watches[WatchNo]:
		DefFormations[defNo][tactNo].Watches[WatchNo] );
      
      if (tactNo==-1)
	return -1;

      for (int i=1;i<12;i++)
	if (world->ourTeam[i].tacticalNo==tactNo)
	  {
	    return i;
	  };
    }
  return UNKNOWN;
}

double PositionEval::GetOffensive() const
{
  return offensive;
};

void PositionEval::readAgent_conf(){
  char valueString[100];
  double x,y;
  for (int k=0; k<=2; k++)
  for (int i=0; i<5; i++){
    //defense
    sprintf(valueString,"Form%i_def_PosX",k);
    x=Config::getValueAsDoubleFromList(valueString,i);
    sprintf(valueString,"Form%i_def_PosY",k);
    y=Config::getValueAsDoubleFromList(valueString,i);//*FIELD_WIDTH;
    DefFormations[k][i].Pos = Vector2d(x,y);

    sprintf(valueString,"Form%i_def_tact",k);
    (int) DefFormations[k][i].tact = Config::getValueAsIntFromList(valueString,i);
    
    sprintf(valueString,"Form%i_def_BforceX",k);
    DefFormations[k][i].ballforceX = 
      Config::getValueAsDoubleFromList(valueString,i);

    sprintf(valueString,"Form%i_def_BforceY",k);
    DefFormations[k][i].ballforceY = 
      Config::getValueAsDoubleFromList(valueString,i);
    
    sprintf(valueString,"Form%i_def_AreaX",k);
    x=Config::getValueAsDoubleFromList(valueString,i);
    sprintf(valueString,"Form%i_def_AreaY",k);
    y=Config::getValueAsDoubleFromList(valueString,i)*FIELD_WIDTH;
    DefFormations[k][i].Area = Vector2d(x,y);
    
    sprintf(valueString,"Form%i_def_BModX",k);
    x=Config::getValueAsDoubleFromList(valueString,i);
    sprintf(valueString,"Form%i_def_BModY",k);
    y=Config::getValueAsDoubleFromList(valueString,i);//*FIELD_WIDTH;
    DefFormations[k][i].BMod = Vector2d(x,y);
        
    sprintf(valueString,"Form%i_def_FModX",k);
    x=Config::getValueAsDoubleFromList(valueString,i);
    sprintf(valueString,"Form%i_def_FModY",k);
    y=Config::getValueAsDoubleFromList(valueString,i);//*FIELD_WIDTH;
    DefFormations[k][i].FMod = Vector2d(x,y);
        
    for (int j=0;j<3;j++){
      sprintf(valueString,"Form%i_def_Watch%d",k,j+1);
      DefFormations[k][i].Watches[j] = 
	Config::getValueAsIntFromList(valueString,i);
      } 

    //offense
    sprintf(valueString,"Form%i_off_PosX",k);
    x=Config::getValueAsDoubleFromList(valueString,i);
    sprintf(valueString,"Form%i_off_PosY",k);
    y=Config::getValueAsDoubleFromList(valueString,i);//*FIELD_WIDTH;
    OffFormations[k][i].Pos = Vector2d(x,y);

    sprintf(valueString,"Form%i_off_tact",k);
    (int) OffFormations[k][i].tact = Config::getValueAsIntFromList(valueString,i);
     
    sprintf(valueString,"Form%i_off_BforceX",k);
    OffFormations[k][i].ballforceX = 
      Config::getValueAsDoubleFromList(valueString,i);
    sprintf(valueString,"Form%i_off_BforceY",k);
    OffFormations[k][i].ballforceY = 
      Config::getValueAsDoubleFromList(valueString,i);
    
    sprintf(valueString,"Form%i_off_AreaX",k);
    x=Config::getValueAsDoubleFromList(valueString,i);
    sprintf(valueString,"Form%i_off_AreaY",k);
    y=Config::getValueAsDoubleFromList(valueString,i)*FIELD_WIDTH;
    OffFormations[k][i].Area = Vector2d(x,y);
    
    sprintf(valueString,"Form%i_off_BModX",k);
    x=Config::getValueAsDoubleFromList(valueString,i);
    sprintf(valueString,"Form%i_off_BModY",k);
    y=Config::getValueAsDoubleFromList(valueString,i);//*FIELD_WIDTH;
    OffFormations[k][i].BMod = Vector2d(x,y);
        
    sprintf(valueString,"Form%i_off_FModX",k);
    x=Config::getValueAsDoubleFromList(valueString,i);
    sprintf(valueString,"Form%i_off_FModY",k);
    y=Config::getValueAsDoubleFromList(valueString,i);//*FIELD_WIDTH;
    OffFormations[k][i].FMod = Vector2d(x,y);
     
    for (int j=0;j<3;j++){
      sprintf(valueString,"Form%i_off_Watch%d",k,j+1);
      OffFormations[k][i].Watches[j] = 
	Config::getValueAsIntFromList(valueString,i);
      } 
  
  }
}

