//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

Grade PositionModule::goalieEvaluate()
{
  Grade g;
  g.set_GOOD();

  priority = 1;
  if( world->ball.pos.getX() > -FIELD_LENGTH / 2.0 + 1.5 * PENALTY_AREA_LENGTH )
    {
      priority++;
      if( world->ball.pos.getX() > -FIELD_LENGTH / 2.0 + 2.5 * PENALTY_AREA_LENGTH )
	{
	  priority++;
	  if( world->ball.pos.getX() > -FIELD_LENGTH / 2.0 + 3.5 * PENALTY_AREA_LENGTH )
	    priority++;
	}
    }

  if (!me.inOurPenaltyArea())
    priority = 0;

  return g;
}


bool PositionModule::GoalieAct()
{
  GOALIETRACE("GoalieAct: going to " << posGoal);

  double useableStam = calcUseableStamina();

  goToPrecision = CATCHABLE_AREA_W;

  // position where first dangerous opponent will stop the ball
  Vector2d ballStopPos = world->ball.pos;

  if( tact->firstDangerousOpponent != UNKNOWN )
    ballStopPos = tact->otherTeamReachBallPos[tact->firstDangerousOpponent];
  else
    ballStopPos = world->ball.pos + ( 1.0 / ( 1.0 - BALL_DECAY ) ) * world->ball.v;
 
  posGoal.setX( -0.5 * FIELD_LENGTH + CATCHABLE_AREA_W );

  posGoal.setY( ballStopPos.getY() * GOAL_AREA_LENGTH * 0.5 / ( 0.5 * FIELD_LENGTH - fabs( ballStopPos.getX() ) ) );
  if( posGoal.getY() > 0.5 * GOAL_WIDTH - CATCHABLE_AREA_L )
    posGoal.setY( 0.5 * GOAL_WIDTH - CATCHABLE_AREA_L );
  if( posGoal.getY() < -0.5 * GOAL_WIDTH + CATCHABLE_AREA_L )
    posGoal.setY( -0.5 * GOAL_WIDTH + CATCHABLE_AREA_L );
  if( ballStopPos.getX() > 0.0 )
    posGoal.setY( 0.0 );

  AdvSkills::TurtleCommand tc 
    = skills->goToPosition( posGoal, goToPrecision, useableStam,
			    true, ( world->ball.pos + world->ball.v - me.pos ).getAngle() );
  // if backwards, double stamina
  if( tc.state == AdvSkills::TGOINGBACKWARDS )
    tc = skills->goToPosition( posGoal, goToPrecision, 
			       min( 2 * useableStam, availableStamina ),
			       true, ( world->ball.pos + world->ball.v - me.pos ).getAngle() );

  // if I am at my desired position turn to ball
  if( tc.state == AdvSkills::TFINISHED )
    {
      AdvSkills::TurnCommand tuc = skills->turnTo( world->ball );

      if( ballStopPos.getY() > world->my.pos.getY() + 5.0 )
	tuc = skills->turnTo( world->my.pos + Vector2d( 0.0, 10.0 ) );
      else if( ballStopPos.getY() < world->my.pos.getY() - 5.0 )
	tuc = skills->turnTo( world->my.pos + Vector2d( 0.0, -10.0 ) );

      if( tuc.state == AdvSkills::RFINISHED )
	skills->noAction();
      else
	skills->exec( tuc );
    }
  else
    skills->exec( tc );

  return true;
}
