//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

 
// [C] constant dependend code
// [C]* multiple constants
#include "PositionModule.h"
#include "constants.h"
#include "minmax.h"
    
 
#include "PositionFunctions.C"

using namespace std;


#ifdef MRBDEBUG

// #define POSACTIONTRACE

#endif

#ifdef POSACTIONTRACE
#warning: POSM debug trace ACTIONTRACE is activated!
#define PACTIONTR(output) Comm::sendMCmd(MLog(output))
#else
#define PACTIONTR(output)
#endif 


using namespace std;

const double max_useable_Stamina[4]={500,100,70,30};//max Stamina per cycle
const double save_Stamina[4]={0,.2,.3,.5}; 
// factor how much stamina is saved for higher priorities

const  double FREEPOSAREA_LENGHT=FIELD_LENGTH/8.;
const  double FREEPOSAREA_WIDTH=FIELD_WIDTH/3.;
// the distance from the ball, in which the players goto free positions during
// standartsit. 

#define sign(x) ((x)>0?1:-1)

PositionModule::PositionModule(bool _isGoalie, 
			       Worldmodel * _world,
			       TacticalModule * _tact,
			       AdvSkills * _skills,
			       const PassEval *_passEval,  
			       const GoalShotEval *_goalShotEval,
			       PositionEval *_positionEval) :
  world(_world),
  tact(_tact),
  skills(_skills),
  posGoal(0,0),
  priority(3), 
  oldPosGoal(0,0),
  middleOfGoal(-FIELD_LENGTH/2, 0), 
  nearestFriendNumber(UNKNOWN), 
  nearestOther(UNKNOWN),
  passEval(_passEval), 
  goalShotEval(_goalShotEval),
  positionEval(_positionEval),
  LastBestPos(1000,1000),
  lastActive(0)
{
  isGoalie = _isGoalie;
  leftCorner  = Vector2d(-FIELD_LENGTH/2,-GOAL_WIDTH/2);
  rightCorner = Vector2d(-FIELD_LENGTH/2,GOAL_WIDTH/2);   
  OppDef1 = 0;
  OppDef2 = 0;
}

PositionModule::~PositionModule()
{
}

bool PositionModule::feedbackAudio(AudioMsg &msg)
{
  if (!msg.getState() || !msg.isCoachMsg()) {
    PACTIONTR("got bad msg");
    return false;
  }

  Int3B oppdefs;
  /*
  bool ok(false);

  msg >> ok;

  if (!ok) {
    PACTIONTR("got wrong msg");
    return false;
  }
  */
  msg >> oppdefs;
  OppDef2 = oppdefs % 12;
  OppDef1 = oppdefs / 12;
  

#ifdef POSACTIONTRACE
  ostrstream tmp;
  tmp << "got message from coach opp1: " << OppDef1 << " opp2: " << OppDef2 << char(0) ;
  { char * str= tmp.str();
  PACTIONTR(str);
  delete str; }
#endif
  //cerr << "got message from coach opp1: " << OppDef1 << " opp2: " << OppDef2 << endl ;

  return true;
};  

bool PositionModule::feedbackWorld()
{
  // positionEval needs the worldmodel as well
  positionEval->CalMinMaxX(world);
  positionEval->CalPlayerPositions(world);

  oldPosGoal = posGoal;
   
  // SET OFFSIDE VARIABLES
  ourOffsideLine = world->ourOffsideLine(MAXAGE);
  otherOffsideLine = world->otherOffsideLine(MAXAGE);
 //   Comm::sendMCmd(LINE(Vector2d(ourOffsideLine,-FIELD_WIDTH/2),Vector2d(ourOffsideLine,FIELD_WIDTH/2))); 
//    Comm::sendMCmd(LINE(Vector2d(otherOffsideLine,-FIELD_WIDTH/2),Vector2d(otherOffsideLine,FIELD_WIDTH/2))); 

  //calculate max. useable stamina
  availableStamina = max (0.0, world->my.stamina - 
			  EFFORT_DEC_THR * STAMINA_MAX - 2 );


  // shorten the code
  me = world->my;
  nearestFriendNumber = world->ourTeamClosestTo(me.pos,20,2);

  myTactNo = world->my.tacticalNo;
  if(myTactNo == UNKNOWN) myTactNo = me.number;
  
  nearestOther = world->otherTeamClosestTo(me.pos);
#ifdef POSACTIONTRACE
  ostrstream temp; 
  switch(tact->ourAction.state){
  case TA_REACT:
    temp <<"posm: ourAction.state TA_REACT performer: ";
    break;
  case TA_PASS:
    temp <<"posm: ourAction.state TA_PASS performer: ";
    break;
  case TA_CONTROL:
    temp << "posm: ourAction.state TA_CONTROL performer: ";
    break;
  default:
    temp <<"posm: ourAction.state= was wei ich performer: ";
  };
  temp << tact->ourAction.performer ;
  temp << " age: " << tact->ourAction.age << " quality: ";
  switch (tact->ourAction.quality){
  case TQ_GUESSED:
    temp << "guessed";
    break;
  case TQ_REPORTED:
    temp << "reported";
    break;
  case TQ_AUDIAL_REPORTED:
    temp << "audial reported";
    break;
  case TQ_AUDIAL:
    temp << "audial ";
    break;
  case TQ_UNKNOWN:
    temp << "unknown";
    break;
  default:
    temp << "other: " << tact->ourAction.quality;
  };
  temp << " BallCon: " ;
  switch (tact->ourBallControl){
  case BC_LOST:
    temp <<"lost";
    break;
  case BC_SAFE:
    temp <<"save";
    break;
  case BC_DANGER:
    temp <<"danger";
    break;
  case BC_UNKNOWN:
    temp <<"unknown";
    break;
  default:
    temp <<"other";
  };

  temp << char(0);
  { char * str= temp.str();
  PACTIONTR(MLog(str));
  delete str; }

  ostrstream tmp;
  tmp << "posm: Resp. players: " << tact->firstResponsiblePlayer << "(" 
      << tact->ourTeamReachBallTime[tact->firstResponsiblePlayer] << "), " 
      << tact->secondResponsiblePlayer << "(" 
      << tact->ourTeamReachBallTime[tact->secondResponsiblePlayer] << "), " 
      << tact->thirdResponsiblePlayer 
      << " dang. opp: " << tact->firstDangerousOpponent << "(" 
      << tact->otherTeamReachBallTime[tact->firstDangerousOpponent] << "), " 
      << tact->secondDangerousOpponent << "("  
      << tact->otherTeamReachBallTime[tact->secondDangerousOpponent] << "), "
      << tact->thirdDangerousOpponent  
      << " Oppdefs: " << OppDef1 << ", " << OppDef2 
      << " OurG: " << world->ourGoalieNo << " OtherG: " << world->otherGoalieNo
      << char(0);
  { char * str = tmp.str();
  PACTIONTR(MLog(str));
  delete str; }
#endif
  return true;
} 


Grade PositionModule::evaluate()
{

  Grade g;
  g.set_GOOD();
  coveredOpponent = -1;
  lookObject=NULL;
  lookPos.setXY(0,0);
  didIAct=0; // for turnHeadEval
  action=DEFAULT;
  
  // standard position is my position
  posGoal = me.pos;
  
  // Now everything is checked and fine, so we can start to think.
  bool AttackOpp = false;
  if (  
      ( (tact->ourAction.state==TA_REACT) || (tact->ourBallControl == BC_LOST) ||
	(tact->otherTeamReachBallTime[tact->firstDangerousOpponent] == 0 ) ) &&
      ( me.number == tact->firstResponsiblePlayer ||
	me.number == tact->secondResponsiblePlayer ||
	me.number == tact->thirdResponsiblePlayer ||
	me.number == 
	world->ourTeamClosestTo(tact->otherTeamReachBallPos[tact->firstDangerousOpponent]) ||  
	me.number == 
	world->ourTeamClosestTo(tact->otherTeamReachBallPos[tact->firstDangerousOpponent],1)  ) )
    {
      AttackOpp = EvalPlayerAttack();
      g.set_PERFECT();
    }
  
  // when did I tackle the last time? May I move?
  if (world->my.cyclesSinceTackle<=TACKLE_CYCLES)
    {
      // I am not allowed to move
#ifdef POSACTIONTRACE
      {
	ostrstream t;
	t << " posm: I can't move due to a tackle command "
	  << world->my.cyclesSinceTackle<<" cycles ago" << endl;
	char *str=t.str();
	PACTIONTR(str);
	delete str;
      }
#endif 

      action=TACKLE;
      return(Grade::PERFECT);
    }

  // my ball ?
  if ( world->my.ballKickable( world->ball ) ){ 
    action=NO_ACTION;
    POSTRACE("I can kick ball");
    posGoal = world->ball.pos;
    priority = 2;
    goToPrecision = KICKABLE_MARGIN/2;
    action=NO_ACTION;
    g.set_BAD();
    EvalGrade=g;
    returnGRADE(g);
  }

  if (!isGoalie)
    {
      CalcStdPriority();

      //g.set_ABSOLUTE();
      posGoal=positionEval->GetPlayerPosition(world,me.number);

      goToPrecision = 2;
      
      if ( AttackOpp &&

	  /*( (tact->ourAction.state==TA_REACT) || 
	     (tact->ourBallControl == BC_LOST) ) && 
	  (tact->firstResponsiblePlayer!=UNKNOWN) &&
	  // (tact->secondResponsiblePlayer!=UNKNOWN) &&
	  (//attack player if 
	   //I am firstResp. player
	   ( 
	    (tact->firstResponsiblePlayer==me.number) ||
	    ( (tact->secondResponsiblePlayer==me.number) &&
	      (tact->firstResponsiblePlayer==world->ourGoalieNo) ) 
	    )
	  
 	   ||
	   // I am second resp. and first resp. behind first dangerous opp.
	   ((tact->secondResponsiblePlayer==me.number) &&
	    ((tact->ourTeamDefender(tact->firstResponsiblePlayer)) ||
	     (tact->ourTeamDefMidfield(tact->firstResponsiblePlayer))) &&
	    (tact->firstDangerousOpponent!=UNKNOWN) &&
	    (world->ourTeam[tact->firstResponsiblePlayer].pos.getX()>
	     world->otherTeam[tact->firstDangerousOpponent].pos.getX()))
 	   )

	   &&*/
/*	  !(
	    // der Spieler soll nicht hingehen, wenn beide schon hinter
	    //dem Gegner sind, er selbst ein Abwehrspieler ist 
	    //(der Mittelfeldspieler hat schon abgesichert und wird 
	    //wahrscheinlich der hintere sein) und er nicht firstRespPlayer
	    //ist
	  ((world->ourTeam[tact->firstResponsiblePlayer].pos.getX()>
	     world->otherTeam[tact->firstDangerousOpponent].pos.getX()) &&
	      ((world->ourTeam[tact->secondResponsiblePlayer].pos.getX()>
		world->otherTeam[tact->firstDangerousOpponent].pos.getX())) &&
	      (me.number!=tact->firstResponsiblePlayer) &&
	      (tact->ourTeamDefender(me.number))
	   )) &&
*/
	  (world->playMode==PLAY_MODE_PLAY_ON))
	{
	  //Bedingung nochmal berdenken; Abwehrspieler drfen sich nicht
	  //zu weit rausziehen lassen


	  action=PLAYER_ATTACK;

	}
      /*      else if ( ( tact->ourTeamDefender( me.number ) ) && ( tact->firstResponsiblePlayer == me.number )
		&& ( world->playMode == PLAY_MODE_PLAY_ON) )
	{
	  //defenders don't intercept, this is too dangerous if any opponent is around
	  //PLAYER_ATTACK does nearly the same as intercept, but takes opponents into account
	  action = PLAYER_ATTACK;
	  
	  g.set_PERFECT();
	  }*/
      else if (evalSupportRespPlayer())
	{
	  action=SUPPORT_RESP_PLAYERS;
#ifdef POSACTIONTRACE
	  ostrstream temp;
	  temp << "posm: supp. player no: " << SupportPlayerNo << char(0);
	  char * str=temp.str();
	  PACTIONTR(str);
	  delete str;
#endif
	}
      else if (searchFreePosition())
      {
	  action=GOTO_FREE_POSITION;
      }
      
      else if (shouldICover())
	  {
	      action=COVER_OPPONENT;
	  }
       
    }
  else 
      g =  goalieEvaluate();

#ifdef POSACTIONTRACE
  switch (action)
    {
    case DEFAULT : PACTIONTR("planned action: default");
      break;
    case NO_ACTION: PACTIONTR("planned action: no action");
      break;
    case SUPPORT_RESP_PLAYERS :PACTIONTR("planned action: supp. resp. player");
      break;
    case PLAYER_ATTACK: PACTIONTR("planned action: player attack");
      break;
    case GOTO_FREE_POSITION: PACTIONTR("planned action: free pos.");
      break;
    case COVER_OPPONENT: PACTIONTR("planned action: cover opp.");
      break;
    default: PACTIONTR("planned action: unknown");
    };
#endif


  EvalGrade=g;//I want to know the grade for the turnHeadEvaluate
  returnGRADE(g);
}

bool PositionModule::evalSupportRespPlayer()
{
  int a;
  int b;
  bool act=false;
  int resp[3];

  resp[0]=tact->firstResponsiblePlayer;
  resp[1]=tact->secondResponsiblePlayer;
  resp[2]=tact->thirdResponsiblePlayer;
#ifdef POSACTIONTRACE
  ostrstream temp;
  temp << "evalSupport-> watching player: " << positionEval->GetPlayerWatchNo(world,me.number,0) 
       << ", " << positionEval->GetPlayerWatchNo(world,me.number,1) << char(0);
  char * str=temp.str();
  PACTIONTR(str);
  delete str;
#endif

  for(b=0;(b<3) && (!act);b++)
    for (a=0;(a<1) && (!act);a++)// only first resp.
	if ((resp[a]==positionEval->GetPlayerWatchNo(world,me.number,b)) &&
	    (resp[a]!=UNKNOWN))
	    act=true;
	    
  SupportPlayerNo=positionEval->GetPlayerWatchNo(world,me.number,b-1);
  if (SupportPlayerNo!=UNKNOWN) lookObject=&world->ourTeam[SupportPlayerNo];
  return act;
}

bool PositionModule::shouldICover(){
    if (world->my.tacticalNo == UNKNOWN) return false;
    int opp2cover =world->otherTeamClosestTo(me.pos);
    if (opp2cover == UNKNOWN) return false;
    if (me.pos.getX()>-FIELD_LENGTH+PENALTY_AREA_LENGTH+10) return false;
    if (tact->ourTeamOffender(world->my.number)) return false;
    if (tact->ourTeamOffMidfield(world->my.number)) return false;
    if ( (tact->ourAction.state == TA_CONTROL) && (tact->ourBallControl == BC_SAFE) 
	 && (tact->ourTeamDefMidfield(world->my.number)) )return false;
    if (world->ourTeamClosestTo(world->otherTeam[opp2cover].pos) != me.number) return false;
    if (world->otherTeam[opp2cover].age > 5) return false;
    if ( (world->otherTeam[opp2cover].state != VISUAL) ) return false;
    if (me.distTo( world->otherTeam[opp2cover].pos ) > 10) return false;
    return true;
}

bool PositionModule::searchFreePosition() 
{ 

  //  if (positionEval->IsInPlayerArea(me.number,me.pos) )
  //  {
  if (((world->playMode == PLAY_MODE_FREE_KICK_US) ||(world->playMode == PLAY_MODE_KICK_IN_US)||
       (world->playMode == PLAY_MODE_CORNER_KICK_US)) && 
      (tact->firstResponsiblePlayer!=UNKNOWN) && (world->ourGoalieNo!=UNKNOWN) &&
      (tact->firstResponsiblePlayer!=world->ourGoalieNo))
	{
	  if (world->ball.pos.getX()<me.pos.getX())
	    {
              lastActive = world->time; 
              return true ;
	    }
	  else
	    if (((world->ball.pos.getX()-FREEPOSAREA_LENGHT)< me.pos.getX()) && 
                (world->ball.pos.getY()-FREEPOSAREA_WIDTH<me.pos.getY())&&(world->ball.pos.getY()+FREEPOSAREA_WIDTH>me.pos.getY())) 
	      {
		lastActive=world->time;
	        return true ;
              }
	    else 
	      {
	
#ifdef FREEPOSTRACE
  ostrstream tmp;
  tmp << "NOT FreePos -> Player too far away " << char(0) ;
  char * str = tmp.str();
  FREEPOSTRC(MLog(str));
  delete str;
#endif
		return false ;
	      }
	} 
      else
        if((world->playMode == PLAY_MODE_PLAY_ON)&&
          ((tact->ourAction.state == TA_CONTROL ) || 
	  (tact->ourAction.state == TA_PASS)))
          if ((tact->ourTeamOffender(me.number) )||(tact->ourTeamOffMidfield(me.number)))
	    {
	      lastActive = world->time;
              return true ;
	    }
	  else 
	   {	  
	  
#ifdef FREEPOSTRACE
  ostrstream tmp;
  tmp << "NOT FreePos -> Defense Player " << char(0) ;
  { char * str= tmp.str();
  FREEPOSTRC(MLog(str));
  delete str; }
#endif
	     return false ; 
	   }
        else 
         {

#ifdef FREEPOSTRACE
  ostrstream tmp;
  tmp << "NOT FreePos -> Action not control or pass " << char(0) ;
  { char * str = tmp.str();
  FREEPOSTRC(MLog(str));
  delete str; }
#endif
	    return false ; 
	 }
      // }
      // else
      //{
      // LastBestPos.setX(1000) ;
      //LastBestPos.setY(1000) ;
      //#ifdef FREEPOSTRACE
      //ostrstream tmp;
      //  tmp << "NOT FreePos -> Player not in Playerarea " << char(0) ;
      // char * str = tmp.str();
      // FREEPOSTRC(MLog(str));
      // delete str; 
      //#endif
      //return false ;
      // }
}

TurnHeadWish PositionModule::turnHeadEvaluate()
{
  TurnHeadWish thw;
  thw.lookNowhere( Grade::IMPOSSIBLE, VIEW_WIDTH_NARROW );
  Grade g;
  g.set_GOOD();

  Vector2d thPos[10];
  int thPosCount = 0;

  if( !didIAct )
    {
      g.downgrade();
      // some variables have to be reset if we didn't act
      if( action != PLAYER_ATTACK )
	taReportTime=0;
      if( action != GOTO_FREE_POSITION )
	LastBestPos.setXY( 1000, 1000 );
    }

  // Goalie looks at the ball most of the time
  if( isGoalie ) 
    {
      if( world->ball.inOurPenaltyArea() || world->ball.distTo( world->ourGoal ) < 0.5 * PENALTY_AREA_WIDTH )
	{
	  thw.lookAtBall( g, VIEW_WIDTH_NARROW );
	  PACTIONTR(MLog("posm: turning head to ball (narrow)"));
	}
      else if( world->ball.age > 0 )
	{
	  thw.lookAtBall( g, VIEW_WIDTH_WIDE );
	  PACTIONTR(MLog("posm: turning head to ball"));
	}
      else
	{
	  thw.lookAtPos( g, world->ball.pos + Vector2d( FIELD_LENGTH / 2.0, 0.0 ), VIEW_WIDTH_WIDE );
	  PACTIONTR(MLog("posm: looking around"));
	}
      return thw;
    }

  // now evaluate field player
  switch( action )
    {

    case PLAYER_ATTACK:
      if ( didIAct )
	return plAttackThw;

      if( tact->firstDangerousOpponent!=UNKNOWN && world->ball.age <= 2 && world->ball.state == VISUAL &&
	  fabs( me.bodyAngleTo( world->otherTeam[tact->firstDangerousOpponent].pos ) ) < 
	  fabs( DEG_TO_RAD( MAXNECKANG + 15 ) ) )
	{
	  thw.lookAtOppPlayer( g, tact->firstDangerousOpponent, VIEW_WIDTH_NARROW );
	  PACTIONTR(MLog("posm: turning head to firstDangerousOpponent"));
	}
      else
	{
	  thw.lookAtBall( g, VIEW_WIDTH_NARROW );
	  PACTIONTR(MLog("posm: turning head to ball"));
	}
      break;

    case TACKLE:
      // I am not allowed to move, and I tackled, so I don't know
      // the new speed of the ball. Therefore I look at the ball, since
      // I can't do anything useful
      // I should use VIEW_WIDTH_WIDE because I can not turn my head
      thw.lookAtBall( didIAct ? Grade::PERFECT : Grade::GOOD, VIEW_WIDTH_WIDE );
      PACTIONTR(MLog("posm: TACKLE -> turning head to ball"));
      break;

    default:
      // add positions of objects I want to look at to thPos array
      // the order of the positions determines their importance

      // always add ball
      //if ( world->ball.age > 2 )
	{
	  thPos[thPosCount] = world->ball.pos;
	  thPosCount++;
	  // also add ball in the next cycle
	  thPos[thPosCount] = world->ball.pos + world->ball.v;
	  thPosCount++;
	}      

      if( coveredOpponent != UNKNOWN )
	{
	  thPos[thPosCount] = world->otherTeam[coveredOpponent].pos;
	  thPosCount++;
	}
      if( lookObject != NULL )
	{
	  thPos[thPosCount] = (*lookObject).pos;
	  thPosCount++;
	}
      if( !( lookPos == Vector2d( 0, 0 ) ) )
	{
	  thPos[thPosCount] = lookPos;
	  thPosCount++;
	}
      if( tact->ourTeamOffender( me.number ) && OppDef1 > 0 && world->otherTeam[ OppDef1 ].age > 3 )
	{
	  thPos[thPosCount] = world->otherTeam[OppDef1].pos;
	  thPosCount++;
	}
      if( tact->ourTeamOffender( me.number ) && OppDef2 > 0 && world->otherTeam[ OppDef2 ].age > 3 )
	{
	  thPos[thPosCount] = world->otherTeam[OppDef2].pos;
	  thPosCount++;
	}
      int closest = world->otherTeamClosestTo( me.pos );
      if( closest != UNKNOWN && ( world->otherTeam[closest].state != VISUAL || world->otherTeam[closest].age > 2 ) )
	{
	  thPos[thPosCount] = world->otherTeam[closest].pos;
	  thPosCount++;
	}
      if( tact->firstDangerousOpponent != UNKNOWN &&
	  ( world->otherTeam[tact->firstDangerousOpponent].state != VISUAL ||
	    world->otherTeam[tact->firstDangerousOpponent].age > 2 ) )
	{
	  thPos[thPosCount] = world->otherTeam[tact->firstDangerousOpponent].pos;
	  thPosCount++;
	}
      if( tact->secondDangerousOpponent != UNKNOWN &&
	  ( world->otherTeam[tact->secondDangerousOpponent].state != VISUAL ||
	    world->otherTeam[tact->secondDangerousOpponent].age > 2 ) )
	{
	  thPos[thPosCount] = world->otherTeam[tact->secondDangerousOpponent].pos;
	  thPosCount++;
	}

      // desired view width
      view_width vWidth = VIEW_WIDTH_WIDE;

      // body dir in the next cycle
      Angle myNextBodyDir = me.bodyDir + Angle( skills->getTurnHeadTurn() / me.inertiaMoment() );

      // minimum and maximum head angle (all calculations with double!)
      double minHeadAngle = DEG_TO_RAD( MINNECKANG );
      double maxHeadAngle = DEG_TO_RAD( MAXNECKANG );

      // if first desired position is close, check if it is visible with VIEW_WIDTH_NORMAL
      if( me.distTo( thPos[0] ) < 15.0 )
	{
	  double objAngle = ( thPos[0] - world->my.pos ).getAngle() - myNextBodyDir;
	  if( objAngle > minHeadAngle - fabs( DEG_TO_RAD( 45.0 ) ) && 
	      objAngle < maxHeadAngle + fabs( DEG_TO_RAD( 45.0 ) ) )
	    vWidth = VIEW_WIDTH_NORMAL;
	}

      double vAngle = world->getViewAngle( vWidth );

      // number of desired positions that will be in next view cone
      int numPositions = 0;

      // now try to determine optimum head angle
      for( int i = 0; i < thPosCount; i++ )
	{
	  double objAngle = ( thPos[i] - world->my.pos ).getAngle() - myNextBodyDir;
	  if( objAngle > minHeadAngle - vAngle && objAngle < maxHeadAngle + vAngle )
	    {
	      numPositions++;
	      if( objAngle < maxHeadAngle - vAngle )
		maxHeadAngle = objAngle + vAngle;
	      if( objAngle > minHeadAngle + vAngle )
		minHeadAngle = objAngle - vAngle;
	    }
	}

      if( numPositions > 0 )
	{
	  thw.lookAtPos( g, me.pos + radAng2Vector2d( VISIBLE_DISTANCE + 1.0,
						      myNextBodyDir + Angle( 0.5 * ( maxHeadAngle + minHeadAngle ) ) ),
			 vWidth );

	  POSTRACE( "found " << numPositions << " of " << thPosCount << " positions to look at" );
#ifdef POSACTIONTRACE
	  ostrstream temp;
	  temp << "posm: TURNHEADEVAL-> found " << numPositions << " of " 
	       << thPosCount << " positions to look at" << char(0);
	  char * str=temp.str();
	  PACTIONTR(str);
	  delete str;
#endif
	}
      else
	{
	  thw.lookAround( g, vWidth );
	  POSTRACE( "found no positions to look at, looking around" );
	  PACTIONTR( MLog( "posm: TURNHEADEVAL-> found no positions to look at, looking around" ) );
	}
    } // switch

  return thw;
}


//////////////////////////////////////////////////////////////////////////
//                 The Goalie                                           //
///////////////////////////////////////////////////////////////////////// 
#include "PositionGoalieEval.C"


bool PositionModule::act()
{
    didIAct=1; //for turnHeadEval

    POSDRAWTRACE(POSITION(posGoal));
    POSTRACE("Going to position: " << posGoal);
    POSTRACE("priority " << priority << "; Przision: " << goToPrecision << 
	     "; Stamina to use: " << calcUseableStamina() << 
	     "; overall Stamina:" << world->my.stamina);
    if (isGoalie) return GoalieAct();


    switch(action)
      {
      case TACKLE:
	{
	  skills->noAction();
	  return (true);
	  break;
	};
      case PLAYER_ATTACK:
	{
	  
	  if ((tact->ourAction.state != TA_REACT) ||
	      (taReportTime>6) || (taReportTime==0))
	    {
	      // same ta as before. Occasionally remind others
	      // else be quiet
	      tact->reportOurTacticalAction(TA_REACT, me.number);
	      taReportTime=1;
	      PACTIONTR(MLog("act->reporting TA"));
	    }
	  else 
	    {
	      taReportTime+=1;
	      PACTIONTR(MLog("act->not reporting TA because of taReportTime"));
	    };
	  
	  playerAttack(tact->firstDangerousOpponent);
	  //stdact();
	  break;
	}
	case COVER_OPPONENT: 
	    posGoal=coverOpponent();
	    stdact();
	    break;
	case SUPPORT_RESP_PLAYERS: 
	    SupportPlayer();
	    stdact();
	    break;
	case GOTO_FREE_POSITION:
	    // gotoFreePos();
	    posGoal=gotoFreePos();
	    stdact();
	    break;
	default:
	    stdact();
    };

    return true;
}

void PositionModule::stdact()
{
#ifdef POSACTIONTRACE
    ostrstream temp;
    temp << "Goint to position: "<< posGoal<<" precision: " << goToPrecision
	 << " with priority: " << priority   
	 << " Stamina to use: " << calcUseableStamina() << char(0);
    { char * str= temp.str();
    PACTIONTR(MLog(str));
    delete str; }

    Comm::sendMCmd(CIRCLE(posGoal,goToPrecision));
#endif
    // keep care of standard situation areas
    respectStandardSituations();
    // avoid player collisions
    avoidPlayerCollision();
    AdvSkills::TurtleCommand tc = skills->goToPosition(posGoal, goToPrecision, calcUseableStamina());

    if(tc.state == AdvSkills::TFINISHED)
      {
	// if nothing to do, turn body to ball
	AdvSkills::TurnCommand tuc = skills->turnTo(world->ball);
	if(tuc.state == AdvSkills::RFINISHED)
	  skills->noAction();
	else
	  skills->exec(tuc);
      }
    else
      skills->exec(tc);
}


////////////////////////////////////////////////////////////////////
//                                                                //
// functions used by more than one eval-function                  //
//                                                                //
////////////////////////////////////////////////////////////////////


bool PositionModule::otherPlayerFree(int num, double distance)
{
    // opponent "num" valid ? and our nearest is known
    if(
	(world->otherTeam[num].age < MAXAGE)
	&&
	(world->ourTeamClosestTo(world->otherTeam[num].pos) != UNKNOWN)
	){
	// in standard situations players near to the ball are NOT free
	if(
	    (world->playMode == PLAY_MODE_KICK_IN_THEM)
	    ||
	    (world->playMode == PLAY_MODE_FREE_KICK_THEM)
	    ||
	    (world->playMode == PLAY_MODE_CORNER_KICK_THEM)
	    ||
	    (world->playMode == PLAY_MODE_GOAL_KICK_THEM)
	    ){
	    if(dist(world->otherTeam[num].pos, world->ball.pos) < FREE_KICK_FORBIDDEN_AREA)
		return false;
	}
    
	// our player near to opponent "num"
	if (dist(world->otherTeam[num].pos,
		 world->ourTeam[world->ourTeamClosestTo(world->otherTeam[num].pos)].pos)
	    < distance){ 
	    // if our player is not nearer to another player
	    if  (world->otherTeamClosestTo(world->ourTeam[world->ourTeamClosestTo(world->otherTeam[num].pos)].pos)
		 == num)
		return false;
	    else
		return true;
	}
    }

    // in doubt, return true
    return true;
}


bool PositionModule::ourPlayerFree(int num, double distance){

    // player valid, opponent near ?
    if ((world->ourTeam[num].age < MAXAGE)
	&&(world->otherTeamClosestTo(world->ourTeam[num].pos) != UNKNOWN))
	if (dist(world->ourTeam[num].pos,
		 world->otherTeam[world->otherTeamClosestTo(world->ourTeam[num].pos)].pos)
	    < distance)
	    return false;
	else
	    return true;
    // in doubt, return true
    return true;  
}


double PositionModule::calcUseableStamina(){
    double useableStam=max_useable_Stamina[priority];
    double maxStam=world->my.stamina-(1+save_Stamina[priority])*
	EFFORT_DEC_THR*STAMINA_MAX;
		 
    return max(min(useableStam,maxStam),0.);
}


void PositionModule:: CalcStdPriority()
{
    priority=2; //standard: not to fast and not to slow
    if (world->playMode==PLAY_MODE_PLAY_ON)
    {
	priority=1; 
	if ((tact->ourTeamDefender(me.number)) && 
	    (world->ball.pos.getX()<me.pos.getX()))
	    priority=0;
	//defenders have to hurry if the ball is behind them
	/*
	if ((tact->ourTeamOffender(me.number)) &&
	    (me.pos.getX()-5<world->ourOffsideLine(MAXAGE)) &&
	    (positionEval->GetOffensive()<.5))
	    priority=3;
	*/
	//offenders may save stamina if the ball is in our defense
	   
    }
};

////////////////////////////////////////////////////////////////////
//                                                                //
//              function used by act                              //
//                                                                //
////////////////////////////////////////////////////////////////////



void PositionModule::respectStandardSituations()
{
    switch(world->playMode){
	case PLAY_MODE_UNKNOWN : 
	    break;
	case PLAY_MODE_BEFORE_KICK_OFF :
	    posGoal.setX(min(posGoal.getX(),-PLAYER_SIZE));
	    break;
	case PLAY_MODE_TIME_OVER :
	    posGoal = me.pos;
	    break;
	case PLAY_MODE_PLAY_ON : 
	    break;
	case PLAY_MODE_KICK_OFF_US : 
	case PLAY_MODE_KICK_OFF_THEM :
	    posGoal.setX(min(posGoal.getX(),-PLAYER_SIZE));    
	    break;
	case PLAY_MODE_KICK_IN_US : 
	    break;
	case PLAY_MODE_KICK_IN_THEM : 
	    if((world->ball.inCorridor(me.pos, posGoal, FREE_KICK_FORBIDDEN_AREA))
	       &&(dist(world->ball.pos,me.pos) < 1.1*FREE_KICK_FORBIDDEN_AREA)) // [C]
		if((posGoal - me.pos).getAngle() - (world->ball.pos - me.pos).getAngle() > 0)//right
		    posGoal = (world->ball.pos - me.pos).perp()*(1/norm(world->ball.pos - me.pos));
		else // left
		    posGoal = (me.pos - world->ball.pos).perp()*(1/norm(me.pos - world->ball.pos));
	    break;
	case PLAY_MODE_FREE_KICK_US : 
	    break;
	case PLAY_MODE_FREE_KICK_THEM :
	    if( (!world->ball.inOtherPenaltyArea()) 
		&& (world->ball.inCorridor(me.pos, posGoal, FREE_KICK_FORBIDDEN_AREA))
		&&(dist(world->ball.pos,me.pos) < 1.1*FREE_KICK_FORBIDDEN_AREA)) // [C]
	      {
		if (tact->ourTeamDefender(me.number))
		  { 
		    //look for position at our offside line
		    double y=sqr(1.1 * FREE_KICK_FORBIDDEN_AREA)+sqr(world->ball.pos.getX()-me.pos.getX());
		    if (y>0)
		      posGoal.setY(world->ball.pos.getY()+
				   sign( - world->ball.pos.getY() + me.pos.getY() ) * sqrt(y));
		    else cerr << "Scheie, negative Wurzel. Schimpf mit Martin!!!!" << endl;
		  };

		if ( (world->ball.inCorridor(me.pos, posGoal, FREE_KICK_FORBIDDEN_AREA))
		    && (dist(world->ball.pos,me.pos) < 1.1*FREE_KICK_FORBIDDEN_AREA)) // [C]
		  if((posGoal - me.pos).getAngle() - (world->ball.pos - me.pos).getAngle() > 0)//right
		    posGoal = (world->ball.pos - me.pos).perp()*(1/norm(world->ball.pos - me.pos));
		  else // left
		    posGoal = (me.pos - world->ball.pos).perp()*(1/norm(me.pos - world->ball.pos));
	      }
	    break;    
	case PLAY_MODE_CORNER_KICK_US : 
	    break;
	case PLAY_MODE_CORNER_KICK_THEM :
	    if((world->ball.inCorridor(me.pos, posGoal, FREE_KICK_FORBIDDEN_AREA))
	       &&(dist(world->ball.pos,me.pos) < 1.1*FREE_KICK_FORBIDDEN_AREA)) // [C]
		posGoal = me.pos;
	    break;
	case PLAY_MODE_GOAL_KICK_US : 
	    break;
	case PLAY_MODE_GOAL_KICK_THEM :
	    if((world->ball.inCorridor(me.pos, posGoal, FREE_KICK_FORBIDDEN_AREA))
	       &&(dist(world->ball.pos,me.pos) < 1.1*FREE_KICK_FORBIDDEN_AREA)) // [C]
		posGoal = me.pos;
	    break;
	case PLAY_MODE_GOAL_US :
	    posGoal = me.pos;
	    break;
	case PLAY_MODE_DROP_BALL :
	    break;
	case PLAY_MODE_GOAL_THEM :
	    posGoal = me.pos;
	    break;    
	case PLAY_MODE_OFFSIDE_US :
	    break;
	case PLAY_MODE_OFFSIDE_THEM :
	    if((world->ball.inCorridor(me.pos, posGoal, FREE_KICK_FORBIDDEN_AREA))
	       &&(dist(world->ball.pos,me.pos) < 1.1*FREE_KICK_FORBIDDEN_AREA)) // [C]
		if((posGoal - me.pos).getAngle() - (world->ball.pos - me.pos).getAngle() > 0)//right
		    posGoal = (world->ball.pos - me.pos).perp()*(1/norm(world->ball.pos - me.pos));
		else // left
		    posGoal = (me.pos - world->ball.pos).perp()*(1/norm(me.pos - world->ball.pos));
	    break;    
	case PLAY_MODE_HALF_TIME :
	    posGoal = me.pos;
	    break;        
	case PLAY_MODE_TIME_EXTENDED:
	    posGoal = me.pos;    
	    break;
    default:
      break;
    }

}


void PositionModule::avoidPlayerCollision()
{
    if (nearestOther!=UNKNOWN)
	posGoal=goAroundOpponent(posGoal,world->otherTeam[nearestOther],2,
				 (int) (me.pos-posGoal).getX());
    if (nearestFriendNumber!=UNKNOWN)
	posGoal=goAroundOpponent(posGoal,world->ourTeam[nearestFriendNumber],2,
				 (int) (me.pos-posGoal).getX());

    // if otherPlayer or ourPlayer (incl. me!) in my way
    /*  if(dist(posGoal, me.pos) > PLAYER_SIZE){
	if(
	(world->otherTeamInCorridor(me.pos, me.pos +
	radAng2Vector2d(PLAYER_SPEED_MAX, (posGoal-me.pos).getAngle()),
	4*PLAYER_SIZE) > 0)
	||
	(world->ourTeamInCorridor(me.pos, me.pos +
	radAng2Vector2d(PLAYER_SPEED_MAX, (posGoal-me.pos).getAngle()),
	4*PLAYER_SIZE) > 1)
	){
	if((world->ball.pos - me.pos).getAngle() > 0) // ball right of me
	// ball right => side step right of player
	// to avoid interference with free kick areas
	posGoal = radAng2Vector2d(PLAYER_SPEED_MAX, (posGoal-me.pos).getAngle())
	+ radAng2Vector2d(5*PLAYER_SIZE, (posGoal-me.pos).getAngle()).perp();
	else // vice versa
	posGoal = radAng2Vector2d(PLAYER_SPEED_MAX, (posGoal-me.pos).getAngle())
	- radAng2Vector2d(5*PLAYER_SIZE, (posGoal-me.pos).getAngle()).perp();
	}
	}
    */
}
