//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////


#ifndef __STANDARDSITUATIONMODULE_H__
#define __STANDARDSITUATIONMODULE_H__

#include "grade.h"
#define FUNSSELMACROS
#include "drawMacros.h"
#include "mcommand.h"
#define MOD_STDSIT
#include "Trace.h"

#include "worldmodel.h"
#include "tacticalmodule.h"
#include "audiomgr.h"
#include "AdvSkills.h"
#include "Config.h"

#include "passeval.h"
#include "goalshoteval.h"
#include "PositionEval.h"
#include <sstream>


// for calculation of player to perform a freekick
const double NOWHERE_DIST = 100.0 * ( FIELD_LENGTH + FIELD_WIDTH );


class StandardSituationModule
{
public:

  StandardSituationModule( Worldmodel * _world,
			   TacticalModule * _tact,
			   AdvSkills * _skills,
			   bool isGoalie,
			   const PassEval *_passEval, 
			   const GoalShotEval *_goalShotEval,
			   const PositionEval *_positionEval );

  ~StandardSituationModule() {};

  // feed information from coach into module (player to perform)
  bool feedbackAudio( AudioMsg &msg );

  // feed information about new world and tactics into module
  bool feedbackWorld();

  // Do the evaluation for the next action (for definition of Grade: see grade.h)
  Grade evaluate();

  // Calculate a turn head wish
  TurnHeadWish turnHeadEvaluate();
  
  // Perform the action for this module
  bool act();

private:

  Worldmodel * world;
  TacticalModule * tact;
  AdvSkills * skills;

  int kickUsPlayer;        // player to perform, selected by coach
  Vector2d kickUsOffset;   // offset to ball position when going to ball
  int goToPhase;           // for the player who is going to perform the kick
  int penaltyPlayer;

  bool actedThisCycle;
  bool wantToTurnHead;
  bool weAreKickingOff;
  bool iAmGoalie;
  bool stdSitFinished;
  int maxAge;
  int timeLeft;   // how much time is left until a drop ball is done by the referre
  play_mode playmode;

  TurnHeadWish thwish;

  // player that will kick
  int actPlayer;

  // needed for GOAL_KICK_US only
  Vector2d targetPos[12];             // the targets where players should walk to (tactical number!)

  int waitingTime;                    // time for the first player to wait for the others

  int standardWaitingTime;            // the time we want to wait before doing a standard 
                                      // situation by default
  int waitingTimeAfterGoalieMove;     // the time the goalie waits after a move

  double maxObstacleDist;             // objects nearer than that are avoided
  double keepAwayDist;                // distance to objects when avoiding them

  // determines how fast the players are running when doing dash-commands
  // they should dash (STAMINA_INC_MAX) * ...Factor if they have enough stamina
  double activeStaminaFactor;

  // for doing goalie's kick off
  int    goalieMovePositions;            // number of possible move positions
  double goalieDistToFrontOfPenArea;     // distances kept to the edges of
  double goalieDistToSideOfPenArea;      // the penalty area
  int    goalieMovesInCurrentSituation;  // number of move commands goalie
                                         // has sent in the current situation

  bool goalieDoesFreeKick;
  bool goalieMayMove;                 // goalie may move in free kick-situation 
                                      // (should be true after a catch-command)

  // the evaluation methods of the agent's modules
  const GoalShotEval *goalShotEval;
  const PassEval     *passEval;
  const PositionEval *positionEval;

  // the position the goalie has to move to when starting a penalty kick
  Vector2d penaltyKickGoaliePos;

  // the position the field player has to move to when starting a penalty kick
  Vector2d penaltyKickFieldplayerPos;

  // the number and type of players for stdsits
  struct playerTypeCount {
    int defenders; 
    int midfield;     
    int offenders; 
  } players[3];   // 3 parts of the field

  // calcs the number of each player type to do a stdsit 
  void calcDesiredPlayerTypes( const Vector2d &target, int &noDef, int &noMid, int &noOff ) const;

  // pushes one player to a distance of 9.15 around the given "pos".
  void pushPlayer( WPlayer &player, const Vector2d &pos ) const;

  // calculates where the goalie should move before a free kick
  void calcGoalieMove( Vector2d &moveTarget ) const;

  // walks to target using stamina avoiding obstacles (ball and other players)
  // true if still walking
  // false if target reached and turned to ball
  // if false no action is sent (even not skills->noAction()!)
  bool goingTo( Vector2d &target, const double stamina );

  // calculates alternative target to avoid ball and other players
  void avoidObstacles( Vector2d &target ) const;

  // does all things necessary for kick-off situations (US+THEM)
  bool doingKickOff();

  // does (field player) free kicks, kick ins and corner kicks
  bool doingStdKickUs();

  // calc closest players to target
  // depends on our tactical system and the field part ball is in
  // goalie is never a result
  int calcClosestDesiredPlayerTo( const Vector2d &target );

  // lets tacticalmodule turn player's head to gather information
  void lookAround();

  // goalie doing goalkick
  bool doingGoalieGoalKickUs();

  // fieldplayer finishing goalkick
  bool doingFieldPlayerGoalKickUs();

  // goalie doing free kick after catch
  bool doingGoalieFreeKickUs();

  // does the evaluation for PLAY_MODE_FREE_KICK_US
  Grade evalFreeKickUs();

  // does the evaluation for PLAY_MODE_KICK_IN_US and PLAY_MODE_CORNER_KICK_US
  Grade evalRemainingKicksUs();

  //
  void doPenaltyShootOutOurSetup();
  void doPenaltyShootOutTheirSetup();
  void doPenaltyShootOut();

}; // class StandardSituationModule

#endif
