//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2003 competition
//
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////


void 
StandardSituationModule::calcGoalieMove( Vector2d &moveTarget ) const
{
  moveTarget.setXY( -FIELD_LENGTH / 2.0 + PENALTY_AREA_LENGTH - goalieDistToFrontOfPenArea, 0.0 );

  // advanced ourEval
  PropOf props;
  props.grade.set_BAD();
  props.score = passEval->SCORE_POOR + 1;
  PropOf actProps;

  Worldmodel simWorld = *world;

  WPlayer player;

  Vector2d testMovePos;
  Vector2d startPos = moveTarget;

  double yStep = 0.0;
  if( goalieMovePositions > 1 )
    {
      yStep = ( PENALTY_AREA_WIDTH - 2 * goalieDistToSideOfPenArea ) / ( goalieMovePositions - 1.0 );
      startPos.setXY( -FIELD_LENGTH / 2.0 + PENALTY_AREA_LENGTH - goalieDistToFrontOfPenArea,
		      -PENALTY_AREA_WIDTH / 2.0 + goalieDistToSideOfPenArea );
    }

  for( int i = 0; i < goalieMovePositions; i++ )
    {
      testMovePos = startPos + Vector2d( 0.0, i * yStep );

      // copy all players and push them if needed
      for( int playerNumber = 1; playerNumber <= 11; playerNumber++ )
	{
	  player = world->ourTeam[playerNumber];
	  if( playerNumber != world->ourGoalieNo )
	    pushPlayer( player, testMovePos );
	  simWorld.ourTeam[playerNumber] = player;

	  player = world->otherTeam[playerNumber]; 
	  pushPlayer( player, testMovePos );
	  simWorld.otherTeam[playerNumber] = player;
	}

      // move our goalie
      simWorld.ourTeam[simWorld.ourGoalieNo].pos = testMovePos;
      simWorld.my = simWorld.ourTeam[simWorld.ourGoalieNo];

      // look for the best pass
      for( int passPartner = 1; passPartner <= 11; passPartner++ )
	{
	  if( passPartner == simWorld.ourGoalieNo )
	    continue;   // goalie does not want the ball
	  if( simWorld.ourTeam[passPartner].age > maxAge )
	    continue;   // old players are no good !!
	  
	  actProps = passEval->ourEvalProp( &simWorld, tact, world->ourGoalieNo, passPartner );
	  if( actProps.score < props.score )
	    {
	      props      = actProps;
	      moveTarget = testMovePos;
	    }
	}
    }

  // no passes found? -> calc emergency move position
  if( props.grade.is_worse_than( Grade::POOR ) )
    {
      Vector2d leftPos( -FIELD_LENGTH / 2.0 + PENALTY_AREA_LENGTH - goalieDistToFrontOfPenArea,
			-PENALTY_AREA_WIDTH / 2.0 + goalieDistToSideOfPenArea );

      Vector2d rightPos( -FIELD_LENGTH / 2.0 + PENALTY_AREA_LENGTH - goalieDistToFrontOfPenArea,
			 PENALTY_AREA_WIDTH / 2.0 - goalieDistToSideOfPenArea );

      // [C]*
      double valueLeft =
	- 3.0 * world->otherTeamInCorridor( leftPos, leftPos + Vector2d( FIELD_LENGTH / 3.0, 0.0 ),
					    FIELD_WIDTH / 3.0 )
	+ 1.0 * world->ourTeamInCorridor( leftPos, leftPos + Vector2d( FIELD_LENGTH / 3.0, 0.0 ),
					  FIELD_WIDTH / 3.0 );

      double valueRight =
	- 3.0 * world->otherTeamInCorridor( rightPos, rightPos + Vector2d( FIELD_LENGTH / 3.0, 0.0 ),
					    FIELD_WIDTH / 3.0 )
	+ 1.0 * world->ourTeamInCorridor( rightPos, rightPos + Vector2d( FIELD_LENGTH / 3.0, 0.0 ),
					  FIELD_WIDTH / 3.0 );

      moveTarget = ( valueLeft > valueRight ? leftPos : rightPos );
      DECISION( "no passes found, emergency move to " << moveTarget );
    }
  else
    {
      DECISION( "goalie will move to " << moveTarget );
      MCMD( POSITION(moveTarget) );
      DECISION( "found pass partner " << props.number << " at " << props.pos << ", grade " << props.grade );
      MCMD( POSITION(props.pos) );
      MCMD( SLINE(moveTarget,props.pos,MDash) );
    }
}


void
StandardSituationModule::pushPlayer( WPlayer &player, const Vector2d &pos ) const
{
  if( player.age > maxAge || player.distTo( pos ) > FREE_KICK_FORBIDDEN_AREA )
    return;

  // player has to be pushed outside the circle and his speed has to be set to 0.
  player.v.setXY( 0.0, 0.0 );
  Vector2d distToPos = player.pos - pos;

  if( distToPos.getLength() < 0.1 )   // [C]
    {
      // player is very close, so place him in front of me
      player.pos.setXY( pos.getX() + FREE_KICK_FORBIDDEN_AREA, pos.getY() );
    }
  else
    {
      // player is not so close, so push him the server-like way
      distToPos.stretch( FREE_KICK_FORBIDDEN_AREA / distToPos.getLength() );
      player.pos = pos + distToPos;
    }
}


bool 
StandardSituationModule::doingGoalieGoalKickUs()
{
  // report tactical action
  tact->reportOurTacticalAction( TA_CONTROL, world->my.number );

  // go to ball first
  if( !world->my.ballKickable( world->ball ) )
    {
      DECISION( "Goalie wants to do the goalkick, but has to go to the ball first" );
      DECISION( "dist2Ball = " << norm( world->my.pos - world->ball.pos ) );
      Vector2d targetPos = world->ball.pos - Vector2d( world->my.size + BALL_SIZE, 0.0 );

      if( goingTo( targetPos, STAMINA_INC_MAX ) )
	{
	  DECISION( "Goalie going to " << targetPos );
	}
      else
	{
	  DECISION( "goalie has reached ball, waitingTime = " << waitingTime );
	  skills->noAction();
	}
      lookAround();
      waitingTime = standardWaitingTime;
      return true;
    }

  waitingTime--;

  // - now look to other goal (waitingTime > 0)
  // - turn head until one of our players is in the penalty area (waitingTime == 0)
  // - kick ball to one of our players in our penalty area (turn body to ball first)

  if( waitingTime > 0 )
    {
      DECISION( "Goalie wants to wait because waitingTime = " << waitingTime ); 
      AdvSkills::TurnCommand tcmd = skills->turnTo( world->otherGoal );	  
      if( tcmd.state != AdvSkills::RFINISHED )
	{
	  skills->exec( tcmd );
	  DECISION( "turning to other goal" ); 
	}
      else
	skills->noAction();

      lookAround();
      return true;
    }

  // goalie now looks for our players in our penalty area

  Grade playerGrades[12][12];
  int ourPlayer;

  for( ourPlayer = 1; ourPlayer <= 11; ourPlayer++ )
    {
      for( int j = 0; j <= 11; j++ )
	playerGrades[ourPlayer][j].set_IMPOSSIBLE();

      if( ourPlayer == world->my.number )
	continue;
      if( world->ourTeam[ourPlayer].age > maxAge )
	continue;
      if( !world->ourTeam[ourPlayer].inOurPenaltyArea() )
	continue;

      // check for the player all other players as passpartners
      for( int passPartner = 1; passPartner <= 11; passPartner++ )
	{
	  if( passPartner == ourPlayer )
	    continue;
	  if( passPartner == world->ourGoalieNo )
	    continue;
	  if( world->ourTeam[passPartner].age > maxAge )
	    continue;
	  playerGrades[ourPlayer][passPartner] = passEval->ourEval( world, tact, ourPlayer, passPartner );
	}
    }

  int gradeCount[12][6] = {{0}};

  for( int i = 1; i <= 11; i++ )
    for( int j = 1; j <= 11; j++ )
      {
	if( playerGrades[i][j].is_POOR() )
	  gradeCount[i][0]++;
	else if( playerGrades[i][j].is_AVERAGE() )
	  gradeCount[i][1]++;
	else if( playerGrades[i][j].is_GOOD() )
	  gradeCount[i][2]++;
	else if( playerGrades[i][j].is_PERFECT() )
	  gradeCount[i][3]++;
      }

  int count = 0;
  for( int i = 1; i <= 11; i++ )
    if( gradeCount[i][3] > count )
      {
	count = gradeCount[i][3];
	ourPlayer = i;
	DECISION("found passpartner " << i << " with " << count << " PERFECT passes");
      }

  if( count == 0 )
    for( int i = 1; i <= 11; i++ )
      if( gradeCount[i][2] > count )
	{
	  count = gradeCount[i][2];
	  ourPlayer = i;
	  DECISION("found passpartner" << i << " with " << count << " GOOD passes");
	}

  if( count == 0 )
    for( int i = 1; i <= 11; i++ )
      if( gradeCount[i][1] > count )
	{
	  count = gradeCount[i][1];
	  ourPlayer = i;
	  DECISION("found passpartner" << i << " with " << count << " AVERAGE passes");
	}

  if( count == 0 )
    for( int i = 1; i <= 11; i++ )
      if( gradeCount[i][0] > count )
	{
	  count = gradeCount[i][0];
	  ourPlayer = i;
	  DECISION("found passpartner" << i << " with " << count << " POOR passes");
	}

  if( count != 0 )
    {
      // check if it is necessary to turn to ball
      AdvSkills::TurnCommand tcmd = skills->turnTo( world->ball );
      if( tcmd.state != AdvSkills::RFINISHED )
	{
	  skills->exec( tcmd );
	  DECISION( "Goalie turning to ball, passpartner is no. " << ourPlayer
		    << " at " << world->ourTeam[ourPlayer].pos << ", waitingTime=" << waitingTime );
	  thwish.lookAtOurPlayer( Grade::ABSOLUTE, ourPlayer, VIEW_WIDTH_NARROW );
	  wantToTurnHead = true;
	  waitingTime = 1;
	  return true;
	}

      AdvSkills::ExtKickCommand kcmd = skills->kick( 0.6 * skills->stopableV, world->ourTeam[ourPlayer].pos, 1 );

      // wait until player has reached his position
      if( world->ourTeam[ourPlayer].distTo( targetPos[world->ourTeam[ourPlayer].tacticalNo] ) < 1.0 )
	{
	  skills->exec( kcmd );
	  thwish.lookAtOurPlayer( Grade::ABSOLUTE, ourPlayer, VIEW_WIDTH_NARROW );
	  wantToTurnHead = true;
	  DECISION( "Goalie kicking, waitingTime = " << waitingTime );
	  DECISION( "pass to no. " << ourPlayer << " at " << world->ourTeam[ourPlayer].pos );
	  tact->reportOurTacticalAction( TA_PASS, world->my.number, ourPlayer );

	  waitingTime = -1;
	  stdSitFinished = true;
	  DECISION( "Standardsituation finished" );
	  return true;
	}

      if( timeLeft > 0 )
	{
	  skills->noAction();
	  lookAround();
	  DECISION( "Goalie looking around, waitingTime = " << waitingTime );
	  waitingTime = 1;
	  return true;
	}
    }

  // no passpartners found -> perform emergency kick
  Vector2d kickTarget( -FIELD_LENGTH / 2.0 + 5.0,
		       world->ball.pos.getY() < 0.0 ? -FIELD_WIDTH / 2.0 : FIELD_WIDTH / 2.0 );
  MCMD( POSITION(kickTarget) );
  DECISION( "Goalie found no passpartners, kicking to " << kickTarget );
  skills->exec( skills->kick( BALL_SPEED_MAX, kickTarget, 1 ) );
  waitingTime = -1;
  return false;
}


bool
StandardSituationModule::doingGoalieFreeKickUs()
{
  if( !iAmGoalie )
    {
      if( world->my.inOurPenaltyArea() )
	{
	  DECISION( "I have to keep out of our penalty area" );
	  Vector2d posEvalTarget = positionEval->GetPlayerPosition( world, world->my.number );

	  if( inOurPenaltyArea( posEvalTarget ) )
	    posEvalTarget.setX( -FIELD_LENGTH/2. + PENALTY_AREA_LENGTH + 8. ); // [C]

	  AdvSkills::TurtleCommand tcmd = skills->goToPosition( posEvalTarget, 1.0 );
	  if( tcmd.state != AdvSkills::TFINISHED )
	    skills->exec( tcmd );
	  else
	    skills->noAction();
	}
      else
	{
	  DECISION( "I am not goalie, doing nothing" );
	  skills->noAction();
	}
      lookAround();
      return false;
    }

  // I am Goalie: now do the freekick after a catch
  waitingTime--;

  // wait a while and collect information
  if( waitingTime > 0 )
    {
      // report tactical action
      tact->reportOurTacticalAction( TA_CONTROL, world->my.number );

      AdvSkills::TurnCommand tcmd =
	skills->turnTo( world->otherGoal );
      if( tcmd.state != AdvSkills::RFINISHED )
	skills->exec( tcmd );
      else
	skills->noAction();
      lookAround();
      DECISION( "Goalie looking around, waitingTime = " << waitingTime );
      return true;
    }

  // check if goalie may move or not
  if( goalieMayMove && goalieMovesInCurrentSituation < min( 1, GOALIE_MAX_MOVES ) )
    {
      Vector2d moveTarget;
      calcGoalieMove( moveTarget );

      if( world->my.distTo( moveTarget ) > 1.0 )   // [C]
	{
	  goalieMovesInCurrentSituation++;

	  DECISION( "goalie moving to " << moveTarget
		    << " goalieMovesInCurrSit: " << goalieMovesInCurrentSituation );
	  skills->basicMove( moveTarget.getX(), moveTarget.getY() );
	  MCMD( POSITION(moveTarget) );
	  lookAround();

	  waitingTime = waitingTimeAfterGoalieMove;   // [C]
	  return true;
	}
    }

  // next time another module should do some action
  waitingTime = -1;
  stdSitFinished = true;
  DECISION( "Standardsituation finished" );
  skills->noAction();
  lookAround();

  return true;
}
